#include "Agent.h"

#include "Junction.h"
#include "Way.h"
#include "utils.h"
#include "Intersection.h"
#include "Pathfinding.h"
#include "AgentSet.h"
#include "SoundEngine.h"
#include "Speaker.h"
#include "MemoryLog.h"

#define DEFAULT_MAXIMUM_ACC 0.15f
#define DEFAULT_MINIMUM_ACC 0.08f
#define DEFAULT_MAXIMUM_DEC 0.25f
#define DEFAULT_MINIMUM_DEC 0.18f
#define DISTANCE_CHECK_FOR_AGENT 10.0f
#define AGENT_LENGHT 0.26f

#define COLOR_RANGE 0.1f
static const D3DXVECTOR3 s_agentColors[] =
{
	D3DXVECTOR3(0.3f, 0.3f, 0.3f),
	D3DXVECTOR3(0.6f, 0.6f, 0.6f),
	D3DXVECTOR3(1.0f, 1.0f, 1.0f),
	D3DXVECTOR3(0.2f, 0.2f, 0.6f),
	D3DXVECTOR3(0.5f, 0.5f, 0.9f),
	D3DXVECTOR3(0.8f, 0.2f, 0.2f),
	D3DXVECTOR3(0.2f, 0.6f, 0.2f)
};

Agent::Agent(Speaker *speaker, Intersection &destination, Way &actualWay, AgentSet& agentSet, const Pathfinding& pathfinding):
m_destination(&destination),
m_currentWay(&actualWay),
m_decelerationPerSecond (0.2f),//randomFloat(DEFAULT_MINIMUM_DEC, DEFAULT_MAXIMUM_DEC)),
m_accelerationPerSecond (0.1f),//randomFloat(DEFAULT_MINIMUM_ACC, DEFAULT_MAXIMUM_ACC)),
m_safetyDistance(0.1f),
m_pathfinding(pathfinding),
m_agentSet(agentSet),
m_speaker(speaker),
m_length(AGENT_LENGHT),
m_currentSpeed(0.0f)
{
	m_currentDestination = &actualWay.getEnd();
	m_distanceFromCurrentJunction = randomFloat(0.0f, actualWay.getLength());
	m_maximumSpeed = actualWay.getSpeedLimit();
	m_maximumSpeedPassingJunction = m_decelerationPerSecond * 0.95f;
	// on rpertorie l'agent sur la route
	actualWay.addAgent(*this);
	m_position = m_currentWay->getBegin().getPosition() 
		+ m_currentWay->getVector() * (m_currentWay->getLength() - m_distanceFromCurrentJunction);

	m_color = s_agentColors[randomInt(0, sizeof(s_agentColors) / sizeof(*s_agentColors))];
	m_color *= randomFloatInRange(1.0f, COLOR_RANGE);

#ifndef NO_SOUND
	m_speaker->attachPosition(m_position);
	m_speaker->setVolume(1.0f);
	m_speaker->play(true);
#endif
}

Agent::~Agent(void)
{
	m_currentWay->removeAgent(*this);
#ifndef NO_SOUND
	delete m_speaker;
#endif
}

void Agent::act(float timeElapsed)
{
	// on regarde si on est arriv  destination
	if (&m_currentDestination->getIntersection() == m_destination)
	{
		// si oui, on envoie le signal  AgentSet, qui va nous replacer ailleur
		m_agentSet.replaceAgent(this);
		return;
	}

	// enfin, on applique l'acclration trouve, puis on calcule le mouvement et on se dplace
	m_currentSpeed += computeAcceleration() * timeElapsed;
	if (m_currentSpeed > m_maximumSpeed)
	{
		m_currentSpeed = m_maximumSpeed;
	}
	else if (m_currentSpeed < 0.0f)
	{
		m_currentSpeed = 0.0f;
	}

	float move = m_currentSpeed * timeElapsed;

	do
	{
		// on avance de la distance
		m_distanceFromCurrentJunction -= move;

		if (m_distanceFromCurrentJunction <= 0.0f)
		{
			// on conserve le mouvement qui nous reste, puis on passe  la route suivante sur le chemin
			move = passToNextWay();
		}
		// sinon, on a encore du chemin  parcourir sur la route, le mouvement a t entirement utilis
		else
		{
			move = 0.0f;
		}
	}
	while (move > 0.0f);

	// on met  jours la position de notre agent
	m_position = m_currentWay->getBegin().getPosition() 
		+ m_currentWay->getVector() * (m_currentWay->getLength() - m_distanceFromCurrentJunction);
}

const D3DXVECTOR3& Agent::getPosition() const
{
	return m_position;
}

const D3DXVECTOR3 & Agent::getColor() const
{
	return m_color;
}

const Junction& Agent::getCurrentDestination() const
{
	return *m_currentDestination;
}

const Junction& Agent::getNextDestination() const
{
	return getNextWay()->getEnd();
}

float Agent::getDistanceFromCurrentJunction() const
{
	return m_distanceFromCurrentJunction;
}

float Agent::getCurrentSpeed() const
{
	return m_currentSpeed;
}

float Agent::getAcceleration() const
{
	return m_accelerationPerSecond;
}

float Agent::getDeceleration() const
{
	return m_decelerationPerSecond;
}

float Agent::passToNextWay()
{
	float move = -m_distanceFromCurrentJunction;
	// on commence par se drfrencer de la way
	m_currentWay->removeAgent(*this);
	
	// sinon, on met les variables internes  jours
	m_currentWay = getNextWay();	// route actuelle
	m_currentWay->addAgent(*this);	// on se rfrence sur cette route
	m_currentDestination = &m_currentWay->getEnd(); // prochaine destination
	m_distanceFromCurrentJunction = m_currentWay->getLength(); // distance restant  parcourir sur la route actuelle
	m_maximumSpeed = m_currentWay->getSpeedLimit(); // nouvelle vitesse limite

	return move;
}

const Agent* Agent::getAgentAheadOnPath() const
{
	const Agent *res = m_currentWay->agentAhead(*this);
	if (res)
	{
		return res;
	}
	float dist = m_distanceFromCurrentJunction;
	unsigned i = 1;
	while (dist < DISTANCE_CHECK_FOR_AGENT)
	{
		Way *nextWay = getNextWay(i);
		if (!nextWay)
		{
			return NULL;
		}
		const Agent* behind = nextWay->agentBehind();
		if (behind)
		{
			return behind;
		}
		// on n'a trouv aucun agent, on continue sur la suite du path
		dist += nextWay->getLength();
		i++;
	}
	
	// on n'a trouv aucun agent sur la distance de vrification
	return NULL;
}

float Agent::getNextDistanceDecelerating(float timeElapsed) const
{
	// on considre ici que notre vhicule ralentit au maximum
	float speed = (m_currentSpeed < m_decelerationPerSecond * timeElapsed) ? 0.0f : m_currentSpeed - (m_decelerationPerSecond * timeElapsed);
	return speed * timeElapsed;
}

const Way* Agent::getCurrentWay() const
{
	return m_currentWay;
}

Way* Agent::getNextWay(unsigned ind) const
{
	Way *nextWay = m_currentWay;
	Junction *current = m_currentDestination;
	Intersection *nextIntersection;
	for (unsigned i = 0; i < ind; i++)
	{
		// si je suis arriv, je retourne la way actuelle
		if (&nextWay->getEnd().getIntersection() == m_destination)
		{
			return nextWay;
		}
		// si je suis sur une voie entre deux junctions diffrentes, il n'y a pas de choix possible
		if (&nextWay->getBegin().getIntersection() != &nextWay->getEnd().getIntersection())
		{
			nextWay = nextWay->getEnd().getWaysOut()[0];
		}
		else // je suis dans une intersection
		{
			// je regarde quelle way il me faut atteindre
			nextIntersection = m_pathfinding.next(nextWay->getEnd().getIntersection(), *m_destination);
			// on rcupre la route qui nous permet d'aller  la prochaine tape
			Way* intWay = nextWay->getEnd().getIntersection().getWayTo(*nextIntersection);
			// si cette route n'est pas lie  la notre, on regarde les routes avant celle-ci est on trouve la notre
			if (&intWay->getBegin() != current)
			{
				for (unsigned j = 0; j < intWay->getBegin().getWaysIn().size(); j++)
				{
					if (&intWay->getBegin().getWaysIn()[j]->getBegin() == current)
					{
						nextWay = intWay->getBegin().getWaysIn()[j];
						break;
					}
				}
			}
			else // on a la bonne route
			{
				nextWay = intWay;
			}
		}
		// on passe  la destination suivante
		current = &nextWay->getEnd();
	}

	return nextWay;
}

float Agent::getFreeDistance() const
{
	const Agent *ahead = m_currentWay->agentAhead(*this);
	if (ahead)
	{
		return m_distanceFromCurrentJunction - ahead->getDistanceFromCurrentJunction() - ahead->getLength();
	}
	float dist = m_distanceFromCurrentJunction;
	unsigned i = 1;
	while (dist < DISTANCE_CHECK_FOR_AGENT)
	{
		Way *nextWay = getNextWay(i);
		if (!nextWay)
		{
			return DISTANCE_CHECK_FOR_AGENT;
		}
		ahead = nextWay->agentBehind();
		if (ahead)
		{
			return dist + nextWay->getLength() - ahead->getDistanceFromCurrentJunction() - ahead->getLength();
		}
		// on n'a trouv aucun agent, on continue sur la suite du path
		dist += nextWay->getLength();
		i++;
	}
	
	// on n'a trouv aucun agent sur la distance de vrification
	return DISTANCE_CHECK_FOR_AGENT;
}

float Agent::getSafetyDistance() const
{
	return m_safetyDistance;
}

const Intersection* Agent::getNextIntersection() const
{
	const Intersection* next = m_pathfinding.next(m_currentDestination->getIntersection(), *m_destination);
	if (next == m_destination)
	{
		next = NULL;
	}
	return next;
}

float Agent::getLength() const
{
	return m_length;
}

float Agent::computeAcceleration() const
{
	// on commence par regarde si il y a quelqu'un devant nous
	const Agent* ahead = getAgentAheadOnPath();
	if (ahead)
	{
		// on calcule l'acclaration maximale possible actuellement, calcule par rapport  l'obstacle devant nous
		float distanceNextAgent = getFreeDistance();
		// on retire la distance de scurit  la distance trouve actuelle
		distanceNextAgent -= m_safetyDistance;

		// si la distance est ngative ou nulle, on ne doit pas avancer
		if (distanceNextAgent <= 0.0f)
		{
			return -m_decelerationPerSecond;
		}

		// si on ne peut pas acclrer sans risquer de percuter l'agent devant
		if (		m_accelerationPerSecond * 0.9f	
				<=	(0.5f * (m_currentSpeed*m_currentSpeed - ahead->getCurrentSpeed()*ahead->getCurrentSpeed())) / distanceNextAgent
			&&	m_currentSpeed - ahead->getCurrentSpeed() > 0.0f )
		{
			// on dclre
			return -m_decelerationPerSecond;
		}
	}

	// on continue en regardant si cette acclration va nous emmener en contact avec des croisements
	// on commence par rduire l'acclration si ncessaire (on ne ralentit plus si on est proche de l'intersection, sauf si on ne
	// peut pas passer (voir plus bas)
	if (	m_distanceFromCurrentJunction >= (m_currentWay->getSpeedLimit() / m_decelerationPerSecond) * 0.1f
		&&	m_accelerationPerSecond 
		<=	(0.5f * (m_currentSpeed * m_currentSpeed - m_maximumSpeedPassingJunction * m_maximumSpeedPassingJunction))
				/ m_distanceFromCurrentJunction)
	{
		// il nous faut ralentir
		return -m_decelerationPerSecond;
	}
	// on vrifie que le prochain croisement soit passable
	if (	m_distanceFromCurrentJunction <= (m_currentWay->getSpeedLimit() / m_decelerationPerSecond) * 0.1f
		&&	!m_currentDestination->passable(*this))
	{
		return -m_decelerationPerSecond;
	}

	return m_accelerationPerSecond;
}