#include "AgentSet.h"

#include "Agent.h"
#include "RoadNetwork.h"
#include "Way.h"
#include "Graph.h"
#include "Junction.h"
#include "utils.h"
#include "Renderer.h"
#include "MediaManager.h"
#include "Intersection.h"
#include "Pathfinding.h"
#include "SoundEngine.h"
#include "Control.h"
#include "MemoryLog.h"

#define CAR_SOUND_NAME "car.wav"
#define NB_AGENTS_PER_BATCH	125

DeclarationElement elemsAgentSet[] =
{
	{ 0, D3DDECLUSAGE_POSITION, D3DDECLTYPE_FLOAT3 },
	{ 0, D3DDECLUSAGE_NORMAL, D3DDECLTYPE_FLOAT3 },
	{ 0, D3DDECLUSAGE_BLENDINDICES, D3DDECLTYPE_SHORT4 }
};

DeclarationElement elemsParticles[] =
{
	{ 0, D3DDECLUSAGE_POSITION, D3DDECLTYPE_FLOAT3 },
	{ 0, D3DDECLUSAGE_TEXCOORD, D3DDECLTYPE_FLOAT2 }
};

AgentSet::AgentSet(SoundEngine *soundEngine, RoadNetwork &roadNetwork, MediaManager &media, Control *control):
m_graph(NULL),
m_disposed(false),
m_media(&media),
m_roadNetwork(&roadNetwork),
m_pathfinding(NULL),
m_soundEngine(soundEngine),
m_showParticles(false),
m_averageVelocity(0.0f),
m_vbParticles(NULL),
m_ibParticles(NULL)
{
	m_shader = m_media->getObject<Shader>("Agent.fx");
	m_declaration = m_media->getRenderer().createDeclaration(elemsAgentSet);

	m_shaderParticles = m_media->getObject<Shader>("Particle.fx");
	m_textureParticles = m_media->getObject<Texture>("Particle.png");
	m_declParticles = m_media->getRenderer().createDeclaration(elemsParticles);

	std::vector<D3DXVECTOR3> vertices;
	std::vector<D3DXVECTOR3> normals;
	std::vector<unsigned short> indices;

	createGeometry(vertices, indices);
	computeNormals(vertices, normals, indices);

	// on cr le vertex buffer et l'index buffer simple que l'on utilisera pour l'affichage
	m_vb = m_media->getRenderer().createVertexBuffer<Vertex>(NB_AGENTS_PER_BATCH * vertices.size());
	Vertex *v = m_vb->lock();	for(unsigned i = 0; i < NB_AGENTS_PER_BATCH; ++i)
	{
		for(unsigned j = 0; j < vertices.size(); ++j)
		{
			v[j].position = vertices[j];
			v[j].normal = normals[j];
			v[j].index[0] = static_cast<char>(i);
		}

		v += vertices.size();
	}

	m_vb->unlock();
	m_ib = m_media->getRenderer().createIndexBuffer<unsigned short>(NB_AGENTS_PER_BATCH * indices.size());

	unsigned short *ind = m_ib->lock();
	unsigned short start = 0;

	for(unsigned i = 0; i < NB_AGENTS_PER_BATCH; ++i)
	{
		for(unsigned j = 0; j < indices.size(); ++j)
		{
			ind[j] = start + indices[j];
		}
		
		ind += indices.size();
		start += static_cast<unsigned short>(vertices.size());
	}

	m_ib->unlock();

	control->addParameter("nbagents", "# Cars", 1.0f, 5000.0f, 1000.0f, 10.0f);
	control->addParameter("undefinedpathprop", "Prop. Undefined Path", 0.0f, 1.0f, 0.2f);
	control->addParameter("proprestocom", "Prop. To Commercial Zone", 0.0f, 1.0f, 0.9f);
	control->addParameter("highwayratio", "Highway Ratio", 0.25f, 4.0f, 1.5f);
	control->addParameter("timecoeff", "Time Coeff.", 1.0f, 20.0f, 1.0f);

	control->attachValue("nbagents", m_nbAgents);
	control->attachValue("undefinedpathprop", m_undefinedPathProp);
	control->attachValue("proprestocom", m_resToComProp);
	control->attachValue("highwayratio", m_highwayRatio);
	control->attachValue("timecoeff", m_timeCoeff);
}

AgentSet::~AgentSet(void)
{
	for (unsigned i=0; i<m_agents.size(); i++)
	{
		delete m_agents[i];
	}
	delete m_vb;
	delete m_ib;
	delete m_declaration;
	if (m_graph != NULL)
	{
		delete m_graph;
	}
	delete m_pathfinding;
	delete m_ibParticles;
	delete m_vbParticles;
	delete m_declParticles;
}

void AgentSet::init(int nbAgents)
{
	m_agents.reserve(nbAgents);

	m_vbParticles = m_media->getRenderer().createVertexBuffer<Particle>(nbAgents * 4);
	m_ibParticles = m_media->getRenderer().createIndexBuffer<unsigned short>(nbAgents * 6);

	//Particle *particles = m_vbParticles->lock();
	unsigned short *ind = m_ibParticles->lock();

	for(unsigned short i = 0; i < static_cast<unsigned short>(nbAgents); ++i)
	{
		ind[0] = 4 * i + 0;
		ind[1] = 4 * i + 1;
		ind[2] = 4 * i + 2;
		ind[3] = 4 * i + 2;
		ind[4] = 4 * i + 1;
		ind[5] = 4 * i + 3;

		ind += 6;
	}

	//m_vbParticles->unlock();
	m_ibParticles->unlock();
}

void AgentSet::disposeOnRoads()
{
	if(m_disposed)
	{
		for (unsigned i=0; i < m_agents.size(); i++)
		{
			delete m_agents[i];
		}
		m_agents.clear();
		delete m_vbParticles;
		delete m_ibParticles;
	}
	
	init(static_cast<int>(m_nbAgents));

	if (m_graph == NULL)
	{
		m_graph = new Graph(*m_roadNetwork, *m_media);
		m_pathfinding = new Pathfinding(*m_graph);

		for(unsigned i = 0; i < m_graph->getWays().size(); ++i)
		{
			Way *way = m_graph->getWays()[i];
			if(way->isZoneType(RZT_COMMERCIAL))
			{
				m_commercialWays.push_back(way);
			}
			if(way->isZoneType(RZT_RESIDENTIAL))
			{
				m_residentialWays.push_back(way);
			}
			if(&way->getBegin().getIntersection() != &way->getEnd().getIntersection())
			{
				m_allWays.push_back(way);
			}
		}
	}
	m_pathfinding->FloydWarshall(m_highwayRatio);

	for (unsigned i=0; i < static_cast<unsigned>(m_nbAgents); i++)
	{
		addAgent();
	}

	m_disposed = true;
}

void AgentSet::setShowParticles(bool visible)
{
	m_showParticles = visible;
}

void AgentSet::render(Renderer &renderer) const
{
	D3DXVECTOR4 positions[NB_AGENTS_PER_BATCH];
	D3DXVECTOR4 colors[NB_AGENTS_PER_BATCH];

	renderer.setShader(m_shader);
	renderer.setDeclaration(m_declaration);
	renderer.setVertexBuffer<Vertex>(0, m_vb, 0);

	for (unsigned i = 0; i < m_agents.size(); i += NB_AGENTS_PER_BATCH)
	{
		unsigned count = min(m_agents.size() - i, NB_AGENTS_PER_BATCH);

		for(unsigned j = 0; j < count; ++j)
		{
			const D3DXVECTOR3 &dir = m_agents[i + j]->getCurrentWay()->getVector();
			const D3DXVECTOR3 &p = m_agents[i + j]->getPosition();
			const D3DXVECTOR3 &color = m_agents[i + j]->getColor();
			float coeff = (dir.z >= 0.0f) ? 1.0f : -1.0f;

			positions[j] = D3DXVECTOR4(p.x, p.y, p.z, dir.x);
			colors[j] = D3DXVECTOR4(coeff * color.x, coeff * color.y, coeff * color.z, dir.y);
		}
		m_shader->setValue("Positions", positions, count);
		m_shader->setValue("Colors", colors, count);
		renderer.drawIndexedPrimitives<unsigned short>(D3DPT_TRIANGLELIST, m_ib, 0, 26 * count);
	}

	if(m_graph)	
		m_graph->render(renderer);

	if(m_showParticles)
	{
		Particle *particles = m_vbParticles->lock();
		for(unsigned i = 0; i < m_agents.size(); ++i)
		{
			particles[0].position = m_agents[i]->getPosition();
			particles[1].position = m_agents[i]->getPosition();
			particles[2].position = m_agents[i]->getPosition();
			particles[3].position = m_agents[i]->getPosition();
			
			particles[0].texcoord = D3DXVECTOR2(0.0f, 0.0f);
			particles[1].texcoord = D3DXVECTOR2(1.0f, 0.0f);
			particles[2].texcoord = D3DXVECTOR2(0.0f, 1.0f);
			particles[3].texcoord = D3DXVECTOR2(1.0f, 1.0f);
			particles += 4;
		}
		m_vbParticles->unlock();

		renderer.setShader(m_shaderParticles);
		renderer.setDeclaration(m_declParticles);
		renderer.setTexture("ColorTexture", m_textureParticles);
		renderer.setVertexBuffer<Particle>(0, m_vbParticles, 0);
		renderer.drawIndexedPrimitives<unsigned short>(D3DPT_TRIANGLELIST, m_ibParticles, 0, m_agents.size() * 2);
	}
}

void AgentSet::run(float timeElapsed)
{
	// si les agents n'ont pas encore t placs, on ne fait rien
	if (!m_disposed)
	{
		return;
	}
	timeElapsed *= m_timeCoeff;
	// on met  jour les intersection
	for (unsigned i=0; i<m_graph->getIntersections().size(); i++)
	{
		m_graph->getIntersections()[i]->update(timeElapsed);
	}
	// chaque agent agit
	for (unsigned i=0; i<m_agents.size(); i++)
	{
		m_agents[i]->act(timeElapsed);
	}
	computeAverageVelocity();
}

void AgentSet::reset()
{
	for (unsigned i=0; i<m_agents.size(); i++)
	{
		delete m_agents[i];
	}
	m_agents.clear();
	delete m_graph;
	m_graph = NULL;
	m_disposed = false;
	m_allWays.clear();
	m_residentialWays.clear();
	m_commercialWays.clear();
}

void AgentSet::addAgent()
{
	Intersection *dest;
	Way	*start;
	choosePath(start, dest);

#ifndef NO_SOUND
	m_agents.push_back( new Agent(m_soundEngine->createSpeaker(m_media->getObject<Sound>(CAR_SOUND_NAME)), *dest, *start, *this, *m_pathfinding) );
#else
	m_agents.push_back( new Agent(NULL, *dest, *start, *this, *m_pathfinding) );
#endif
}

const Graph& AgentSet::getGraph() const
{
	return *m_graph;
}

void AgentSet::replaceAgent(Agent* agent)
{
	unsigned index = 0;

	for(unsigned i = 0; i < m_agents.size(); ++i)
	{
		if(m_agents[i] == agent)
		{
			index = i;
			break;
		}
	}

	// on supprime l'agent en question
	delete agent;
	// puis on le recr
	Intersection *dest;
	Way *start;

	choosePath(start, dest);

#ifndef NO_SOUND
	m_agents[index] = new Agent(m_soundEngine->createSpeaker(m_media->getObject<Sound>(CAR_SOUND_NAME)), *dest, *start, *this, *m_pathfinding);
#else
	m_agents[index] = new Agent(NULL, *dest, *start, *this, *m_pathfinding);
#endif
}

void AgentSet::choosePath(Way *&start, Intersection *&dest) const
{
	ERoadZoneType begin = RZT_UNDEFINED;
	ERoadZoneType end = RZT_UNDEFINED;

	if(randomBool(1.0f - m_undefinedPathProp))
	{
		if(randomBool(m_resToComProp))
		{
			begin = RZT_RESIDENTIAL;
			end = RZT_COMMERCIAL; 
		}
		else
		{
			begin = RZT_COMMERCIAL;
			end = RZT_RESIDENTIAL;
		}
	}

	dest = randomDestination(end);
	start = randomWay(dest, begin);
}

Intersection* AgentSet::randomDestination(ERoadZoneType type) const
{
	const std::vector<Way *> &ways = getWays(type);

	int ind = randomInt(0, ways.size());
	return &ways[ind]->getEnd().getIntersection();
}

Way* AgentSet::randomWay(const Intersection *dest, ERoadZoneType type) const
{
	const std::vector<Way *> &ways = getWays(type);
	Way* way;

	do
	{
		int ind = randomInt(0, ways.size());
		way = ways[ind];
	}
	// on trouve une route qui n'est pas sur une intersection et qui est connexe  notre destination
	while (	m_pathfinding->next(way->getEnd().getIntersection(), *dest) == NULL);

	// si on nous demande de faire demi-tour, on se met sur l'autre voie
	if ( m_pathfinding->next( way->getEnd().getIntersection(), *dest ) == &way->getBegin().getIntersection() )
	{
		way = way->getEnd().getIntersection().getWayTo(way->getBegin().getIntersection());
	}

	return way;
}

const std::vector<Way *> & AgentSet::getWays(ERoadZoneType type) const
{
	switch(type)
	{
	case RZT_UNDEFINED:
	default:
		return m_allWays;
	case RZT_COMMERCIAL:
		return m_commercialWays;
	case RZT_RESIDENTIAL:
		return m_residentialWays;
	}
}

void AgentSet::createGeometry(std::vector<D3DXVECTOR3> &vertices, std::vector<unsigned short> &indices) const
{
	vertices.resize(16);

	vertices[0] = D3DXVECTOR3(-0.26f, 0.00f, 0.05f);
	vertices[1] = D3DXVECTOR3(-0.26f, 0.05f, 0.05f);
	vertices[2] = D3DXVECTOR3(-0.22f, 0.05f, 0.05f);
	vertices[3] = D3DXVECTOR3(-0.18f, 0.08f, 0.04f);
	vertices[4] = D3DXVECTOR3(-0.08f, 0.08f, 0.04f);
	vertices[5] = D3DXVECTOR3(-0.05f, 0.05f, 0.05f);
	vertices[6] = D3DXVECTOR3( 0.00f, 0.05f, 0.05f);
	vertices[7] = D3DXVECTOR3( 0.00f, 0.00f, 0.05f);

	vertices[ 8] = D3DXVECTOR3(-0.26f, 0.00f, -0.05f);
	vertices[ 9] = D3DXVECTOR3(-0.26f, 0.05f, -0.05f);
	vertices[10] = D3DXVECTOR3(-0.22f, 0.05f, -0.05f);
	vertices[11] = D3DXVECTOR3(-0.18f, 0.08f, -0.04f);
	vertices[12] = D3DXVECTOR3(-0.08f, 0.08f, -0.04f);
	vertices[13] = D3DXVECTOR3(-0.05f, 0.05f, -0.05f);
	vertices[14] = D3DXVECTOR3( 0.00f, 0.05f, -0.05f);
	vertices[15] = D3DXVECTOR3( 0.00f, 0.00f, -0.05f);

	indices.resize(26 * 3);

	indices[0 * 3 + 0] = 0;
	indices[0 * 3 + 1] = 2;
	indices[0 * 3 + 2] = 1;

	indices[1 * 3 + 0] = 0;
	indices[1 * 3 + 1] = 5;
	indices[1 * 3 + 2] = 2;

	indices[2 * 3 + 0] = 0;
	indices[2 * 3 + 1] = 7;
	indices[2 * 3 + 2] = 5;

	indices[3 * 3 + 0] = 2;
	indices[3 * 3 + 1] = 4;
	indices[3 * 3 + 2] = 3;

	indices[4 * 3 + 0] = 2;
	indices[4 * 3 + 1] = 5;
	indices[4 * 3 + 2] = 4;

	indices[5 * 3 + 0] = 5;
	indices[5 * 3 + 1] = 7;
	indices[5 * 3 + 2] = 6;


	indices[6 * 3 + 0] = 8 + 0;
	indices[6 * 3 + 1] = 8 + 1;
	indices[6 * 3 + 2] = 8 + 2;

	indices[7 * 3 + 0] = 8 + 0;
	indices[7 * 3 + 1] = 8 + 2;
	indices[7 * 3 + 2] = 8 + 5;

	indices[8 * 3 + 0] = 8 + 0;
	indices[8 * 3 + 1] = 8 + 5;
	indices[8 * 3 + 2] = 8 + 7;

	indices[9 * 3 + 0] = 8 + 2;
	indices[9 * 3 + 1] = 8 + 3;
	indices[9 * 3 + 2] = 8 + 4;

	indices[10 * 3 + 0] = 8 + 2;
	indices[10 * 3 + 1] = 8 + 4;
	indices[10 * 3 + 2] = 8 + 5;

	indices[11 * 3 + 0] = 8 + 5;
	indices[11 * 3 + 1] = 8 + 6;
	indices[11 * 3 + 2] = 8 + 7;

	for(unsigned short j = 0; j < 7; ++j)
	{
		indices[(12 + j * 2 + 0) * 3 + 0] = j + 0;
		indices[(12 + j * 2 + 0) * 3 + 1] = j + 8 + 1;
		indices[(12 + j * 2 + 0) * 3 + 2] = j + 8 + 0;

		indices[(12 + j * 2 + 1) * 3 + 0] = j + 0;
		indices[(12 + j * 2 + 1) * 3 + 1] = j + 1;
		indices[(12 + j * 2 + 1) * 3 + 2] = j + 8 + 1;
	}
}

void AgentSet::computeNormals(const std::vector<D3DXVECTOR3> &vertices, std::vector<D3DXVECTOR3> &normals, const std::vector<unsigned short> &indices) const
{
	normals.resize(vertices.size(), D3DXVECTOR3(0.0f, 0.0f, 0.0f));

	for(unsigned i = 0; i < indices.size(); i += 3)
	{
		D3DXVECTOR3 u = vertices[indices[i + 1]] - vertices[indices[i]];
		D3DXVECTOR3 v = vertices[indices[i + 2]] - vertices[indices[i]];
		D3DXVECTOR3 w;
		D3DXVec3Cross(&w, &u, &v);

		if(D3DXVec3LengthSq(&w) > 0.00000001f)
		{
			D3DXVec3Normalize(&w, &w);

			for(unsigned j = 0; j < 3; ++j)
			{
				normals[indices[i + j]] += w;
			}
		}
	}

	for(unsigned i = 0; i < normals.size(); ++i)
	{
		D3DXVec3Normalize(&normals[i], &normals[i]);
	}
}

void AgentSet::computeAverageVelocity()
{
	float sumVelocities = 0.0f;
	for(unsigned i = 0; i < m_agents.size(); i++)
	{
		sumVelocities += m_agents[i]->getCurrentSpeed();
	}

	m_averageVelocity = sumVelocities * 1 / m_nbAgents; 
}

float AgentSet::getAverageVelocity()const
{
	return m_averageVelocity;
}