#include "BlockSet.h"

#include "Block.h"
#include "MediaManager.h"
#include "Renderer.h"
#include "Environment.h"
#include "utils.h"
#include "MemoryLog.h"

DeclarationElement elemsBlockSet[] =
{
	{ 0, D3DDECLUSAGE_POSITION, D3DDECLTYPE_FLOAT3 },
	{ 0, D3DDECLUSAGE_COLOR, D3DDECLTYPE_FLOAT3 }
};

BlockSet::Patch::Patch():
vb(NULL),
ib(NULL),
nbPoints(0)
{
}

BlockSet::BlockSet(MediaManager &media):
m_media(&media)
{
	m_shader = media.getObject<Shader>("Road.fx");
	m_declaration = media.getRenderer().createDeclaration(elemsBlockSet);
}

BlockSet::~BlockSet(void)
{
	for(int i = 0; i < NB_PATCHES_BLOCKSET*NB_PATCHES_BLOCKSET; ++i)
	{
		delete m_patches[i].vb;
		delete m_patches[i].ib;
	}
	// on supprime les blocks
	for (std::vector<Block*>::iterator it = m_blocks.begin(); it != m_blocks.end(); ++it)
	{
		delete *it;
	}

	delete m_declaration;
}

void BlockSet::addBlock(Block &b)
{
	const D3DXVECTOR3 &p = b.getVertices()[0];
	int i = static_cast<int>(p.x / (1 << GRID_PRECISION) * NB_PATCHES_BLOCKSET);
	int j = static_cast<int>(p.z / (1 << GRID_PRECISION) * NB_PATCHES_BLOCKSET);

	m_patches[i + j * NB_PATCHES_BLOCKSET].newBlocks.push_back(&b);
	m_blocks.push_back(&b);
}

void BlockSet::commitChange()
{
	recomputeBuffer();
}

void BlockSet::recomputeBuffer()
{
	for(int i = 0; i < NB_PATCHES_BLOCKSET*NB_PATCHES_BLOCKSET; ++i)
	{
		std::vector<Block *> &newBlocks = m_patches[i].newBlocks;
		std::vector<Block *> &blocks = m_patches[i].blocks;

		if (blocks.empty() && newBlocks.empty())
		{
			delete m_patches[i].vb;
			delete m_patches[i].ib;
			m_patches[i].vb = NULL;
			m_patches[i].ib = NULL;
		}
		else if(!newBlocks.empty())
		{
			unsigned short nbPoints = 0;
			for (std::vector<Block*>::iterator it = newBlocks.begin(); it != newBlocks.end(); ++it)
			{
				nbPoints += static_cast<unsigned short>((*it)->getVertices().size());
			}

			if(!m_patches[i].vb || m_patches[i].vb->getSize() < static_cast<unsigned>(m_patches[i].nbPoints + nbPoints))
			{
				unsigned size = (static_cast<unsigned>(m_patches[i].nbPoints + nbPoints) * 3) / 2;

				Assert(size < 65536);
				Assert(static_cast<unsigned>(m_patches[i].nbPoints) + static_cast<unsigned>(nbPoints) < 65536/2);

				delete m_patches[i].vb;
				delete m_patches[i].ib;
				m_patches[i].vb = m_media->getRenderer().createVertexBuffer<Vertex>(size);
				m_patches[i].ib = m_media->getRenderer().createIndexBuffer<unsigned short>(2 * size);

				Vertex *vertices = m_patches[i].vb->lock(0, m_patches[i].nbPoints);
				unsigned short *indices = m_patches[i].ib->lock(0, 2 * m_patches[i].nbPoints);

				unsigned short indVertices = 0;
				unsigned short indIndices = 0;
				for (std::vector<Block*>::iterator it = blocks.begin(); it != blocks.end(); ++it)
				{
					unsigned short firstIndice = indVertices;
					// on ajoute la liste de vecteurs
					for (	unsigned j = 0; j < (*it)->getVertices().size(); ++j)
					{
						const D3DXVECTOR3 &vertex = (*it)->getVertices()[j];
						vertices[indVertices].Position = vertex;
						vertices[indVertices].Color = D3DXVECTOR3(0.5f, 0.3f, 0.8f);
						// et on ajoute son indice au polygone que l'on est en train de former
						indices[indIndices++] = indVertices++;
						indices[indIndices++] = (j == (*it)->getVertices().size() - 1) ? firstIndice : indVertices;
					}
				}

				m_patches[i].vb->unlock();
				m_patches[i].ib->unlock();
			}

			Vertex *vertices = m_patches[i].vb->lock(m_patches[i].nbPoints, nbPoints);
			unsigned short *indices = m_patches[i].ib->lock(m_patches[i].nbPoints * 2, nbPoints * 2);
			
			unsigned short indVertices = 0;
			unsigned short indIndices = 0;
			for (std::vector<Block*>::iterator it = newBlocks.begin(); it != newBlocks.end(); ++it)
			{
				unsigned short firstIndice = indVertices;
				// on ajoute la liste de vecteurs
				for (unsigned j = 0; j < (*it)->getVertices().size(); ++j)
				{
					const D3DXVECTOR3 &vertex = (**it).getVertices()[j];
					vertices[indVertices].Position = vertex;
					vertices[indVertices].Color = D3DXVECTOR3(0.5f, 0.3f, 0.8f);
					// et on ajoute son indice au polygone que l'on est en train de former
					indices[indIndices++] = m_patches[i].nbPoints + indVertices++;
					indices[indIndices++] = m_patches[i].nbPoints + ((j == (**it).getVertices().size() - 1) ? firstIndice : indVertices);
				}

				blocks.push_back(*it);
			}

			m_patches[i].vb->unlock();
			m_patches[i].ib->unlock();

			m_patches[i].nbPoints += nbPoints;
			newBlocks.clear();
		}
	}
}

void BlockSet::render(Renderer &renderer) const
{
	renderer.setShader(m_shader);
	m_shader->setValue("World", renderer.getMatrix());
	renderer.setDeclaration(m_declaration);
	for(int i = 0; i < NB_PATCHES_BLOCKSET*NB_PATCHES_BLOCKSET; ++i)
	{
		if(m_patches[i].vb)
		{
			renderer.setVertexBuffer<Vertex>(0, m_patches[i].vb, 0);
			renderer.drawIndexedPrimitives<unsigned short>(D3DPT_LINELIST, m_patches[i].ib, 0, m_patches[i].nbPoints);
		}
	}
}

void BlockSet::reset()
{	
	for(int i = 0; i < NB_PATCHES_BLOCKSET*NB_PATCHES_BLOCKSET; ++i)
	{
		m_patches[i].blocks.clear();
		m_patches[i].newBlocks.clear();
		m_patches[i].nbPoints = 0;
	}

	for(std::vector<Block *>::iterator it = m_blocks.begin(); it != m_blocks.end(); ++it)
	{
		delete *it;
	}
	m_blocks.clear();

	recomputeBuffer();
}

std::vector<Block *>& BlockSet::getBlocks()
{
	return m_blocks;
}

const std::vector<Block *> & BlockSet::getBlocks() const
{
	return m_blocks;
}