#include "BuildingSet.h"
#include "Block.h"
#include "Building.h"
#include "MediaManager.h"
#include "Renderer.h"
#include "Environment.h"
#include "utils.h"
#include "MemoryLog.h"

DeclarationElement elemsBuildingSet[] =
{
	{ 0, D3DDECLUSAGE_POSITION, D3DDECLTYPE_FLOAT3 },
	{ 0, D3DDECLUSAGE_NORMAL, D3DDECLTYPE_FLOAT3 }
};

BuildingSet::Patch::Patch():
vb(NULL),
ib(NULL),
nbVertices(0),
nbIndices(0)
{
}

BuildingSet::BuildingSet(MediaManager &media):
m_media(&media)
{
	m_shader = media.getObject<Shader>("Building.fx");
	m_declaration = media.getRenderer().createDeclaration(elemsBuildingSet);
}

BuildingSet::~BuildingSet(void)
{
	for(int i = 0; i < NB_ZONE_TYPE * NB_PATCHES_BUILDINGSET*NB_PATCHES_BUILDINGSET; ++i)
	{
		delete m_patches[i].vb;
		delete m_patches[i].ib;
	}
	// on supprime les blocks
	for (std::vector<Building*>::iterator it = m_buildings.begin(); it != m_buildings.end(); ++it)
	{
		delete *it;
	}

	delete m_declaration;
}

void BuildingSet::addBuilding(Building &b, ZoneType type)
{
	const D3DXVECTOR3 &p = b.getVertices()[0];
	int i = static_cast<int>(p.x / (1 << GRID_PRECISION) * NB_PATCHES_BUILDINGSET);
	int j = static_cast<int>(p.z / (1 << GRID_PRECISION) * NB_PATCHES_BUILDINGSET);

	m_patches[i + j * NB_PATCHES_BUILDINGSET + type * NB_PATCHES_BUILDINGSET * NB_PATCHES_BUILDINGSET].newBuildings.push_back(&b);
	m_buildings.push_back(&b);
}

void BuildingSet::recomputeBuffer()
{
	for(int i = 0; i < NB_ZONE_TYPE * NB_PATCHES_BUILDINGSET * NB_PATCHES_BUILDINGSET; ++i)
	{
		std::vector<Building*> &newBuildings = m_patches[i].newBuildings;
		std::vector<Building *> &buildings = m_patches[i].buildings;

		if (buildings.empty() && newBuildings.empty())
		{
			delete m_patches[i].vb;
			delete m_patches[i].ib;
			m_patches[i].vb = NULL;
			m_patches[i].ib = NULL;
		}
		else if(!newBuildings.empty())
		{
			unsigned short nbVertices = 0;
			unsigned short nbIndices = 0;
			for (std::vector<Building*>::iterator it = newBuildings.begin(); it != newBuildings.end(); ++it)
			{
				nbVertices += static_cast<unsigned short>((*it)->getVertices().size());
				nbIndices += static_cast<unsigned short>((*it)->getIndices().size());
			}

			if(!m_patches[i].vb || m_patches[i].vb->getSize() < static_cast<unsigned>(m_patches[i].nbVertices + nbVertices))
			{
				unsigned size = (static_cast<unsigned>(m_patches[i].nbVertices + nbVertices) * 3) / 2;

				Assert(size < 65536);

				delete m_patches[i].vb;

				m_patches[i].vb = m_media->getRenderer().createVertexBuffer<Vertex>(size);

				Vertex *vertices = m_patches[i].vb->lock(0, m_patches[i].nbVertices);

				for (std::vector<Building*>::iterator it = buildings.begin(); it != buildings.end(); ++it)
				{
					for(unsigned j = 0; j < (*it)->getVertices().size(); ++j)
					{
						vertices->Position = (*it)->getVertices()[j];
						vertices->Normal = (*it)->getNormals()[j];
						++vertices;
					}
				}

				m_patches[i].vb->unlock();
			}

			if(!m_patches[i].ib || m_patches[i].ib->getSize() < static_cast<unsigned>(m_patches[i].nbIndices + nbIndices))
			{
				unsigned size = (static_cast<unsigned>(m_patches[i].nbIndices + nbIndices) * 3) / 2;

				Assert(size < 65536);

				delete m_patches[i].ib;

				m_patches[i].ib = m_media->getRenderer().createIndexBuffer<unsigned short>(size);

				unsigned short *indices = m_patches[i].ib->lock(0, m_patches[i].nbIndices);

				unsigned short startIndices = 0;
				for (std::vector<Building*>::iterator it = buildings.begin(); it != buildings.end(); ++it)
				{
					for(std::vector<unsigned>::const_iterator it2 = (*it)->getIndices().begin(); it2 != (*it)->getIndices().end(); ++it2)
					{
						*indices = startIndices + static_cast<unsigned short>(*it2);
						++indices;
					}

					startIndices += static_cast<unsigned short>((*it)->getVertices().size());
				}

				m_patches[i].ib->unlock();
			}

			Vertex *vertices = m_patches[i].vb->lock(m_patches[i].nbVertices, nbVertices);
			unsigned short *indices = m_patches[i].ib->lock(m_patches[i].nbIndices, nbIndices);
			
			unsigned short startIndices = m_patches[i].nbVertices;
			for (std::vector<Building*>::iterator it = newBuildings.begin(); it != newBuildings.end(); ++it)
			{
				for(unsigned j = 0; j < (*it)->getVertices().size(); ++j)
				{
					vertices->Position = (*it)->getVertices()[j];
					vertices->Normal = (*it)->getNormals()[j];
					++vertices;
				}

				for(std::vector<unsigned>::const_iterator it2 = (*it)->getIndices().begin(); it2 != (*it)->getIndices().end(); ++it2)
				{
					*indices = startIndices + static_cast<unsigned short>(*it2);
					++indices;
				}

				startIndices += static_cast<unsigned short>((*it)->getVertices().size());
				buildings.push_back(*it);
			}

			m_patches[i].vb->unlock();
			m_patches[i].ib->unlock();

			m_patches[i].nbVertices += nbVertices;
			m_patches[i].nbIndices += nbIndices;
			newBuildings.clear();
		}
	}
}

void BuildingSet::commitChange()
{
	recomputeBuffer();
}

static const D3DXVECTOR3 zoneColor[NB_ZONE_TYPE] =
{
	D3DXVECTOR3(0.5f, 0.3f, 0.8f),
	D3DXVECTOR3(1.0f, 0.4f, 0.0f),
	D3DXVECTOR3(0.3f, 0.5f, 0.8f),
	D3DXVECTOR3(0.0f, 0.5f, 0.0f),
	D3DXVECTOR3(0.5f, 0.5f, 0.5f)
};
	
void BuildingSet::render(Renderer &renderer) const
{
	renderer.setShader(m_shader);
	m_shader->setValue("World", renderer.getMatrix());
	renderer.setDeclaration(m_declaration);
	for(unsigned k = 0; k < NB_ZONE_TYPE; ++k)
	{
		m_shader->setValue("Color", &zoneColor[k]);
		for(unsigned i = 0; i < NB_PATCHES_BUILDINGSET*NB_PATCHES_BUILDINGSET; ++i)
		{
			unsigned index = k * NB_PATCHES_BUILDINGSET * NB_PATCHES_BUILDINGSET + i;
			if(m_patches[index].vb)
			{
				renderer.setVertexBuffer<Vertex>(0, m_patches[index].vb, 0);
				renderer.drawIndexedPrimitives<unsigned short>(D3DPT_TRIANGLELIST, m_patches[index].ib, 0, m_patches[index].nbIndices / 3);
			}
		}
	}
}

void BuildingSet::reset()
{
	for(int i = 0; i < NB_ZONE_TYPE * NB_PATCHES_BUILDINGSET*NB_PATCHES_BUILDINGSET; ++i)
	{
		m_patches[i].buildings.clear();
		m_patches[i].newBuildings.clear();
		m_patches[i].nbIndices = 0;
		m_patches[i].nbVertices = 0;
	}

	for(std::vector<Building *>::iterator it = m_buildings.begin(); it != m_buildings.end(); ++it)
	{
		delete *it;
	}
	m_buildings.clear();

	recomputeBuffer();
}

void BuildingSet::setHeight(float height)
{
	m_height = height;
}