#include "City.h"

#include "MediaManager.h"
#include "Renderer.h"
#include "GenEnvironment.h"
#include "GenCenters.h"
#include "GenDensity.h"
#include "GenRoad.h"
#include "GenBlocks.h"
#include "GenAllotment.h"
#include "CommandFrame.h"
#include "GenBuilding.h"
#include "MemoryLog.h"
#include "Graph.h"

#define PROCEED_TIME 0.01f
#define PROCEED_TIME_STEP 0.0001f
#define NB_AGENTS 2000

#define RENDER_WATER				0x0001
#define RENDER_LANDSCAPE			0x0002
#define RENDER_TEXTUREDLANDSCAPE	0x0004
#define RENDER_ROADS				0x0008
#define RENDER_ROADS_GEO			0x0010
#define RENDER_BLOCKS				0x0020
#define RENDER_ALLOTMENTS			0x0040
#define RENDER_BUILDINGS			0x0080
#define RENDER_AGENTS				0x0100

City::City(SoundEngine *soundEngine, MediaManager &media, CommandFrame *frame):
m_mediaManager(&media),
m_environment(media, GRID_PRECISION),
m_density(media, m_environment),
m_actStep(0),
m_run(false),
m_stopAtNextStep(false),
m_roadNetwork(media),
m_roadGeometry(media),
m_blocks(media),
m_allotments(media),
m_agentSet(soundEngine, m_roadNetwork, media, frame->getSimulationControl()),
m_interface(frame),
m_buildings(media)
{
	m_interface->onStep += new BlindDelegate<City, void>(this, &City::runStep);
	m_interface->onRun += new BlindDelegate<City, void>(this, &City::run);
	m_interface->onRunAll += new BlindDelegate<City, void>(this, &City::runAll);
	m_interface->onClear += new BlindDelegate<City, void>(this, &City::reset);

	m_interface->onSetupAgents += new BlindDelegate<City, void>(this, &City::setupAgents);
	m_interface->onStartSimulation += new BlindDelegate<City, void>(this, &City::startSimulation);
	m_interface->onStopSimulation += new BlindDelegate<City, void>(this, &City::stopSimulation);
	m_interface->onShowBottlenecks += new BlindDelegate<City, bool>(this, &City::showBottlenecks);

	setRenderState();
	setGenerators();
	createControls();
}

City::~City(void)
{
	while(!m_pipeline.empty())
	{
		delete m_pipeline.back();
		m_pipeline.pop_back();
	}
}

void City::render(Renderer &renderer) const
{
	unsigned renderState = m_renderStates[m_actStep];

	if(renderState & RENDER_WATER)
		m_environment.renderWater(renderer);
	if(renderState & RENDER_LANDSCAPE)
		m_environment.render(renderer);	
	if(renderState & RENDER_TEXTUREDLANDSCAPE)
		m_density.render(renderer);
	if(renderState & RENDER_ROADS)
		m_roadNetwork.render(renderer);
	if(renderState & RENDER_ROADS_GEO)
		m_roadGeometry.render(renderer);
	if(renderState & RENDER_BLOCKS)
		m_blocks.render(renderer);
	if(renderState & RENDER_ALLOTMENTS)
		m_allotments.render(renderer);
	if(renderState & RENDER_BUILDINGS)
		m_buildings.render(renderer);
	if(renderState & RENDER_AGENTS)
		m_agentSet.render(renderer);

}

void City::update (float timeElapsed)
{
	if(m_run && m_actStep < m_pipeline.size())
	{
		m_pipeline[m_actStep]->proceed(PROCEED_TIME);

		if(m_pipeline[m_actStep]->isFinished())
		{
			++m_actStep;

			if(m_actStep >= m_pipeline.size() || m_stopAtNextStep)
			{
				m_run = false;
			}
			m_interface->setCurrentStage(m_actStep);

			if(m_actStep >= m_pipeline.size())
			{
				m_interface->setMode(CFM_SIMULATION);
			}
		}
	}
	else if (m_run)
	{
		// lancement du moteur de simulation
		m_agentSet.run(timeElapsed);
		m_interface->setAverageSpeed(m_agentSet.getAverageVelocity());
	}
}

void City::runAll()
{
	m_run = true;
	m_stopAtNextStep = false;
}

void City::run()
{
	m_run = true;
	m_stopAtNextStep = true;
}

void City::runStep()
{
	if(m_actStep < m_pipeline.size())
	{
		m_pipeline[m_actStep]->proceed(PROCEED_TIME_STEP);

		if(m_pipeline[m_actStep]->isFinished())
		{
			++m_actStep;
			m_interface->setCurrentStage(m_actStep);

			if(m_actStep >= m_pipeline.size())
			{
				m_interface->setMode(CFM_SIMULATION);
			}
		}
	}
	m_run = false;
}

void City::reset()
{
	m_actStep = 0;
	m_interface->setMode(CFM_GENERATION);
	m_interface->setCurrentStage(0);

	m_run = false;

	deleteGenerators();

	m_environment.reset();
	m_density.reset();
	m_roadNetwork.reset();
	m_roadGeometry.reset();
	m_blocks.reset();
	m_allotments.reset();
	m_buildings.reset();

	m_agentSet.reset();

	setGenerators();
	attachControls();
}

void City::setupAgents()
{
	m_run = false;
	m_interface->setAverageSpeed(-1.0f);
	m_agentSet.disposeOnRoads();
}

void City::startSimulation()
{
	Assert(m_actStep >= m_pipeline.size());
	m_run = true;
}

void City::stopSimulation()
{
	Assert(m_actStep >= m_pipeline.size());
	m_run = false;
}

void City::showBottlenecks(bool show)
{
	m_agentSet.setShowParticles(show);
}

void City::setRenderState()
{
	m_renderStates.push_back(RENDER_LANDSCAPE);
	m_renderStates.push_back(RENDER_WATER | RENDER_TEXTUREDLANDSCAPE);
	m_renderStates.push_back(RENDER_WATER | RENDER_TEXTUREDLANDSCAPE);
	m_renderStates.push_back(RENDER_WATER | RENDER_TEXTUREDLANDSCAPE | RENDER_ROADS);
	m_renderStates.push_back(RENDER_WATER | RENDER_LANDSCAPE | RENDER_ROADS | RENDER_ROADS_GEO | RENDER_BLOCKS);
	m_renderStates.push_back(RENDER_WATER | RENDER_LANDSCAPE | RENDER_BLOCKS | RENDER_ROADS_GEO | RENDER_ALLOTMENTS);
	m_renderStates.push_back(RENDER_WATER | RENDER_LANDSCAPE | RENDER_ROADS_GEO | RENDER_ALLOTMENTS | RENDER_BUILDINGS);
	m_renderStates.push_back(RENDER_WATER | RENDER_LANDSCAPE | RENDER_BUILDINGS | RENDER_ROADS_GEO | RENDER_AGENTS);
}

void City::setGenerators()
{
	m_pipeline.push_back(new GenEnvironment(m_environment));
	m_pipeline.push_back(new GenCenters(m_density));
	m_pipeline.push_back(new GenDensity(m_density));
	m_pipeline.push_back(new GenRoad(m_roadNetwork, m_environment, m_density));
	m_pipeline.push_back(new GenBlocks(m_roadNetwork, m_roadGeometry, m_blocks, m_environment, m_density));
	m_pipeline.push_back(new GenAllotment(m_blocks, m_allotments));
	m_pipeline.push_back(new GenBuilding(m_allotments, m_buildings, m_density));
}

void City::deleteGenerators()
{
	for(unsigned i = 0; i < m_pipeline.size(); ++i)
	{
		delete m_pipeline[i];
	}
	m_pipeline.clear();
}

void City::createControls()
{
	for(unsigned i = 0; i < m_pipeline.size(); ++i)
	{
		Control *control = m_pipeline[i]->newControl(m_interface->getGUI());
		m_interface->addControl(control);
		m_pipeline[i]->attachControl(control);
	}
	m_interface->setCurrentStage(0);
}

void City::attachControls()
{
	for(unsigned i = 0; i < m_pipeline.size(); ++i)
	{
		m_pipeline[i]->attachControl(m_interface->getControl(i));
	}
}
