#include "Control.h"
#include "GUILabel.h"
#include "GUITrackbar.h"
#include "Debug.h"
#include <sstream>
#include <iomanip>
#include <cmath>
#include "MemoryLog.h"

Control::Control(GUI *gui, const std::string &name):
GUIWindow(gui, name),
m_nbParameters(0)
{
}

void Control::addParameter(const std::string &name, const std::string &label)
{
	addParameter(name, label, 0.0f, 100.0f);
}

void Control::addParameter(const std::string &name, const std::string &label, float minValue, float maxValue)
{
	addParameter(name, label, minValue, maxValue, minValue);
}

void Control::addParameter(const std::string &name, const std::string &label, float minValue, float maxValue, float defaultValue)
{
	addParameter(name, label, minValue, maxValue, defaultValue, 0.01f);
}

void Control::addParameter(const std::string &name, const std::string &label, float minValue, float maxValue, float defaultValue, float step)
{
	float height = m_nbParameters * 40.0f;
	GUILabel *graphicLabel = new GUILabel(getGUI(), getName() + "." + name + "_label", this, label);
	graphicLabel->getPosition().setPosition(D3DXVECTOR2(4.0f, height), GA_TOPLEFT, GA_TOPLEFT);

	GUILabel *graphicValue = new GUILabel(getGUI(), getName() + "." + name + "_value", this, "");
	graphicValue->getPosition().setPosition(D3DXVECTOR2(-4.0f, height), GA_TOPRIGHT, GA_TOPRIGHT);

	GUITrackbar *trackbar = new GUITrackbar(getGUI(), getName() + "." + name, this);
	trackbar->getSize().setRelativeSize(D3DXVECTOR2(-8.0f, 16.0f), D3DXVECTOR2(1.0f, 0.0f));
	trackbar->getPosition().setPosition(D3DXVECTOR2(0.0f, height + 20.0f), GA_TOP, GA_TOP);
	trackbar->setStart(minValue);
	trackbar->setEnd(maxValue);
	trackbar->setStep(step);
	trackbar->onValueChanged += new Delegate<Control, float>(this, &Control::valueChanged);

	Parameter &param = m_parameters[name];
	param.label = graphicValue;
	param.trackbar = trackbar;
	param.value = NULL;

	++m_nbParameters;

	trackbar->setValue(defaultValue, true);
}

void Control::attachValue(const std::string &name, float &value)
{
	Assert(m_parameters.isExist(name));

	Parameter &param = m_parameters[name];
	param.value = &value;
	value = param.trackbar->getValue();
}

void Control::saveParameter(const std::string &name, const Parameter &value, std::ofstream &file)
{
	file << name << "\t" << value.label->getText() << "\n";
}

void Control::saveParameters(std::ofstream &file) const
{
	m_parameters.forEach<std::ofstream &>(saveParameter, file);
}

void Control::loadParameters(std::ifstream &file)
{
	std::string line;

	while(getline(file, line))
	{
		if(line == "")
		{
			return;
		}

		std::istringstream buffer(line);
		std::string name;
		float value;

		buffer >> name;

		if(m_parameters.isExist(name))
		{
			buffer >> value;
			m_parameters[name].trackbar->setValue(value, true);
		}
	}
}

void Control::valueChanged(GUIWindow *sender, float value)
{
	const std::string &name = sender->getName().substr(getName().size() + 1);

	Parameter &param = m_parameters[name];
	int prec = -static_cast<int>(floor(std::log(param.trackbar->getStep()) / std::log(10.0f)));
	prec = max(0, prec);

	std::ostringstream stream;
	stream << std::setprecision(prec) << std::setiosflags(std::ios_base::fixed) << value;
	param.label->setText(stream.str());
	
	if(param.value)
	{
		*param.value = value;
	}
}