#include "Crossing.h"
#include "Road.h"
#include "utils.h"
#include <cmath>
#include <algorithm>
#include "MemoryLog.h"

Crossing::Crossing(const D3DXVECTOR3 &pos):
m_position(pos)
{
	// initialisation de la liste
	m_roads.empty();
}

Crossing::~Crossing(void)
{
}

/**************************************************************************************************
					Getters et Setters
**************************************************************************************************/
const D3DXVECTOR3& Crossing::getPosition() const
{
	return m_position;
}

D3DXVECTOR3 & Crossing::getPosition()
{
	return m_position;
}

D3DXVECTOR2 Crossing::get2DPosition() const
{
	return D3DXVECTOR2(m_position.x, m_position.z);
}

const std::vector<const Road*>& Crossing::getRoads() const
{
	return m_roads;
}

void Crossing::addRoad(const Road &r)
{
	m_roads.push_back(&r);
}

void Crossing::removeRoad(const Road &r)
{
	for(std::vector<const Road *>::const_iterator it = m_roads.begin(); it != m_roads.end(); ++it)
	{
		if(*it == &r)
		{
			m_roads.erase(it);
			return;
		}
	}
}

void Crossing::setPosition(const D3DXVECTOR3& position)
{
	m_position = position;
}

bool Crossing::isNeighbor(const Crossing *c) const
{
	for(std::vector<const Road *>::const_iterator it = m_roads.begin(); it != m_roads.end(); ++it)
	{
		if(&(*it)->getEndCrossing(*this) == c)
		{
			return true;
		}
	}

	return false;
}

bool Crossing::isNeighbor(const Road *r) const
{
	for(std::vector<const Road *>::const_iterator it = m_roads.begin(); it != m_roads.end(); ++it)
	{
		if(*it == r)
		{
			return true;
		}
	}

	return false;
}

bool Crossing::hasColinearRoad(const Road *r, float cosTreshold) const
{
	D3DXVECTOR2 dir = r->get2DVector();
	D3DXVec2Normalize(&dir, &dir);

	if(cosTreshold < 0.0f)
	{
		dir = -dir;
		cosTreshold = -cosTreshold;
	}

	for(std::vector<const Road *>::const_iterator it = m_roads.begin(); it != m_roads.end(); ++it)
	{
		if(*it == r)
		{
			continue;
		}

		D3DXVECTOR2 dir2 = (*it)->get2DVector();
		D3DXVec2Normalize(&dir2, &dir2);

		if(this != &(*it)->getStartCrossing())
		{
			dir2 = -dir2;
		}

		if(D3DXVec2Dot(&dir, &dir2) > cosTreshold)
		{
			return true;
		}
	}

	return false;
}

unsigned Crossing::getNbStreet() const
{
	unsigned nb = 0;
	for(unsigned i = 0; i < m_roads.size(); ++i)
	{
		if(m_roads[i]->getType() == RT_STREET)
			++nb;
	}
	return nb;
}


class RoadAngleComp
{
public:
	RoadAngleComp(const Crossing *start);
	float getRoadAngle(const Road *r) const;
	bool operator()(const Road *r1, const Road *r2) const;

private:
	const Crossing *m_start;
};

RoadAngleComp::RoadAngleComp(const Crossing *start):
m_start(start)
{
}

float RoadAngleComp::getRoadAngle(const Road *r) const
{
	D3DXVECTOR2 vec = r->get2DVector();
	D3DXVec2Normalize(&vec, &vec);
	if(&r->getStartCrossing() != m_start)
	{
		vec = -vec;
	}
	clamp(vec.x, -1.0f, 1.0f);

	float angle = std::acos(vec.x);
	return (vec.y < 0.0f) ? -angle : angle;
}

bool RoadAngleComp::operator ()(const Road *r1, const Road *r2) const
{
	return getRoadAngle(r1) < getRoadAngle(r2);
}


void Crossing::sortRoad()
{
	std::sort<std::vector<const Road *>::iterator, const RoadAngleComp &>(
		m_roads.begin(), m_roads.end(), RoadAngleComp(this));
}

int Crossing::getRoadIndex(const Road *r) const
{
	for(unsigned i = 0; i < m_roads.size(); ++i)
	{
		if(m_roads[i] == r)
		{
			return i;
		}
	}
	return -1;
}

void Crossing::setJunctionPositions(std::vector<D3DXVECTOR3> &positions)
{
	m_junctionPositions.swap(positions);
}

D3DXVECTOR3 Crossing::getJunctionPosition(unsigned index) const
{
	return m_junctionPositions[index];
}