#include "Density.h"

#include "Environment.h"
#include "Texture.h"
#include "Renderer.h"
#include "MediaManager.h"
#include "Shader.h"
#include "Declaration.h"
#include "MemoryLog.h"

#define NB_POINTS_CENTERS 32

DeclarationElement densityElems[] =
{
	{ 0, D3DDECLUSAGE_POSITION, D3DDECLTYPE_FLOAT3 }
};

Density::Density(MediaManager &media, Environment &env):
m_env(&env),
m_gridSize(1 << env.getGridPrecision()),
m_densityMap(m_gridSize, m_gridSize),
m_textureMap(m_gridSize, m_gridSize),
m_media(&media),
m_vbCenters(NULL),
m_ibCenters(NULL)
{
	// initialisation de la texture, du shader, de la textureMap et de la dclaration
	m_texture = media.getRenderer().createDynamicTexture(m_textureMap);
	m_shader = media.getObject<Shader>("TexturedLandscape.fx");
	m_shaderCenter = media.getObject<Shader>("Line.fx");
	m_declaration = media.getRenderer().createDeclaration(densityElems);
	
	reset();
}

Density::~Density(void)
{
	delete m_texture;
	delete m_vbCenters;
	delete m_ibCenters;
	delete m_declaration;
}

int Density::getGridSize() const
{
	return m_gridSize;
}

const Image<float> & Density::getDensityMap() const
{
	return m_densityMap;
}

Image<unsigned short> & Density::getTextureMap()
{
	return m_textureMap;
}

Image<float> & Density::getDensityMap()
{
	return m_densityMap;
}

const std::vector<Center> & Density::getCenters() const
{
	return m_centers;
}

std::vector<Center>& Density::getCenters()
{
	return m_centers;
}

const Environment& Density::getEnvironment() const
{
	return *m_env;
}

Environment& Density::getEnvironment()
{
	return *m_env;
}

void Density::addCenter(const Center &toAdd)
{
	m_centers.push_back(toAdd);
}

void Density::addCenter(const D3DXVECTOR2 &pos, const float innerRadius, const float outerRadius, ECenterType type)
{
	Center toAdd;
	toAdd.pos = pos;
	toAdd.innerRadius = innerRadius;
	toAdd.outerRadius = outerRadius;
	toAdd.type = type;
	
	addCenter(toAdd);
}

void Density::update()
{
	// on synchronise la texture
	m_texture->update();
}

void Density::updateCenters()
{
	if (m_vbCenters == NULL)
	{
		return;
	}

	D3DXVECTOR3 *vertices = m_vbCenters->lock();
	unsigned short *indices = m_ibCenters->lock();

	const float angle = D3DX_PI * 2.0f / static_cast<float>(NB_POINTS_CENTERS);

	// pour chaque centre
	unsigned short nb = 0;
	for (std::vector<Center>::const_iterator it = m_centers.begin(); it != m_centers.end(); ++it)
	{
		// on calcule donc les positions des points du vertex buffer
		for (int i=0; i<NB_POINTS_CENTERS; i++)
		{
			float x = it->pos.x + it->innerRadius * std::cos(static_cast<float>(i) * angle);
			float z = it->pos.y + it->innerRadius * std::sin(static_cast<float>(i) * angle);
			vertices[i + NB_POINTS_CENTERS * nb].x = x;
			vertices[i + NB_POINTS_CENTERS * nb].z = z;
			(x <= 0.0f) ? x = 0.0f : x;
			(x > getGridSize()) ? x = static_cast<float>(m_env->getGridSize()) - 1.0f : x;
			(z <= 0.0f) ? z = 0.0f : z;
			(z > getGridSize()) ? z = static_cast<float>(m_env->getGridSize()) - 1.0f : z;
			vertices[i + NB_POINTS_CENTERS * nb].y = m_env->getHeightMap().interpolate(x, z) + 0.1f;
		}

		// puis les indices de l'index buffer
		for (unsigned short i=0; i<NB_POINTS_CENTERS+4; i++)
		{
			indices[i + (NB_POINTS_CENTERS + 4) * nb] = (i % NB_POINTS_CENTERS) + nb * NB_POINTS_CENTERS;
			if (i >= NB_POINTS_CENTERS)
			{
				i++;
				indices[i + (NB_POINTS_CENTERS + 4) * nb] = NB_POINTS_CENTERS / 2 + nb * NB_POINTS_CENTERS;
				i++;
				indices[i + (NB_POINTS_CENTERS + 4) * nb] = 3 * NB_POINTS_CENTERS / 4 + nb * NB_POINTS_CENTERS;
				i++;
				indices[i + (NB_POINTS_CENTERS + 4) * nb] = NB_POINTS_CENTERS / 4 + nb * NB_POINTS_CENTERS;
			}
		}
		nb++;
	}

	m_vbCenters->unlock();
	m_ibCenters->unlock();
}

void Density::render(Renderer &renderer) const
{
	renderer.setShader(m_shader);
	m_shader->setValue("World", renderer.getMatrix());

	m_shader->setValue("Size", getGridSize() - 1.0f);
	renderer.setTexture("density", m_texture);
	D3DXVECTOR3 color(1.0f, 0.1f, 0.2f);	
	m_shader->setValue("Color", &color);
	m_env->renderLandscape(renderer);

	if (m_vbCenters)
	{
		renderer.setShader(m_shaderCenter);
		m_shaderCenter->setValue("World", renderer.getMatrix());
		renderer.setDeclaration(m_declaration);
		renderer.setVertexBuffer<D3DXVECTOR3>(0, m_vbCenters, 0);
		for (unsigned i = 0; i < m_centers.size(); i++)
		{
			renderer.drawIndexedPrimitives<unsigned short>(D3DPT_LINESTRIP, m_ibCenters, i * (NB_POINTS_CENTERS + 4), NB_POINTS_CENTERS );
			renderer.drawIndexedPrimitives<unsigned short>(D3DPT_LINELIST, m_ibCenters, i * (NB_POINTS_CENTERS + 4) + NB_POINTS_CENTERS, 1 );
			renderer.drawIndexedPrimitives<unsigned short>(D3DPT_LINELIST, m_ibCenters, i * (NB_POINTS_CENTERS + 4) + NB_POINTS_CENTERS + 2, 1 );
		}
	}
}

void Density::reset()
{
	// on efface l'affichage des centres
	displayCenters(false);

	for (int i=0; i<m_gridSize; i++)
	{
		for (int j=0; j<m_gridSize; j++)
		{
			m_densityMap(i, j) = 0.0f;
		}
	}

	m_centers.clear();

	for (int i=0; i<m_gridSize; i++)
	{
		for (int j=0; j<m_gridSize; j++)
		{
			m_textureMap(i, j) = 0;
		}
	}

	update();
}

void Density::displayCenters(bool disp)
{
	
	if (disp)	// on demande d'afficher les centres
	{
		// si on n'a pas de centres, on ne fait rien
		if (m_centers.size() == 0)
		{
			return;
		}
		m_vbCenters = m_media->getRenderer().createVertexBuffer<D3DXVECTOR3>(m_centers.size() * NB_POINTS_CENTERS);
		m_ibCenters = m_media->getRenderer().createIndexBuffer<unsigned short>(m_centers.size() * (NB_POINTS_CENTERS+4));

		updateCenters();

	}
	else		// on doit effacer les centres
	{
		if (m_vbCenters != NULL)
		{
			delete m_vbCenters;
			m_vbCenters = NULL;
		}
		if (m_ibCenters != NULL)
		{
			delete m_ibCenters;
			m_ibCenters = NULL;
		}
	}		
}

const float Density::getLocalDensity(const D3DXVECTOR2& pos) const
{
	float localDensity = 0.0f;
	D3DXVECTOR2 posCenter;

	// on regarde la distance du point actuel avec tous les centres
	for (	std::vector<Center>::const_iterator it = m_centers.begin();
			it != m_centers.end();
			++it )
	{
		posCenter = pos - (*it).pos;
		float distSq = D3DXVec2LengthSq(&posCenter);
		if (distSq < it->innerRadius * it->innerRadius)
		{
			// on a une densit max sur le point, pas la peine de continuer
			return 1.0f;
		}
		else if (distSq < it->outerRadius * it->outerRadius)
		{
			float tmp = 1 -	(distSq - it->innerRadius*it->innerRadius ) 
							/	(it->outerRadius*it->outerRadius - it->innerRadius*it->innerRadius);
			// si on dpasse une densit de 1, on sort
			if (tmp >= 1.0f)
			{
				return 1.0f;
			} 
			else if (tmp > localDensity)
			{
				localDensity = tmp;
			}
		}
	}

	return localDensity;
}