
#include "Environment.h"

#include "MediaManager.h"
#include "Renderer.h"
#include "Shader.h"
#include "MemoryLog.h"

static DeclarationElement elemsLandscape[] =
{
	{0, D3DDECLUSAGE_POSITION, D3DDECLTYPE_FLOAT2},
	{1, D3DDECLUSAGE_NORMAL, D3DDECLTYPE_FLOAT4}
};

static DeclarationElement elemsWater[] =
{
	{0, D3DDECLUSAGE_POSITION, D3DDECLTYPE_FLOAT2}
};

/*****************************************************************************************************************************
											CONSTRUCTEUR / DESTRUCTEUR
*****************************************************************************************************************************/

Environment::Environment(MediaManager &media, int gridPrecision, float waterLevel):
m_gridPrecision(gridPrecision),
m_gridSize((1 << gridPrecision) +1),
m_waterLevel(waterLevel),
m_heightsMap(m_gridSize, m_gridSize),
m_minHeight(0.0f),
m_maxHeight(0.0f)
{
	m_vbLandscape = media.getRenderer().createVertexBuffer<D3DXVECTOR2>(m_gridSize * m_gridSize);
	m_ibLandscape = media.getRenderer().createIndexBuffer<INDEX_TYPE>((m_gridSize - 1) * (m_gridSize * 2 + 1) - 1);
	m_normals = media.getRenderer().createVertexBuffer<D3DXVECTOR4>(m_gridSize * m_gridSize);
	m_declLandscape = media.getRenderer().createDeclaration(elemsLandscape);
	m_shaderLandscape = media.getObject<Shader>("Landscape.fx");

	m_shaderWater = media.getObject<Shader>("Water.fx");
	m_vbWater = media.getRenderer().createVertexBuffer<D3DXVECTOR2>(4);
	m_declWater = media.getRenderer().createDeclaration(elemsWater);

	// initialisation de la heightMap  0
	for (int i=0; i<m_gridSize; i++)
	{
		for (int j=0; j<m_gridSize; j++)
		{
			m_heightsMap(i, j) = 0.0f;
		}
	}

	// initialisation des variables graphiques
	synchroHeights();
	D3DXVECTOR2 *vertices = m_vbLandscape->lock();
	for(int j = 0; j < m_gridSize; ++j)
	{
		for(int i = 0; i < m_gridSize; ++i)
		{
			vertices[j * m_gridSize + i] = D3DXVECTOR2(static_cast<float>(i), static_cast<float>(j));
		}
	}
	m_vbLandscape->unlock();

	INDEX_TYPE *indices = m_ibLandscape->lock();

#define INDEX_OF(i, j) (static_cast<INDEX_TYPE>((j) * m_gridSize + (i)))

	for(int j = 0; j < m_gridSize - 1; j += 2)
	{
		if(j > 0)
		{
			*indices++ = INDEX_OF(0, j);
		}

		for(int i = 0; i < m_gridSize; ++i)
		{
			*indices++ = INDEX_OF(i, j);
			*indices++ = INDEX_OF(i, j + 1);
		}

		*indices++ = INDEX_OF(m_gridSize - 1, j + 1);

		for(int i = m_gridSize - 1; i >= 0; --i)
		{
			*indices++ = INDEX_OF(i, j + 1);
			*indices++ = INDEX_OF(i, j + 2);
		}
	}

	m_ibLandscape->unlock();

	vertices = m_vbWater->lock();

	vertices[0] = D3DXVECTOR2(0.0f, 0.0f);
	vertices[1] = D3DXVECTOR2(0.0f, static_cast<float>(m_gridSize));
	vertices[2] = D3DXVECTOR2(static_cast<float>(m_gridSize), static_cast<float>(m_gridSize));
	vertices[3] = D3DXVECTOR2(static_cast<float>(m_gridSize), 0.0f);

	m_vbWater->unlock();
}

Environment::~Environment(void)
{
	delete m_vbLandscape;
	delete m_normals;
	delete m_ibLandscape;
	delete m_declLandscape;
	delete m_vbWater;
	delete m_declWater;
}

/*****************************************************************************************************************************
											FONCTIONS PUBLIQUES
*****************************************************************************************************************************/

void Environment::render(Renderer &renderer) const
{
	renderer.setShader(m_shaderLandscape);
	m_shaderLandscape->setValue("World", renderer.getMatrix());
	D3DXVECTOR3 color(1.0f, 0.8f, 0.2f);	
	m_shaderLandscape->setValue("Color", &color);
	renderLandscape(renderer);
}

void Environment::renderLandscape(Renderer &renderer) const
{
	renderer.setDeclaration(m_declLandscape);
	renderer.setVertexBuffer<D3DXVECTOR2>(0, m_vbLandscape, 0);
	renderer.setVertexBuffer<D3DXVECTOR4>(1, m_normals, 0);
	renderer.drawIndexedPrimitives<INDEX_TYPE>(D3DPT_TRIANGLESTRIP, m_ibLandscape, 0, (m_gridSize - 1) * (m_gridSize * 2 + 1) - 3);
	renderer.clearZBuffer();
}

void Environment::renderWater(Renderer &renderer) const
{
	renderer.setDeclaration(m_declWater);
	renderer.setVertexBuffer<D3DXVECTOR2>(0, m_vbWater, 0);
	renderer.setShader(m_shaderWater);
	m_shaderWater->setValue("Height", m_waterLevel);
	m_shaderWater->setValue("World", renderer.getMatrix());
	renderer.drawPrimitives(D3DPT_TRIANGLEFAN, 0, 2);
}

void Environment::synchroHeights()
{
	// on locke le vertex buffer
	D3DXVECTOR4 *normals = m_normals->lock();

	// et on y enregistre les hauteurs de chaque point
	for(int j = 0; j < m_gridSize; ++j)
	{
		for(int i = 0; i < m_gridSize; ++i)
		{
			float dx = computeDX((i == 0) ? 0 : (i - 1), (i == m_gridSize - 1) ? m_gridSize - 1 : i, j);
			float dy = computeDY((j == 0) ? 0 : (j - 1), (j == m_gridSize - 1) ? m_gridSize - 1 : j, i);
			D3DXVECTOR4 normal(dx, 1.0f, dy, 0.0f);
			D3DXVec4Normalize(&normal, &normal);
			normal.w = m_heightsMap(i, j);
			normals[j * m_gridSize + i] = normal;
		}
	}

	m_normals->unlock();
}

void Environment::reset()
{
	for(int j = 0; j < m_gridSize; ++j)
	{
		for(int i = 0; i < m_gridSize; ++i)
		{
			m_heightsMap(i, j) = 0.0f;
		}
	}

	synchroHeights();
}

float Environment::computeDX(int x1, int x3, int y) const
{
	return (m_heightsMap(x1, y) - m_heightsMap(x3, y)) / 2.0f;
}

float Environment::computeDY(int y1, int y3, int x) const
{
	return (m_heightsMap(x, y1) - m_heightsMap(x, y3)) / 2.0f;
}

/*****************************************************************************************************************************
											GETTERS / SETTERS
*****************************************************************************************************************************/

int Environment::getGridSize() const
{
	return m_gridSize;
}

int Environment::getGridPrecision() const
{
	return m_gridPrecision;
}

const Image<float> & Environment::getHeightMap() const
{
	return m_heightsMap;
}

Image<float> & Environment::getHeightMap()
{
	return m_heightsMap;
}

float Environment::getWaterLevel() const
{
	return m_waterLevel;
}

float & Environment::getWaterLevel()
{
	return m_waterLevel;
}

void Environment::setWaterLevel(float waterLevel)
{
	m_waterLevel = waterLevel;
}

float Environment::getMinHeight() const
{
	return m_minHeight;
}

float Environment::getMaxHeight() const
{
	return m_maxHeight;
}

void Environment::setMinHeight(float minHeight)
{
	m_minHeight = minHeight;
}

void Environment::setMaxHeight(float maxHeight)
{
	m_maxHeight = maxHeight;
}
