#include "FFT.h"
#include <cmath>
#include "MemoryLog.h"

FFT fastFourierTransform;

FFT::FFT()
{
	int denum = 1;
	for(unsigned i = 0; i < FFT_LOG_POINTS; ++i)
	{
		for(unsigned j = 0; j < FFT_NB_POINTS; ++j)
		{
			float re = std::cos((M_PI * j) / denum);
			float im = -std::sin((M_PI * j) / denum);
			m_exponentials[i * FFT_NB_POINTS + j] = std::complex<float>(re, im);
		}
		denum <<= 1;
	}

	m_reverseOrder[0] = 0;
	for(unsigned i = 1; i < FFT_NB_POINTS; ++i)
	{
		m_reverseOrder[i] = m_reverseOrder[i - 1];

		unsigned mask = FFT_NB_POINTS >> 1;
		while(m_reverseOrder[i] & mask)
		{
			m_reverseOrder[i] &= ~mask;
			mask >>= 1;
		}
		m_reverseOrder[i] |= mask;
	}
}

void FFT::operator()(const float *data, std::complex<float> *outBuffer)
{	
	convertData(data, outBuffer);

	for(unsigned i = 0; i < FFT_LOG_POINTS; ++i)
	{
		unsigned step = 1 << i;
		unsigned inc = 2 << i;
		for(unsigned j = 0; j < step; ++j)
		{
			const std::complex<float> &exp = m_exponentials[i * FFT_NB_POINTS + j];

			for(unsigned k = j; k < FFT_NB_POINTS; k += inc)
			{
				std::complex<float> temp = exp * outBuffer[k + step];
				outBuffer[k + step] = outBuffer[k] - temp;
				outBuffer[k] += temp;
			}
		}
	}
}

void FFT::operator()(std::complex<float> *data, float *outBuffer)
{	
	convertData(data);

	for(unsigned i = 0; i < FFT_LOG_POINTS; ++i)
	{
		unsigned step = 1 << i;
		unsigned inc = 2 << i;
		for(unsigned j = 0; j < step; ++j)
		{
			std::complex<float> exp = m_exponentials[i * FFT_NB_POINTS + j];
			exp.imag(-exp.imag());

			for(unsigned k = j; k < FFT_NB_POINTS; k += inc)
			{
				std::complex<float> temp = exp * data[k + step];
				data[k + step] = data[k] - temp;
				data[k] += temp;
			}
		}
	}
	
	outputData(data, outBuffer);
}

void FFT::convertData(const float *data, std::complex<float> *buffer)
{
	for(unsigned i = 0; i < FFT_NB_POINTS; ++i)
	{
		buffer[m_reverseOrder[i]] = std::complex<float>(data[i]);
	}
}

void FFT::convertData(std::complex<float> *buffer)
{
	for(unsigned i = 0; i < FFT_NB_POINTS; ++i)
	{
		if(i < m_reverseOrder[i])
		{
			std::swap(buffer[i], buffer[m_reverseOrder[i]]);
		}
	}
}

void FFT::outputData(const std::complex<float> *data, float *outBuffer)
{
	const float scale = 1.0f / FFT_NB_POINTS;
	for(unsigned i = 0; i < FFT_NB_POINTS; ++i)
	{
		outBuffer[i] = data[i].real() * scale;
	}
}