#include "GUI.h"
#include "GUIDesktopWindow.h"
#include "IOManager.h"
#include "MediaManager.h"
#include "Renderer.h"
#include "GUITheme.h"
#include "MemoryLog.h"

static const unsigned short indices[] = { 0, 1, 2, 2, 1, 3 };

static DeclarationElement elems[] =
{
	{0, D3DDECLUSAGE_POSITION, D3DDECLTYPE_FLOAT2},
	{0, D3DDECLUSAGE_TEXCOORD, D3DDECLTYPE_FLOAT2}
};

GUI::GUI(IOManager &io, MediaManager &media, Appli &appli):
m_root(NULL),
m_focus(NULL),
m_captureMouse(NULL),
m_media(&media),
m_ib(NULL),
m_decl(NULL)
{
	io.registerListener(this, KEYDOWN);
	io.registerListener(this, KEYPRESSED);
	io.registerListener(this, KEYRELEASED);
	io.registerListener(this, KEYCHAR);
	io.registerListener(this, MOUSEMOVE);
	io.registerListener(this, MOUSEDRAG);
	io.registerListener(this, MOUSECLICK);
	io.registerListener(this, MOUSERELEASE);
	io.registerListener(this, MOUSEWHEEL);

	m_ib = media.getRenderer().createIndexBuffer<unsigned short>(6, indices);
	m_decl = media.getRenderer().createDeclaration(elems);
	m_shader = media.getObject<Shader>("UI.fx");

	m_focus = m_root = new GUIDesktopWindow(this, &appli, 800, 600);
	m_theme = new GUITheme(media);
}

GUI::~GUI()
{
	delete m_root;
	delete m_ib;
	delete m_decl;
	delete m_theme;
}

void GUI::onKeyDown(const IOManager &, unsigned key, float elapsedTime)
{
	if(m_focus)
	{
		m_focus->dispatchKeyDown(key, elapsedTime);
	}
}

void GUI::onKeyPressed(const IOManager &, unsigned key)
{
	if(m_focus)
	{
		m_focus->dispatchKeyPress(key, true, false);
	}
}

void GUI::onKeyReleased(const IOManager &, unsigned key)
{
	if(m_focus)
	{
		m_focus->dispatchKeyPress(key, false, false);
	}
}

void GUI::onKeyChar(const IOManager &, unsigned key)
{
	if(m_focus)
	{
		m_focus->dispatchKeyPress(key, true, true);
	}
}

void GUI::onMouseMove(const IOManager &io, int dx, int dy)
{
	int x = io.getMouseX();
	int y = io.getMouseY();

	int lastX = x - dx;
	int lastY = y - dy;

	if(m_captureMouse)
	{
		m_captureMouse->onMouseMove(
			D3DXVECTOR2(static_cast<float>(lastX), static_cast<float>(lastY)),
			D3DXVECTOR2(static_cast<float>(x), static_cast<float>(y)));
	}
	else
	{
		m_root->dispatchMouseMove(
			D3DXVECTOR2(static_cast<float>(lastX), static_cast<float>(lastY)),
			D3DXVECTOR2(static_cast<float>(x), static_cast<float>(y)));
	}
}

void GUI::onMouseDrag(const IOManager &io, int dx, int dy)
{
	int x = io.getMouseX();
	int y = io.getMouseY();

	int lastX = x - dx;
	int lastY = y - dy;

	if(m_captureMouse)
	{
		m_captureMouse->onMouseDrag(
			D3DXVECTOR2(static_cast<float>(lastX), static_cast<float>(lastY)),
			D3DXVECTOR2(static_cast<float>(x), static_cast<float>(y)),
			io.getMouseState());
	}
	else
	{
		m_root->dispatchMouseDrag(
			D3DXVECTOR2(static_cast<float>(lastX), static_cast<float>(lastY)),
			D3DXVECTOR2(static_cast<float>(x), static_cast<float>(y)),
			io.getMouseState());
	}
}

void GUI::onMouseClick(const IOManager &io, unsigned button)
{
	D3DXVECTOR2 pos(static_cast<float>(io.getMouseX()), static_cast<float>(io.getMouseY()));
	if(m_captureMouse)
	{
		m_captureMouse->onMouseClick(pos, button);
	}
	else
	{
		m_root->dispatchMouseButton(pos, button, true);
	}
}

void GUI::onMouseRelease(const IOManager &io, unsigned button)
{
	D3DXVECTOR2 pos(static_cast<float>(io.getMouseX()), static_cast<float>(io.getMouseY()));
	if(m_captureMouse)
	{
		m_captureMouse->onMouseRelease(pos, button);
	}
	else
	{
		m_root->dispatchMouseButton(pos, button, false);
	}
}

void GUI::onMouseWheel(const IOManager &io, float delta)
{
	D3DXVECTOR2 pos(static_cast<float>(io.getMouseX()), static_cast<float>(io.getMouseY()));
	if(m_captureMouse)
	{
		m_captureMouse->onMouseWheel(pos, delta);
	}
	else
	{
		m_root->dispatchMouseWheel(pos, delta);
	}
}

void GUI::onSize(unsigned width, unsigned height)
{
	m_root->setResolution(width, height);
}

void GUI::setFocus(GUIWindow *focusWindow)
{
	m_focus = focusWindow ? focusWindow : m_root;
}

void GUI::update(float elapsedTime)
{
	m_root->update(elapsedTime);
}

void GUI::render(Renderer &renderer) const
{
	renderer.setDeclaration(m_decl);
	renderer.setShader(m_shader);

	const D3DXVECTOR2 &screenSize = m_root->getSize().getAbsoluteSize();
	D3DXVECTOR2 oneOverScreenSize(1.0f / screenSize.x, 1.0f / screenSize.y);
	m_shader->setValue("OneOverScreenSize", &oneOverScreenSize);

	m_root->render(renderer);
}

void GUI::drawRect(Renderer &renderer, VertexBuffer<GUIVertex>::Type *vb, Texture *tex, float alpha) const
{
	renderer.setTexture("ColorMap", tex);
	m_shader->setValue("Alpha", alpha);
	renderer.setVertexBuffer<GUIVertex>(0, vb, 0);
	renderer.drawIndexedPrimitives<unsigned short>(D3DPT_TRIANGLELIST, m_ib, 0, 2);
}

GUITheme * GUI::getTheme() const
{
	return m_theme;
}

GUIWindow * GUI::getRoot() const
{
	return m_root;
}

void GUI::addWindow(GUIWindow *window)
{
	m_root->addChild(window);
}

void GUI::captureMouse(GUIWindow *window)
{
	m_captureMouse = window;
}

void GUI::releaseMouse(GUIWindow *window)
{
	if(m_captureMouse == window)
	{
		m_captureMouse = NULL;
	}
}

MediaManager & GUI::getMedia() const
{
	return *m_media;
}