#include "GUITrackbar.h"
#include "GUITexture.h"
#include "Debug.h"
#include "utils.h"
#include <cmath>
#include "MemoryLog.h"

GUITrackbar::GUITrackbar(GUI *gui, const std::string &name, GUIWindow *parent):
GUIWindow(gui, name, parent),
m_start(0.0f),
m_end(100.0f),
m_value(0.0f),
m_step(0.01f)
{
	m_textures[0] = new GUITexture(gui->getMedia(), "Trackbar_bar.png", *this);
	m_textures[0]->getSize().setRelativeSize(D3DXVECTOR2(-8.0f, 6.0f), D3DXVECTOR2(1.0f, 0.0f));
	m_textures[0]->getPosition().setPosition(D3DXVECTOR2(0.0f, 0.0f), GA_CENTER, GA_CENTER);

	m_textures[1] = new GUITexture(gui->getMedia(), "Trackbar_tracker.png", *this);
	m_textures[1]->getSize().setAbsoluteSize(D3DXVECTOR2(8.0f, 16.0f));
	m_textures[1]->getPosition().setPosition(D3DXVECTOR2(0.0f, 0.0f), GA_LEFT, GA_CENTER);
}

GUITrackbar::~GUITrackbar()
{
	delete m_textures[0];
	delete m_textures[1];
}

float GUITrackbar::getStart() const
{
	return m_start;
}

float GUITrackbar::getEnd() const
{
	return m_end;
}

float GUITrackbar::getValue() const
{
	return m_value;
}

float GUITrackbar::getStep() const
{
	return m_step;
}

void GUITrackbar::setStart(float start)
{
	m_start = start;
	if(m_value < m_start)
	{
		m_value = m_start;
	}
}

void GUITrackbar::setEnd(float end)
{
	m_end = end;
	if(m_value > m_end)
	{
		m_value = m_end;
	}
}

void GUITrackbar::setValue(float value, bool forceNotify)
{
	float lastValue = m_value;
	if(value < m_start)
	{
		m_value = m_start;
	}
	else if(value > m_end)
	{
		m_value = m_end;
	}
	else
	{
		m_value = std::floor(value / m_step + 0.5f) * m_step;
	}

	if(forceNotify || lastValue != m_value)
	{
		onValueChanged(this, m_value);
	}
}

void GUITrackbar::setStep(float step)
{
	m_step = step;
	setValue(m_value);
}

void GUITrackbar::updateSelf(float)
{
	setTrackerPosition();
}

void GUITrackbar::renderSelf(Renderer &renderer) const
{
	m_textures[0]->render(renderer, *getGUI());
	m_textures[1]->render(renderer, *getGUI());
}

bool GUITrackbar::onMouseClick(const D3DXVECTOR2 &pos, unsigned button)
{
	if(isInTracker(pos) && button == MK_LBUTTON)
	{
		getGUI()->captureMouse(this);
		return true;
	}
	return false;
}

bool GUITrackbar::onMouseRelease(const D3DXVECTOR2 &, unsigned button)
{
	if(button == MK_LBUTTON)
	{
		getGUI()->releaseMouse(this);
		return true;
	}
	return false;
}

bool GUITrackbar::onMouseDrag(const D3DXVECTOR2 &, const D3DXVECTOR2 &pos, unsigned buttons)
{
	if(buttons == MK_LBUTTON)
	{
		float start = m_textures[0]->getPosition().getAbsolutePosition().x;
		float d = (pos.x - start) / m_textures[0]->getSize().getAbsoluteSize().x;

		setValue(m_start + d * (m_end - m_start));
	}

	return true;
}

void GUITrackbar::setTrackerPosition()
{
	Assert(m_start < m_end);

	float d = (m_value - m_start) / (m_end - m_start);
	m_textures[1]->getPosition().setPosition(D3DXVECTOR2(4.0f + m_textures[0]->getSize().getAbsoluteSize().x * d, 0.0f));
}

bool GUITrackbar::isInTracker(const D3DXVECTOR2 &point) const
{
	const D3DXVECTOR2 &min = m_textures[1]->getPosition().getAbsolutePosition();

	return isInRect(min, min + m_textures[1]->getSize().getAbsoluteSize(), point);
}