#include "GUIWindow.h"
#include "GUI.h"
#include "utils.h"
#include "MemoryLog.h"

GUIWindow::GUIWindow(GUI *gui, const std::string &name, GUIWindow *parent):
m_size(parent ? &parent->getClientSize() : NULL),
m_position(parent ? &parent->getClientPosition() : NULL, m_size),
m_gui(gui),
m_parent(parent),
m_name(name),
m_visible(true),
m_closed(false)
{
	if(parent)
	{
		m_parent->addChild(this);
	}
}

GUIWindow::~GUIWindow()
{
	while(!m_children.empty())
	{
		delete *m_children.begin();
	}

	if(m_parent)
	{
		m_parent->removeChild(this);
	}
}

GUIWindow * GUIWindow::getParent() const
{
	return m_parent;
}

const Size & GUIWindow::getSize() const
{
	return m_size;
}

Size & GUIWindow::getSize()
{
	return m_size;
}

const Position & GUIWindow::getPosition() const
{
	return m_position;
}

Position & GUIWindow::getPosition()
{
	return m_position;
}

const std::string & GUIWindow::getName() const
{
	return m_name;
}

GUIWindow * GUIWindow::getChildrenByName(const std::string &name)
{
	if(m_name == name)
	{
		return this;
	}
	else
	{
		for(std::vector<GUIWindow *>::iterator it = m_children.begin(); it != m_children.end(); ++it)
		{
			GUIWindow *win = (*it)->getChildrenByName(name);
			if(win)
			{
				return win;
			}
		}

		return NULL;
	}
}

const GUIWindow * GUIWindow::getChildrenByName(const std::string &name) const
{
	if(m_name == name)
	{
		return this;
	}
	else
	{
		for(std::vector<GUIWindow *>::const_iterator it = m_children.begin(); it != m_children.end(); ++it)
		{
			GUIWindow *win = (*it)->getChildrenByName(name);
			if(win)
			{
				return win;
			}
		}

		return NULL;
	}
}
Position & GUIWindow::getClientPosition()
{
	return getPosition();
}

Size & GUIWindow::getClientSize()
{
	return getSize();
}

const Position & GUIWindow::getClientPosition() const
{
	return getPosition();
}

const Size & GUIWindow::getClientSize() const
{
	return getSize();
}

D3DXVECTOR2 GUIWindow::clientToWindowSize(const D3DXVECTOR2 &size)
{
	return size;
}

void GUIWindow::addChild(GUIWindow *child)
{
	if(child->getParent())
	{
		child->getParent()->removeChild(child);
	}

	child->m_parent = this;
	child->m_gui = m_gui;
	child->m_size.setParentSize(&getClientSize());
	child->m_position.setParentPosition(&getClientPosition());
	m_children.push_back(child);
}

void GUIWindow::removeChild(GUIWindow *child)
{
	child->m_parent = NULL;
	child->m_size.setParentSize(NULL);
	child->m_position.setParentPosition(NULL);
	for(std::vector<GUIWindow *>::iterator it = m_children.begin(); it != m_children.end(); ++it)
	{
		if(*it == child)
		{
			m_children.erase(it);
			return;
		}
	}
}

bool GUIWindow::onMouseClick(const D3DXVECTOR2 &, unsigned)
{
	return false;
}

bool GUIWindow::onMouseWheel(const D3DXVECTOR2 &, float)
{
	return false;
}

bool GUIWindow::onMouseRelease(const D3DXVECTOR2 &, unsigned)
{
	return false;
}

void GUIWindow::onMouseMove(const D3DXVECTOR2 &, const D3DXVECTOR2 &)
{
}

void GUIWindow::onMouseEnter(const D3DXVECTOR2 &, const D3DXVECTOR2 &)
{
}

void GUIWindow::onMouseLeave(const D3DXVECTOR2 &, const D3DXVECTOR2 &)
{
}

bool GUIWindow::onMouseDrag(const D3DXVECTOR2 &, const D3DXVECTOR2 &, unsigned )
{
	return false;
}

bool GUIWindow::onKeyDown(unsigned, float)
{
	return false;
}

bool GUIWindow::onKeyPressed(unsigned)
{
	return false;
}

bool GUIWindow::onKeyReleased(unsigned)
{
	return false;
}

bool GUIWindow::onChar(unsigned)
{
	return false;
}

bool GUIWindow::dispatchMouseButton(const D3DXVECTOR2 &pos, unsigned button, bool down)
{
	if(isVisible())
	{
		for(std::vector<GUIWindow *>::reverse_iterator it = m_children.rbegin(); it != m_children.rend(); ++it)
		{
			if((*it)->dispatchMouseButton(pos, button, down))
			{
				return true;
			}
		}

		if(isInWindow(pos))
		{
			if(down ? onMouseClick(pos, button) : onMouseRelease(pos, button))
			{
				return true;
			}
		}
	}
	return false;
}

bool GUIWindow::dispatchMouseWheel(const D3DXVECTOR2 &pos, float delta)
{
	if(isVisible())
	{
		for(std::vector<GUIWindow *>::reverse_iterator it = m_children.rbegin(); it != m_children.rend(); ++it)
		{
			if((*it)->dispatchMouseWheel(pos, delta))
			{
				return true;
			}
		}

		if(isInWindow(pos))
		{
			if(onMouseWheel(pos, delta))
			{
				return true;
			}
		}
	}
	return false;
}

void GUIWindow::dispatchMouseMove(const D3DXVECTOR2 &lastPos, const D3DXVECTOR2 &pos)
{
	if(isVisible())
	{
		for(std::vector<GUIWindow *>::reverse_iterator it = m_children.rbegin(); it != m_children.rend(); ++it)
		{
			(*it)->dispatchMouseMove(lastPos, pos);
		}

		bool a = isInWindow(lastPos);
		bool b = isInWindow(pos);

		if(a && b)
		{
			onMouseMove(lastPos, pos);
		}
		else if(a)
		{
			onMouseLeave(lastPos, pos);
		}
		else if(b)
		{
			onMouseEnter(lastPos, pos);
		}
	}
}

bool GUIWindow::dispatchMouseDrag(const D3DXVECTOR2 &lastPos, const D3DXVECTOR2 &pos, unsigned buttons)
{
	if(isVisible())
	{
		for(std::vector<GUIWindow *>::reverse_iterator it = m_children.rbegin(); it != m_children.rend(); ++it)
		{
			if((*it)->dispatchMouseDrag(lastPos, pos, buttons))
			{
				return true;
			}
		}

		if(isInWindow(lastPos))
		{
			return onMouseDrag(lastPos, pos, buttons);
		}
	}
	return false;
}

void GUIWindow::dispatchKeyPress(unsigned key, bool down, bool translated)
{
	if(!(translated ? onChar(key) : 
		(down ? onKeyPressed(key) : onKeyReleased(key))))
	{
		if(m_parent)
		{
			m_parent->dispatchKeyPress(key, down, translated);
		}
	}
}

void GUIWindow::dispatchKeyDown(unsigned key, float elapsedTime)
{
	if(!onKeyDown(key, elapsedTime))
	{
		if(m_parent)
		{
			m_parent->dispatchKeyDown(key, elapsedTime);
		}
	}
}

bool GUIWindow::isInWindow(const D3DXVECTOR2 &point) const
{
	const D3DXVECTOR2 &min = m_position.getAbsolutePosition();

	return isInRect(min, min + m_size.getAbsoluteSize(), point);
}

bool GUIWindow::isVisible() const
{
	return m_visible;
}

void GUIWindow::show()
{
	m_visible = true;
}

void GUIWindow::hide()
{
	m_visible = false;
}

void GUIWindow::close()
{
	hide();
	m_closed = true;
}

void GUIWindow::getFocus()
{
	m_gui->setFocus(this);
}

void GUIWindow::update(float elapsedTime)
{
	std::vector<GUIWindow *> closedWindow;

	updateSelf(elapsedTime);

	for(std::vector<GUIWindow *>::iterator it = m_children.begin(); it != m_children.end(); ++it)
	{
		if((*it)->m_closed)
		{
			closedWindow.push_back(*it);
		}
	}

	for(std::vector<GUIWindow *>::iterator it = closedWindow.begin(); it != closedWindow.end(); ++it)
	{
		delete *it;
	}

	for(std::vector<GUIWindow *>::iterator it = m_children.begin(); it != m_children.end(); ++it)
	{
		(*it)->update(elapsedTime);
	}
}

void GUIWindow::render(Renderer &renderer) const
{
	if(isVisible())
	{
		renderSelf(renderer);

		for(std::vector<GUIWindow *>::const_iterator it = m_children.begin(); it != m_children.end(); ++it)
		{
			(*it)->render(renderer);
		}
	}
}

void GUIWindow::updateSelf(float)
{
}

void GUIWindow::renderSelf(Renderer &) const
{
}

GUI * GUIWindow::getGUI() const
{
	return m_gui;
}