#include "GenAllotment.h"
#include "Allotment.h"
#include "Block.h"
#include "utils.h"
#include "ZoneType.h"
#include <cmath>
#include <limits>
#include "Control.h"
#include "MemoryLog.h"

#define MIN_RAND	-0.25f
#define MAX_RAND	0.25f

GenAllotment::GenAllotment(const BlockSet &blocks, BlockSet &allotments):
Generator("Allotments"),
m_blocks(&blocks),
m_allotments(&allotments),
m_actBlock(0)
{
}

GenAllotment::~GenAllotment(void)
{
}

/*
Fonction permettant de trouver le milieu d'un segement
*/
static D3DXVECTOR3 midpoint(Edge edge)
{	
	float random = randomFloat(MIN_RAND, MAX_RAND);
	D3DXVECTOR3 point = edge.v1 + edge.vector * (0.5f + random);
	return point;
}

/*
Fonction permettant de trouver le point d'intersection de la perpendiculaire au ct et de la base passant par le midpoint  
*/
static float intersect(const D3DXVECTOR3 &A, const D3DXVECTOR3 &n, const Edge &e, float &t2)
{
	D3DXVECTOR2 u(n.x, n.z);
	D3DXVECTOR2 v(e.vector.x, e.vector.z);
	D3DXVECTOR2 w = D3DXVECTOR2(A.x, A.z) - D3DXVECTOR2(e.v1.x, e.v1.z);
	float d;
	float t1;
	d = D3DXVec2CCW(&u, &v);

	if(std::fabs(d) < 0.0001f)
	{
		return std::numeric_limits<float>::infinity();
	}

	d = 1.0f / d;
	t1 = D3DXVec2CCW(&v, &w) * d;
	t2 = D3DXVec2CCW(&u, &w) * d;
	
	if(t1 <= 0.0001f || t2 < 0.0f || t2 > 1.0f)
	{
		return std::numeric_limits<float>::infinity();
	}

	return t1;
}

/*
Fonction permettant de trouver le segment le plus parrallle  un autre
*/
static Edge* intersectedEdge(const D3DXVECTOR3 &A, const D3DXVECTOR3 &n, std::vector<Edge> &edges, float &tOut)
{
	float tMin= std::numeric_limits<float>::infinity();
	Edge *pEdge = NULL;
	for(unsigned i = 0; i < edges.size(); i++)
	{	
		float t2 = 0.0f;
		float t = intersect(A, n, edges[i], t2);

		if(t < tMin)
		{
			tMin = t;
			pEdge = &edges[i];
			tOut = t2;
		}
	}

	return pEdge;
}

/*
Fonction permettant de trouver le plus grand ct d'un polygone
*/
static Edge& maxEdge(std::vector<Edge> &edges)
{
		Edge *maxEdge = &edges[0];
		for(unsigned i = 1; i < edges.size(); i++)
		{
			if(edges[i].length > maxEdge->length) maxEdge = &edges[i];
		}
		maxEdge->vector = maxEdge->v2 - maxEdge->v1;
		return *maxEdge;
}

/*
Fonction permettant de calculer l'aire d'un triangle  partir de deux vecteurs de dfinition
*/

static float computeTriangleArea(const D3DXVECTOR2 &u, const D3DXVECTOR2 &v)
{
	return D3DXVec2CCW(&v, &u) * 0.5f;
}

/*
Fonction permettant de calculer l'aire d'un polygone
*/
static float computePolygonArea(Allotment *A)
{   
	std::vector<D3DXVECTOR3> &vertices = A->getVertices();

	float area = 0.0f;
	D3DXVECTOR2 origin(vertices[0].x, vertices[0].z);
	for(unsigned i = 0; i < vertices.size() - 2; ++i)
	{
		D3DXVECTOR2 u = D3DXVECTOR2(vertices[i + 1].x, vertices[i + 1].z) - origin;
		D3DXVECTOR2 v = D3DXVECTOR2(vertices[i + 2].x, vertices[i + 2].z) - origin; 
		area += computeTriangleArea(u, v);
	}
	return area;
}


/*
Fonction permettant de diviser un polygone en deux, renvoyant ces deux derniers
*/
static void subdivide(Allotment *A, Edge edge1, Edge edge2, D3DXVECTOR3 p1, D3DXVECTOR3 p2, Allotment *&out1, Allotment *&out2)
{
	std::vector<bool> markers1, markers2;
	unsigned size = A->getVertices().size();
	const std::vector<Edge> &edges = A->getEdges();
	std::vector<D3DXVECTOR3> vertices1, vertices2;
	
	//Dterminaison du premier subAllotment
	vertices1.push_back(p1);
	vertices1.push_back(p2);
	markers1.push_back(true);
	vertices1.push_back(edge2.v2);
	markers1.push_back(edge2.marker);
	unsigned bf = edge1.index;
	unsigned bi = edge2.index;
	for(unsigned i = (bi + 1) % size ; i != bf && i < size; i = (i + 1) % size)
	{	
		vertices1.push_back(edges[i].v2);
		markers1.push_back(edges[i].marker);
	}
	markers1.push_back(edge1.marker);
	
	//Dterminaison du second subAllotment
	vertices2.push_back(p2);
	vertices2.push_back(p1);
	markers2.push_back(true);
	vertices2.push_back(edge1.v2);
	markers2.push_back(edge1.marker);
	for(unsigned i = (bf + 1) % size; i != bi % size && i < size; i = (i + 1) % size)
	{
		vertices2.push_back(edges[i].v2);
		markers2.push_back(edges[i].marker);
	}
	markers2.push_back(edge2.marker);

	//on obtient ainsi nos deux subAllotments
	out1 = new Allotment(A->getZoneType(), vertices1, markers1);
	out2 = new Allotment(A->getZoneType(), vertices2, markers2);
}

std::list<Allotment*> GenAllotment::generate(Allotment *allotment)
{
	std::list<Allotment*> inlist, outlist;
	float area = computePolygonArea(allotment);
	//for(unsigned i = 0; i < allotment->getEdges().size(); i++) allotment->getEdges()[i].marker = false;
	if(area < m_maxArea) outlist.push_back(allotment);
	else inlist.push_back(allotment);
	if(!inlist.empty()) 
	{
		for(std::list<Allotment*>::iterator it = inlist.begin(); it != inlist.end(); ++it)
		{
			//on cherche le plus grand ct
			Edge biggestEdge = maxEdge((*it)->getEdges());

			//on cherche le point du milieu
			D3DXVECTOR3 v1 = midpoint(biggestEdge);

			//on cherche le segment coupant le allotment selon ce point et la perpendiculaire du ct
			float t = 0.0f;
			Edge *parallelEdge = intersectedEdge(v1, biggestEdge.n, (*it)->getEdges(), t); 
			
			if(!parallelEdge)
			{
				outlist.push_back(*it);
			}
			else
			{
				D3DXVECTOR3 pp = parallelEdge->v1 + t * parallelEdge->vector;
				
				//on subdivise notre allotment en deux
				Allotment *sub1, *sub2;
				subdivide(*it, biggestEdge, *parallelEdge, v1, pp, sub1, sub2);
				
				//on dtermine si on devra subdiviser les subAllotments 
				float area1, area2;
				if(!sub1->isIn())
				{
					area1 = computePolygonArea(sub1);
					if(area1 < m_maxArea) outlist.push_back(sub1);
					else inlist.push_back(sub1);
				}
				else
				{
					outlist.push_back(sub1);
				}
				if(!sub2->isIn())
				{
					area2 = computePolygonArea(sub2);
					if(area2 < m_maxArea) outlist.push_back(sub2);
					else inlist.push_back(sub2);
				}
				else
				{
					outlist.push_back(sub2);
				}
				delete *it;
			}
		}
	}
	
	return outlist;
}

bool GenAllotment::isFinished() const
{
	return m_actBlock >= m_blocks->getBlocks().size();
}

void GenAllotment::commitChange()
{
	m_allotments->commitChange();
}

void GenAllotment::createParameters(Control *control)
{
	control->addParameter("maxarea", "Max Lots Area", 0.1f, 30.0f, 2.0f, 0.1f);
	control->addParameter("minarea", "Min Lots Area", 0.1f, 30.0f, 0.7f, 0.1f);
}

void GenAllotment::attachControl(Control *control)
{
	control->attachValue("maxarea", m_maxArea);
	control->attachValue("minarea", m_minArea);
}

void GenAllotment::proceed()
{
	const Block &block = *m_blocks->getBlocks()[m_actBlock++];

	if(block.getZoneType() == PARK)
	{
		m_allotments->addBlock(*new Block(block));
	}
	else
	{
		const std::list<Allotment*> &listAllotments = generate(new Allotment(block.getZoneType(), block.getVertices()));

		for(std::list<Allotment *>::const_iterator it = listAllotments.begin(); it != listAllotments.end(); ++it)
		{
			if((*it)->isIn() || computePolygonArea(*it) < m_minArea)
			{
				m_allotments->addBlock(*new Block((*it)->getVertices(), ROAD));
			}
			else
			{
				m_allotments->addBlock(*new Block((*it)->getVertices(), block.getZoneType()));
			}
			delete *it;
		}
	}
}