#include "GenBlocks.h"

#include "Block.h"
#include "RoadNetwork.h"
#include "Road.h"
#include "Crossing.h"
#include "BlockSet.h"
#include "Debug.h"
#include "Environment.h"
#include "Density.h"
#include "Image.h"
#include "utils.h"
#include "ZoneType.h"
#include <cmath>
#include <limits>
#include "Control.h"
#include "MemoryLog.h"

#define BLOCK_PARK_MIN_POINTS 24
#define BLOCK_POLYGON_MAX_POINTS 100
#define MIN_ANGLE_BETWEEN_BLOCK_VERTICES 0.9999f

GenBlocks::GenBlocks(RoadNetwork &roadNetwork, Geometry &roadsGeometry, BlockSet &blockSet, const Environment &env, const Density &density):
Generator("Blocks"),
m_roads(&roadNetwork.getRoads()),
m_crossings(&roadNetwork.getCrossings()),
m_roadsGeometry(&roadsGeometry),
m_blockSet(&blockSet),
m_actualCrossing(0),
m_actualRoad(0),
m_env(&env),
m_density(&density)
{
	// on initialise la liste des marques
	for (std::vector<Road*>::const_iterator it = m_roads->begin(); it != m_roads->end(); ++it)
	{
		m_roadMarks[*it] = RoadMark();
	}	
}

GenBlocks::~GenBlocks(void)
{
}


bool GenBlocks::isFinished() const
{
	return m_actualRoad >= m_roads->size();
}

void GenBlocks::commitChange()
{
	m_roadsGeometry->commitChange();
	m_blockSet->commitChange();
}

void GenBlocks::createParameters(Control *control)
{
	control->addParameter("waywidth", "Way Width", 0.1f, 0.3f, 0.2f, 0.005f);
	control->addParameter("streetsidewalkwidth", "Street Sidewalk Width", 0.03f, 0.3f, 0.08f, 0.005f);
	control->addParameter("highwaysidewalkwidth", "Highway Sidewalk Width", 0.03f, 0.3f, 0.12f, 0.005f);
	control->addParameter("sidewalkheight", "Sidewalk Height", 0.005f, 0.03f, 0.012f, 0.001f);
	control->addParameter("centercommercialprop", "Center Commercial Proportion", 0.0f, 1.0f, 0.9f);
	control->addParameter("outskirtcommercialprop", "Outskirt Commercial Proportion", 0.0f, 1.0f, 0.1f);
	control->addParameter("centerdensitytreshold", "Center Density Treshold", 0.0f, 1.0f, 0.7f);
}

void GenBlocks::attachControl(Control *control)
{
	control->attachValue("waywidth", m_wayWidth);
	control->attachValue("streetsidewalkwidth", m_sidewalkStreetWidth);
	control->attachValue("highwaysidewalkwidth", m_sidewalkHighwayWidth);
	control->attachValue("sidewalkheight", m_sidewalkHeight);
	control->attachValue("centercommercialprop", m_centerCommercialProportion);
	control->attachValue("outskirtcommercialprop", m_outskirtCommercialProportion);
	control->attachValue("centerdensitytreshold", m_centerDensityTreshold);
}

void GenBlocks::finish()
{
	m_roadsGeometry->computeNormals();
}

void GenBlocks::proceed()
{
	if(m_actualCrossing < m_crossings->size())
	{
		computeCrossingGeometry(*(*m_crossings)[m_actualCrossing++]);
	}
	else
	{
		const Road *actual = (*m_roads)[m_actualRoad];
		Block *tmp;
		// pour cette route, on trouve le polygone  droite
		if (!m_roadMarks[actual].endBegin)
		{
			tmp = computeAdjancentBlock(*actual, true);
			if (tmp)
			{
				// et on l'ajoute  la liste
				m_blockSet->addBlock(*tmp);
			}
		}
		// et le polygone  gauche
		if (!m_roadMarks[actual].beginEnd)
		{
			tmp = computeAdjancentBlock(*actual, false);
			if (tmp)
			{
				// et on l'ajoute  la liste
				m_blockSet->addBlock(*tmp);
			}
		}

		// Et la route
		computeRoadGeometry(actual);

		// on passe ensuite  la route suivante
		m_actualRoad++;
	}
}

void GenBlocks::computeCrossingGeometry(Crossing &c)
{
	unsigned short nbRoads = static_cast<unsigned short>(c.getRoads().size());
	std::vector<Geometry::VertexStruct> vertices[3];

	computeCrossingVertices(c, vertices);

	unsigned short start[3];
	for(unsigned i = 0; i < 3; ++i)
	{
		start[i] = m_roadsGeometry->addVertices(vertices[i], i);
	}

	std::vector<unsigned short>	indices[3];
	indices[0].resize(3 * 4 * nbRoads);
	indices[1].resize(3 * 4 * nbRoads);
	indices[2].resize(3 * 3 * nbRoads);
	for(unsigned short i = 0; i < nbRoads; ++i)
	{
		unsigned short prec = (i == 0) ? (nbRoads - 1) : (i - 1);

		indices[0][3 * (4 * i + 0) + 0] = start[0] + 2 * i + 0;
		indices[0][3 * (4 * i + 0) + 1] = start[0] + 2 * i + 1;
		indices[0][3 * (4 * i + 0) + 2] = start[0] + 2 * nbRoads + 4 * i + 0;

		indices[0][3 * (4 * i + 1) + 0] = start[0] + 2 * nbRoads + 4 * i + 1;
		indices[0][3 * (4 * i + 1) + 1] = start[0] + 2 * nbRoads + 4 * i + 0;
		indices[0][3 * (4 * i + 1) + 2] = start[0] + 2 * i + 1;

		indices[0][3 * (4 * i + 2) + 0] = start[0] + 2 * prec + 1;
		indices[0][3 * (4 * i + 2) + 1] = start[0] + 2 * prec + 0;
		indices[0][3 * (4 * i + 2) + 2] = start[0] + 2 * nbRoads + 4 * i + 3;

		indices[0][3 * (4 * i + 3) + 0] = start[0] + 2 * nbRoads + 4 * i + 3;
		indices[0][3 * (4 * i + 3) + 1] = start[0] + 2 * nbRoads + 4 * i + 2;
		indices[0][3 * (4 * i + 3) + 2] = start[0] + 2 * prec + 1;


		indices[1][3 * (4 * i + 0) + 0] = start[1] + 2 * i + 0;
		indices[1][3 * (4 * i + 0) + 1] = start[1] + 2 * i + 1;
		indices[1][3 * (4 * i + 0) + 2] = start[1] + 2 * nbRoads + 4 * i + 0;

		indices[1][3 * (4 * i + 1) + 0] = start[1] + 2 * nbRoads + 4 * i + 1;
		indices[1][3 * (4 * i + 1) + 1] = start[1] + 2 * nbRoads + 4 * i + 0;
		indices[1][3 * (4 * i + 1) + 2] = start[1] + 2 * i + 1;

		indices[1][3 * (4 * i + 2) + 0] = start[1] + 2 * prec + 1;
		indices[1][3 * (4 * i + 2) + 1] = start[1] + 2 * prec + 0;
		indices[1][3 * (4 * i + 2) + 2] = start[1] + 2 * nbRoads + 4 * i + 3;

		indices[1][3 * (4 * i + 3) + 0] = start[1] + 2 * nbRoads + 4 * i + 3;
		indices[1][3 * (4 * i + 3) + 1] = start[1] + 2 * nbRoads + 4 * i + 2;
		indices[1][3 * (4 * i + 3) + 2] = start[1] + 2 * prec + 1;


		indices[2][3 * 3 * i + 0] = start[2] + nbRoads;
		indices[2][3 * 3 * i + 1] = start[2] + i;
		indices[2][3 * 3 * i + 2] = start[2] + prec;

		indices[2][3 * (3 * i + 1) + 0] = start[2] + prec;
		indices[2][3 * (3 * i + 1) + 1] = start[2] + i;
		indices[2][3 * (3 * i + 1) + 2] = start[2] + nbRoads + 1 + 2* i + 0;

		indices[2][3 * (3 * i + 2) + 0] = start[2] + i;
		indices[2][3 * (3 * i + 2) + 1] = start[2] + nbRoads + 1 + 2* i + 1;
		indices[2][3 * (3 * i + 2) + 2] = start[2] + nbRoads + 1 + 2* i + 0;
	}

	for(unsigned i = 0; i < 3; ++i)
	{
		m_roadsGeometry->addIndices(indices[i], i);
	}

	std::vector<unsigned short> &roadsIndices = m_crossingVertices[&c];
	roadsIndices.resize(nbRoads * 10);
	for(unsigned short i = 0; i < nbRoads; ++i)
	{
		roadsIndices[10 * i + 0] = start[0] + 2 * nbRoads + 4 * i + 0;
		roadsIndices[10 * i + 1] = start[0] + 2 * nbRoads + 4 * i + 1;
		roadsIndices[10 * i + 2] = start[1] + 2 * nbRoads + 4 * i + 0;
		roadsIndices[10 * i + 3] = start[1] + 2 * nbRoads + 4 * i + 1;
		roadsIndices[10 * i + 4] = start[2] + nbRoads + 1 + 2 * i + 0;
		roadsIndices[10 * i + 5] = start[2] + nbRoads + 1 + 2 * i + 1;
		roadsIndices[10 * i + 6] = start[1] + 2 * nbRoads + 4 * i + 2;
		roadsIndices[10 * i + 7] = start[1] + 2 * nbRoads + 4 * i + 3;
		roadsIndices[10 * i + 8] = start[0] + 2 * nbRoads + 4 * i + 2;
		roadsIndices[10 * i + 9] = start[0] + 2 * nbRoads + 4 * i + 3;
	}
}

D3DXVECTOR3 translatePoint(const D3DXVECTOR3 &point, const D3DXVECTOR2 &dir, float distance)
{
	float d = point.x * dir.x + point.z * dir.y;
	return point + D3DXVECTOR3(dir.x, 0.0f, dir.y) * (distance - d);
}

void GenBlocks::computeCrossingVertices(Crossing &c, std::vector<Geometry::VertexStruct> *vertices)
{
	unsigned short nbRoads = static_cast<unsigned short>(c.getRoads().size());
	std::vector<D3DXVECTOR3> junctions(3 * nbRoads);
	std::vector<D3DXVECTOR3> &wallVertices = m_wallVertices[&c];

	vertices[0].resize(2 * nbRoads + 4 * nbRoads);
	vertices[1].resize(2 * nbRoads + 4 * nbRoads);
	vertices[2].resize(nbRoads + 1 + 2 * nbRoads);

	wallVertices.resize(3 * nbRoads);

	for(unsigned short i = 0; i < nbRoads; ++i)
	{
		D3DXVECTOR2 sidewalk, wall;
		computeCorner(*c.getRoads()[i], *c.getRoads()[(i + 1) % nbRoads], c, sidewalk, wall);
		D3DXVECTOR3 toSidewalk(sidewalk.x, 0.0f, sidewalk.y);
		vertices[2][i].position = c.getPosition() + toSidewalk;
		vertices[1][2 * i + 0].position = c.getPosition() + toSidewalk;
		vertices[1][2 * i + 1].position = c.getPosition() + D3DXVECTOR3(sidewalk.x, m_sidewalkHeight, sidewalk.y);
		vertices[0][2 * i + 0].position = c.getPosition() + D3DXVECTOR3(sidewalk.x, m_sidewalkHeight, sidewalk.y);
		vertices[0][2 * i + 1].position = c.getPosition() + D3DXVECTOR3(wall.x, m_sidewalkHeight, wall.y);

		junctions[3 * i + 1] =  c.getPosition() + toSidewalk * 0.5f;

		wallVertices[3 * i + 1] = vertices[0][2 * i + 1].position; 
	}
	vertices[2][nbRoads].position = c.getPosition();

	for(unsigned short i = 0; i < nbRoads; ++i)
	{
		unsigned short prec = (i == 0) ? (nbRoads - 1) : (i - 1);
		D3DXVECTOR2 roadDir = c.getRoads()[i]->get2DVector();
		D3DXVec2Normalize(&roadDir, &roadDir);

		if(&c.getRoads()[i]->getStartCrossing() != &c)
		{
			roadDir = -roadDir;
		}

		const D3DXVECTOR3 &point3D1 = vertices[0][2 * i + 1].position;
		const D3DXVECTOR3 &point3D2 = vertices[0][2 * prec + 1].position;
		D3DXVECTOR2 point1 = D3DXVECTOR2(point3D1.x, point3D1.z);
		D3DXVECTOR2 point2 = D3DXVECTOR2(point3D2.x, point3D2.z);
		D3DXVECTOR2 center = c.get2DPosition();

		float proj1 = D3DXVec2Dot(&point1, &roadDir);
		float proj2 = D3DXVec2Dot(&point2, &roadDir);
		float projCenter = D3DXVec2Dot(&center, &roadDir);
		float distance = max(proj1, proj2);
		if(distance < projCenter) distance = projCenter + 0.01f;

		vertices[2][nbRoads + 1 + 2 * i + 0].position = translatePoint(vertices[2][prec].position, roadDir, distance);
		vertices[2][nbRoads + 1 + 2 * i + 1].position = translatePoint(vertices[2][i].position, roadDir, distance);
		vertices[1][2 * nbRoads + 4 * i + 0].position = translatePoint(vertices[1][2 * i + 0].position, roadDir, distance);
		vertices[1][2 * nbRoads + 4 * i + 1].position = translatePoint(vertices[1][2 * i + 1].position, roadDir, distance);
		vertices[1][2 * nbRoads + 4 * i + 2].position = translatePoint(vertices[1][2 * prec + 1].position, roadDir, distance);
		vertices[1][2 * nbRoads + 4 * i + 3].position = translatePoint(vertices[1][2 * prec + 0].position, roadDir, distance);
		vertices[0][2 * nbRoads + 4 * i + 0].position = translatePoint(vertices[0][2 * i + 0].position, roadDir, distance);
		vertices[0][2 * nbRoads + 4 * i + 1].position = translatePoint(vertices[0][2 * i + 1].position, roadDir, distance);
		vertices[0][2 * nbRoads + 4 * i + 2].position = translatePoint(vertices[0][2 * prec + 1].position, roadDir, distance);
		vertices[0][2 * nbRoads + 4 * i + 3].position = translatePoint(vertices[0][2 * prec + 0].position, roadDir, distance);

		junctions[3 * i + 0] = translatePoint(junctions[3 * i + 1], roadDir, distance);
		junctions[3 * prec + 2] = translatePoint(junctions[3 * prec + 1], roadDir, distance);

		wallVertices[3 * i + 0] = vertices[0][2 * nbRoads + 4 * i + 1].position;
		wallVertices[3 * prec + 2] = vertices[0][2 * nbRoads + 4 * i + 2].position;
	}

	c.setJunctionPositions(junctions);
}

void GenBlocks::computeRoadGeometry(const Road *r)
{
	std::vector<unsigned short> &indices1 = m_crossingVertices[&r->getStartCrossing()];
	std::vector<unsigned short> &indices2 = m_crossingVertices[&r->getEndCrossing()];

	unsigned index1 = r->getStartCrossing().getRoadIndex(r);
	unsigned index2 = r->getEndCrossing().getRoadIndex(r);

	std::vector<unsigned short> geometry[3];

	geometry[0].resize(12);
	geometry[1].resize(12);
	geometry[2].resize(6);

	static const unsigned layer[5] = { 0, 1, 2, 1, 0 };
	static const unsigned offset[5] = { 0, 0, 0, 1, 1 };

	for(unsigned i = 0; i < 5; ++i)
	{
		std::vector<unsigned short> &array = geometry[layer[i]];
		array[6 * offset[i] + 0] = indices1[10 * index1 + 2*i];
		array[6 * offset[i] + 1] = indices1[10 * index1 + 2*i + 1];
		array[6 * offset[i] + 2] = indices2[10 * index2 + 8 - 2*i];

		array[6 * offset[i] + 3] = indices1[10 * index1 + 2*i];
		array[6 * offset[i] + 4] = indices2[10 * index2 + 8 - 2*i];
		array[6 * offset[i] + 5] = indices2[10 * index2 + 9 - 2*i];
	}

	for(unsigned i = 0; i < 3; ++i)
	{
		m_roadsGeometry->addIndices(geometry[i], i);
	}
}

void GenBlocks::computeCorner(const Road &r1, const Road &r2, const Crossing &common, D3DXVECTOR2 &sidewalk, D3DXVECTOR2 &wall) const
{
	D3DXVECTOR2 v1 = r1.get2DVector();
	D3DXVECTOR2 v2 = r2.get2DVector();

	if(&r1.getStartCrossing() != &common) v1 = -v1;
	if(&r2.getStartCrossing() != &common) v2 = -v2;

	D3DXVec2Normalize(&v1, &v1);
	D3DXVec2Normalize(&v2, &v2);

	D3DXVECTOR2 n1(v2.y, -v2.x);
	float cos = D3DXVec2Dot(&v1, &n1);
	if(fabs(cos) < 0.1f)
	{
		float width = max(getRoadWidth(r2), getRoadWidth(r1));
		wall = (width + max(getSidewalkWidth(r2), getSidewalkWidth(r1))) * n1;
		sidewalk = width * n1;
	}
	else
	{
		float width1 = getRoadWidth(r1);
		float width2 = getRoadWidth(r2);
		wall = ((width2 + getSidewalkWidth(r2)) * v1 + (width1 + getSidewalkWidth(r1)) * v2) / cos;
		sidewalk = (width2 * v1 + width1 * v2) / cos;
	}
}

D3DXVECTOR3 GenBlocks::computeNormal(const Road &r1, const Road &r2, const Crossing &common) const
{
	D3DXVECTOR3 v1 = r1.get3DVector(&common);
	D3DXVECTOR3 v2 = r2.get3DVector(&common);
	D3DXVECTOR3 right1(v1.z, 0.0f, -v1.x);
	D3DXVECTOR3 right2(v2.z, 0.0f, -v2.x);

	D3DXVECTOR3 up1, up2;
	D3DXVec3Cross(&up1, &v1, &right1);
	D3DXVec3Cross(&up2, &v2, &right2);
	D3DXVec3Normalize(&up1, &up1);
	D3DXVec3Normalize(&up2, &up2);

	up1 += up2;
	D3DXVec3Normalize(&up1, &up1);
	return up1;
}

float GenBlocks::getRoadWidth(const Road &r) const
{
	return (r.getType() == RT_STREET) ? m_wayWidth : 2.0f * m_wayWidth;
}

float GenBlocks::getSidewalkWidth(const Road &r) const
{
	return (r.getType() == RT_STREET) ? m_sidewalkStreetWidth : m_sidewalkHighwayWidth;
}

Block* GenBlocks::computeAdjancentBlock(const Road &r, bool direction)
{
	const Crossing* beginning = (direction) ? &r.getStartCrossing()	: &r.getEndCrossing();
	std::vector<D3DXVECTOR3> points;
	std::vector<Road *> roads;
	points.clear();

	// tant que le "bot" de la route retourn est diffrent de la fin
	const Crossing* currentCrossing = beginning;
	const Road* currentRoad = &r, *nextRoad = NULL;
	int roadCount = 0;
	bool hasBridge = false;

	do
	{
		// /!\ HACK /!\ 
		roads.push_back(const_cast<Road *>(currentRoad));
		if(currentRoad->isBridge())
		{
			hasBridge = true;
		}
		// on prend la route la plus  droite suivante
		nextRoad = findRightRoad(*currentRoad, *currentCrossing);

		// on ajoute les points ncessaire  la liste
		addPoints(*currentRoad, *currentCrossing, points);

		// puis le ct oppos sur cette route, et on marque le passage sur la route, selon le sens
		if (currentCrossing == &nextRoad->getStartCrossing())
		{
			currentCrossing = &nextRoad->getEndCrossing();
			// on part du dbut de la route vers la fin
			Assert(!m_roadMarks[nextRoad].beginEnd);
			m_roadMarks[nextRoad].beginEnd = true;
		}
		else
		{
			currentCrossing = &nextRoad->getStartCrossing();
			// on part de la fin de la route vers le dbut
			Assert(!m_roadMarks[nextRoad].endBegin);
			m_roadMarks[nextRoad].endBegin = true;
		}

		roadCount++;
		currentRoad = nextRoad;
	}
	while (beginning != currentCrossing);

	if(hasBridge || points.size() > BLOCK_POLYGON_MAX_POINTS)
	{
		return NULL;
	}

	if(points.size() > BLOCK_PARK_MIN_POINTS)
	{
		return new Block(points, PARK);
	}
	else
	{
		D3DXVECTOR2 midPoint(0.0f, 0.0f);
		for(unsigned i = 0; i < points.size(); ++i)
		{
			midPoint += D3DXVECTOR2(points[i].x, points[i].z);
		}
		midPoint /= static_cast<float>(points.size());
		ZoneType type = RESIDENTIAL;

		if(m_density->getDensityMap().interpolate(midPoint.x, midPoint.y) >= m_centerDensityTreshold)
		{
			if(randomBool(m_centerCommercialProportion))
			{
				type = COMMERCIAL;
			}
		}
		else
		{
			if(randomBool(m_outskirtCommercialProportion))
			{
				type = COMMERCIAL;
			}
		}

		ERoadZoneType roadZoneType= (type == COMMERCIAL) ? RZT_COMMERCIAL : RZT_RESIDENTIAL;

		for(unsigned i = 0; i < roads.size(); ++i)
		{
			roads[i]->addZoneType(roadZoneType);
		}
		
		return new Block(points, type);
	}
}

const Road* GenBlocks::findRightRoad(const Road &r, const Crossing &c) const
{
	return c.getRoads()[(c.getRoadIndex(&r) + 1) % c.getRoads().size()];
}

void GenBlocks::addPoints(const Road &r1, const Crossing &common, std::vector<D3DXVECTOR3> &points) const
{
	/*D3DXVECTOR2 sidewalk, wall;
	computeCorner(r1, r2, common, sidewalk, wall);
	points.push_back(common.getPosition() + D3DXVECTOR3(wall.x, 0.012f, wall.y));*/

	std::vector<D3DXVECTOR3> &vertices = *m_wallVertices.get(&common);
	unsigned index = common.getRoadIndex(&r1);
	for(unsigned i = 0; i < 3; ++i)
	{
		points.push_back(vertices[3 * index + i]);
	}
}