#include "GenBuilding.h"
#include "Block.h"
#include "utils.h"
#include <cmath>
#include "Control.h"
#include "MemoryLog.h"

#define MIN_HEIGHT	0.2f
#define MAX_HEIGHT	0.4f
#define PROB_BUILDING	0.8f

GenBuilding::GenBuilding(const BlockSet &allotments, BuildingSet &buildings, Density &density):
Generator("Buildings"),
m_allotments(&allotments),
m_buildings(&buildings),
m_actBlock(0),
m_density(&density)
{
}

GenBuilding::~GenBuilding(void)
{
}

static float maxY(const std::vector<D3DXVECTOR3> &vertices)
{
	float maxY = vertices[0].y;
	for(unsigned i = 1; i < vertices.size(); i++)
	{
		if (vertices[i].y > maxY) maxY = vertices[i].y;
	}
	return maxY;
}

/*
Fonction permettant de dterminer le toit d'un building
*/
static void roof(const Building &building, Face &outface, float height)
{
	D3DXVECTOR3 vertex;
	float roofLevel = maxY(building.getAllotment().getVertices()) + height;
	const std::vector<Edge> &edges = building.getAllotment().getEdges();
	for(unsigned i = 0; i < edges.size(); i++)
	{	
		Edge e;
		e.v1 = D3DXVECTOR3(edges[i].v1.x, roofLevel, edges[i].v1.z);
		e.v2 = D3DXVECTOR3(edges[i].v2.x, roofLevel, edges[i].v2.z);
		outface.vertices.push_back(e.v1);
		outface.edges.push_back(e);
	}
}

/*
Fonction permettant de crer les faces autour d'un building
*/
static void around(std::vector<Edge>& basis, std::vector<Edge>& roof, Building &building)
{
	D3DXVECTOR3 up(0.0f, 1.0f, 0.0f);
	for(unsigned i = 0; i < basis.size(); i++)
	{
		D3DXVECTOR3 edge = basis[i].v2 - basis[i].v1;
		D3DXVECTOR3 normal;
		edge.y = 0.0f;
		D3DXVec3Normalize(&edge, &edge);
		D3DXVec3Cross(&normal, &edge, &up);

		building.addVertex(basis[i].v1);
		building.addVertex(roof[i].v1);
		building.addVertex(basis[i].v2);
		building.addVertex(roof[i].v2);

		building.addIndex((4 * i + 2));
		building.addIndex(4 * i + 1);
		building.addIndex(4 * i + 0);

		building.addIndex((4 * i + 2));
		building.addIndex((4 * i + 3));
		building.addIndex(4 * i + 1);
	}
}

static bool isIntersect(const D3DXVECTOR3 &edge1, const D3DXVECTOR3 &edge2, const D3DXVECTOR3 &A, const D3DXVECTOR3 &B)
{
	D3DXVECTOR2 u = D3DXVECTOR2(edge1.x, edge1.z);
	D3DXVECTOR2 v = D3DXVECTOR2(edge2.x, edge2.z);
	D3DXVECTOR2 w = D3DXVECTOR2(A.x - B.x, A.z - B.z);

	float d = D3DXVec2CCW(&u, &v);

	if(std::fabs(d) < 0.000001f)
	{
		return false;
	}

	d = 1.0f / d;

	float t1 = D3DXVec2CCW(&v, &w) * d;
	float t2 = D3DXVec2CCW(&u, &w) * d;

	if(t1 < 0.0001f || t1 > 0.9999f || t2 < 0.0001f || t2 > 0.9999f)
	{
		return false;
	}

	return true;
}

static float earCos(const std::vector<D3DXVECTOR3> &vertices, std::vector<unsigned> &indices, unsigned prec, unsigned index, unsigned next)
{
	D3DXVECTOR3 v1 = vertices[indices[prec]] - vertices[indices[index]];
	D3DXVECTOR3 v2 = vertices[indices[next]] - vertices[indices[index]];
	D3DXVECTOR3 cross;
	D3DXVec3Cross(&cross, &v2, &v1);

	if(cross.y < 0.0f)
	{
		return -1.0f;
	}

	D3DXVECTOR3 edge1 = vertices[indices[next]] - vertices[indices[prec]];

	for(unsigned i = 0; i < indices.size() - 1; ++i)
	{
		D3DXVECTOR3 edge2 = vertices[indices[i]] - vertices[indices[i+1]];

		if(isIntersect(edge1, edge2, vertices[indices[prec]], vertices[indices[i + 1]]))
		{
			return -1.0f;
		}
	}

	D3DXVec3Normalize(&v1, &v1);
	D3DXVec3Normalize(&v2, &v2);
	return D3DXVec3Dot(&v1, &v2);
}

static void constructRoofRec(const std::vector<D3DXVECTOR3> &vertices, std::vector<unsigned> &indices, Building *b, unsigned start)
{
	if(indices.size() == 3)
	{
		b->addIndex(start + indices[0]);
		b->addIndex(start + indices[1]);
		b->addIndex(start + indices[2]);
	}
	else
	{
		unsigned index = 0;
		float maxCos = -1.0f;
		for(unsigned i = 0; i < indices.size(); ++i)
		{
			unsigned prec = (i == 0) ? (indices.size() - 1) : (i - 1);
			unsigned next = (i == indices.size() - 1) ? 0 : (i + 1);
			float cos = earCos(vertices, indices, prec, i, next);
			if(cos > maxCos)
			{
				index = i;
				maxCos = cos;
			}
		}

		b->addIndex(start + indices[(index == 0) ? (indices.size() - 1) : (index - 1)]);
		b->addIndex(start + indices[index]);
		b->addIndex(start + indices[(index == indices.size() - 1) ? 0 : (index + 1)]);

		indices.erase(indices.begin() + index);
		constructRoofRec(vertices, indices, b, start);
	}
}

static void constructRoof(const std::vector<Edge> &roof, Building *b)
{
	std::vector<D3DXVECTOR3>	vertices(roof.size());
	std::vector<unsigned>		indices(roof.size());
	unsigned start = b->getVertices().size();

	for(unsigned i = 0; i < roof.size(); ++i)
	{
		vertices[i] = roof[i].v1;
		indices[i] = i;

		b->addVertex(roof[i].v1);
	}
	constructRoofRec(vertices, indices, b, start);
}

Building* GenBuilding::generate(Allotment *allotment)
{
	Building *building = NULL;
	if(allotment->getZoneType() >= PARK)
	{
		Face floorFaces;
		building = new Building(allotment, 0.0f);
		floorFaces.vertices = building->getAllotment().getVertices();
		floorFaces.edges = building->getAllotment().getEdges();
		constructRoof(floorFaces.edges, building);
	}
	else
	{
		float density = m_density->getDensityMap().interpolate(allotment->getPosition().x, allotment->getPosition().y);
		float height = randomFloatInRange(m_heightBuilding, m_buildingRange);
		
		if(randomBool(density - m_thresholdDensity1))
		{
			height *= m_coeffHeight1;
		}

		if(randomBool((density - m_thresholdDensity2) * 2.0f))
		{
			height *= m_coeffHeight2;
		}
		
		Face roofFaces;
		building = new Building(allotment, height);
		roof(*building, roofFaces, height);
		around(building->getAllotment().getEdges(), roofFaces.edges, *building);
		constructRoof(roofFaces.edges, building);
	}
	building->computeNormals();
	return building;
}

bool GenBuilding::isFinished() const
{
	return m_actBlock >= m_allotments->getBlocks().size();
}

void GenBuilding::commitChange()
{
	m_buildings->commitChange();
}

void GenBuilding::createParameters(Control *control)
{
	control->addParameter("heightbuilding", "Buildings Height", 0.1f, 1.0f, 0.3f);
	control->addParameter("buildingrange", "Buildings Range", 0.0f, 1.0f, 0.3f);
	control->addParameter("thresholddensity1", "Treshold Density First Zone", 0.0f, 1.0f, 0.5f);
	control->addParameter("coeffheight1", "Coeff. Height In First Zone", 1.0f, 3.0f, 2.0f);
	control->addParameter("thresholddensity2", "Treshold Density Second Zone", 0.0f, 1.0f, 0.9f);
	control->addParameter("coeffheight2", "Coeff. Height In Second Zone", 1.0f, 6.0f, 4.0f);
}

void GenBuilding::attachControl(Control *control)
{
	Generator::attachControl(control);

	control->attachValue("heightbuilding",		m_heightBuilding);
	control->attachValue("buildingrange",		m_buildingRange);
	control->attachValue("thresholddensity1",	m_thresholdDensity1);
	control->attachValue("coeffheight1",		m_coeffHeight1);
	control->attachValue("thresholddensity2",	m_thresholdDensity2);
	control->attachValue("coeffheight2",		m_coeffHeight2);
}

void GenBuilding::proceed(void)
{
	const Block &block = *m_allotments->getBlocks()[m_actBlock++];
	Allotment a(block.getZoneType(), block.getVertices());
	Building *b = generate(&a); 
	m_buildings->addBuilding(*b, block.getZoneType());
}