#include "GenCenters.h"

#include "Density.h"
#include "utils.h"
#include "Environment.h"
#include <cmath>
#include <vector>
#include <queue>
#include "Control.h"
#include "MemoryLog.h"

#define POSITIONNING_STRENGTH_BORDERS		6.0f
#define POSITIONNING_STRENGTH_CENTERS		2.0f
#define POSITIONNING_ZERO					0.0001f
#define POSITIONNING_STRENGTH_REDUCTION		0.01f
#define MIN_HEIGHT_OUT_OF_WATER				10.0f

#define OPTIMUM_PRECISION					0.008f

#define NB_EXECUTIONS_BEFORE_STOP			5000


/*****************************************************************************************************************************
										CONSTRUCTEUR / DESTRUCTEUR
*****************************************************************************************************************************/
GenCenters::GenCenters(Density &density):
Generator("Centers"),
m_density(&density),
m_nbExecutions(0),
m_startExec(true),
m_closestPointOutOfWater(NULL)
{
}

GenCenters::~GenCenters(void)
{
	delete[] m_closestPointOutOfWater;
}

/*****************************************************************************************************************************
												FONCTIONS PUBLIQUES
*****************************************************************************************************************************/

void GenCenters::init()
{
	m_history.clear();

	// on gnre les centres
	for (unsigned i = 0; i < static_cast<unsigned>(m_nbCentersBasic); ++i)
	{
		genCenter(CENTER_BASIC);
	}

	for (unsigned i = 0; i < static_cast<unsigned>(m_nbCentersRaster); ++i)
	{
		genCenter(CENTER_RASTER);
	}

	for (unsigned i = 0; i < static_cast<unsigned>(m_nbCentersRadial); ++i)
	{
		genCenter(CENTER_RADIAL);
	}

	// on vrifie que le niveau de l'eau ne soit pas trop important
	if (m_density->getEnvironment().getWaterLevel() > m_density->getEnvironment().getMaxHeight())
	{
		m_density->getEnvironment().setWaterLevel(m_density->getEnvironment().getMaxHeight() - MIN_HEIGHT_OUT_OF_WATER);
	}

	computeClosestPointOutOfWater();
}

void GenCenters::genCenter(ECenterType type)
{
	D3DXVECTOR2 center;
	float innerRadius = randomFloatInRange(m_innerRadius, m_innerRadiusRange);
	float outerRadius = innerRadius * randomFloatInRange(m_outerRadiusProportion, m_outerRadiusProportionRange);
	center.x = randomFloat(0.0f, static_cast<float>(m_density->getGridSize()));
	center.y = randomFloat(0.0f, static_cast<float>(m_density->getGridSize()));
	m_density->addCenter(center, innerRadius, outerRadius, type);
	// initialisation de l'historique correspondante
	m_history.push_back(D3DXVECTOR2(-1.0f, -1.0f));
}

bool GenCenters::isFinished() const
{
	// on  fini si on a excut le gnrateur le nombre maximum de fois
	if (m_nbExecutions > NB_EXECUTIONS_BEFORE_STOP)
	{
		return true;
	}

	// Sinon, on a fini si on est sur un optimum local. Pour cela, on regarde l'historique des positions prcdentes.
	// Si sur les trois positions prcdentes, on ne bouge pas ou on revient sur notre position, on s'arrte.
	std::vector<Center>::const_iterator itCenters = m_density->getCenters().begin();
	for (	std::vector<D3DXVECTOR2>::const_iterator itHistory = m_history.begin();
			itHistory != m_history.end();
			++itHistory
			)
	{
		// si notre position actuelle est loigne de la positions prcdente...
		if (	itCenters->pos.x < itHistory->x - OPTIMUM_PRECISION || itCenters->pos.x > itHistory->x + OPTIMUM_PRECISION 
			||	itCenters->pos.y < itHistory->y - OPTIMUM_PRECISION || itCenters->pos.y > itHistory->y + OPTIMUM_PRECISION )
		{
			// ... on est encore en mouvement, donc on n'est pas sur un optimum
			return false;
		}

		if (itCenters != m_density->getCenters().end())
		{
			++itCenters;
		}
	}
	// tout les points sont sur un optimum local, on a fini de les placer
	//m_density->displayCenters(false); // on enlve l'affichage des centres
	return true;
}

void GenCenters::commitChange()
{
	m_density->updateCenters();
}

void GenCenters::createParameters(Control *control)
{
	control->addParameter("nbcentersbasic", "# Centers Basic", 0.0f, 10.0f, 4.0f, 1.0f);
	control->addParameter("nbcentersraster", "# Centers Raster", 0.0f, 10.0f, 0.0f, 1.0f);
	control->addParameter("nbcentersradial", "# Centers Radial", 0.0f, 10.0f, 0.0f, 1.0f);
	control->addParameter("innerradius", "Inner Radius", 5.0f, 50.0f, 15.0f);
	control->addParameter("innerradiusrange", "Inner Radius Variance", 0.0f, 1.0f, 0.2f);
	control->addParameter("outerradiusproportion", "Outer Radius Proportion", 1.0f, 10.0f, 4.0f);
	control->addParameter("outerradiusrange", "Outer Radius Variance", 0.0f, 1.0f, 0.10f);
	control->addParameter("waterlevel", "Water Level", -50.0f, 50.0f, -5.0f);
}

void GenCenters::attachControl(Control *control)
{
	Generator::attachControl(control);

	control->attachValue("nbcentersbasic", m_nbCentersBasic);
	control->attachValue("nbcentersraster", m_nbCentersRaster);
	control->attachValue("nbcentersradial", m_nbCentersRadial);
	control->attachValue("innerradius", m_innerRadius);
	control->attachValue("innerradiusrange", m_innerRadiusRange);
	control->attachValue("outerradiusproportion", m_outerRadiusProportion);
	control->attachValue("outerradiusrange", m_outerRadiusProportionRange);
	control->attachValue("waterlevel", m_density->getEnvironment().getWaterLevel());
}

/*****************************************************************************************************************************
												FONCTIONS PROTECTED
*****************************************************************************************************************************/

void GenCenters::proceed()
{
	if (m_startExec)
	{
		m_startExec = false;
		m_density->displayCenters(true);
	}
	std::vector<D3DXVECTOR2>::iterator itHistory = m_history.begin();

	// les centres se repoussent entre eux, les bord repoussent les centres, ainsi que l'eau
	for (	std::vector<Center>::iterator it = m_density->getCenters().begin();
			it != m_density->getCenters().end();
			++it )
	{
		// on met  jours l'historique correspondant
		*itHistory = it->pos;
		// puis on dplace la position comme voulu
		it->pos += findPosition(*it);
		// et on passe  l'historique suivante, si ce n'est pas la dernire
		if (itHistory != m_history.end())
		{
			++itHistory; 
		}
	}

	// et on note une excution en plus
	m_nbExecutions++;
}

/*****************************************************************************************************************************
											FONCTIONS PRIVEES
*****************************************************************************************************************************/

D3DXVECTOR2 GenCenters::findPosition(const Center &c) const
{
	D3DXVECTOR2 dir(0.0f, 0.0f);

	/* 
	On commence par faire repousser le centre par tout les bords. Chaque bord repousse le centre avec une force inversement 
	proportionnelle  la  distance qui le separre du centre (en considrant le rayon).
	*/
	// par rapport au bord Y 0
	float dist = (c.pos.x - c.innerRadius <= 0.0f) ? 
						POSITIONNING_ZERO : 
						c.pos.x - c.innerRadius;
	dir += D3DXVECTOR2(POSITIONNING_STRENGTH_BORDERS / (dist * dist), 0.0f);
	// par rapport au bord Y gridSize
	dist = (m_density->getGridSize() - (c.pos.x + c.innerRadius) <= 0.0f) ? 
						POSITIONNING_ZERO : 
						m_density->getGridSize() - (c.pos.x + c.innerRadius);
	dir += D3DXVECTOR2( - POSITIONNING_STRENGTH_BORDERS / (dist * dist), 0.0f);
	// par rapport au bord X 0
	dist = (c.pos.y - c.innerRadius <= 0.0f) ? 
						POSITIONNING_ZERO : 
						c.pos.y - c.innerRadius;
	dir += D3DXVECTOR2(0.0f, POSITIONNING_STRENGTH_BORDERS / (dist * dist));
	// par rapport au bord X gridSize
	dist = (m_density->getGridSize() - (c.pos.y + c.innerRadius) <= 0.0f) ? 
						POSITIONNING_ZERO : 
						m_density->getGridSize() - (c.pos.y + c.innerRadius);
	dir += D3DXVECTOR2(0.0f, - POSITIONNING_STRENGTH_BORDERS / (dist * dist));

	// ensuite, chaque autres centres nous repousse
	for (	std::vector<Center>::const_iterator it = m_density->getCenters().begin();
			it != m_density->getCenters().end();
			++it )
	{
		if (&c == &*it)
		{
			continue;
		}

		D3DXVECTOR2 vDist = c.pos - it->pos ;
		float distFromCenter = D3DXVec2Length(&vDist) - (c.innerRadius+it->innerRadius);
		if (distFromCenter <= 0.0f)
		{
			distFromCenter = POSITIONNING_ZERO;
		}
		else
		{
			distFromCenter *= distFromCenter;
		}
		D3DXVec2Normalize(&vDist, &vDist);
		dir += ( POSITIONNING_STRENGTH_CENTERS / (distFromCenter * distFromCenter)) * vDist;
	}

	// en dernier lieu, on divise la force applique sur le centre
	dir *= POSITIONNING_STRENGTH_REDUCTION;
	float l;
	// on s'assure que la force applique ne nous fasse pas dplacer de plus de 3
	if((l = D3DXVec2LengthSq(&dir)) > 9.0f)
	{
		dir *= 3.0f / std::sqrt(l);
	}

	// si le centre est sous l'eau, on cherche le point le plus proche hors de l'eau et on s'y place
	if (	m_density->getEnvironment().getHeightMap()(static_cast<int>((c.pos + dir).x), static_cast<int>((c.pos + dir).y)) 
		<=	m_density->getEnvironment().getWaterLevel())
	{
		int pos[2] = { static_cast<int>((c.pos + dir).x), static_cast<int>((c.pos + dir).y) };
		int newPosX = m_closestPointOutOfWater[pos[0] * m_density->getEnvironment().getGridSize() + pos[1]].x;
		int newPosY = m_closestPointOutOfWater[pos[0] * m_density->getEnvironment().getGridSize() + pos[1]].y;
		dir = D3DXVECTOR2(static_cast<float>(newPosX), static_cast<float>(newPosY)) - c.pos;
		/*
		int pos[2] = { static_cast<int>((c.pos + dir).x), static_cast<int>((c.pos + dir).y) };
		int inc[2] = { 0, 0 };
		bool ind = 0;

		do
		{
			// on passe  l'incrment suivant
			inc[ind] *= -1;
			(inc[ind] >= 0) ? ++inc[ind] : --inc[ind];
			// on dplace la position de manire quivalente
			pos[ind] += inc[ind];
			// on varie ensuite sur l'autre axe
			ind = !ind;
			// on vrifie enfin que l'on ne soit pas hors de la map
			if (pos[0] < 0)
			{
				pos[0] = 0;
			}
			else if (pos[0] >= m_density->getEnvironment().getGridSize())
			{
				pos[0] = m_density->getEnvironment().getGridSize() - 1;
			}
			if (pos[1] < 0)
			{
				pos[1] = 0;
			}
			else if (pos[1] >= m_density->getEnvironment().getGridSize())
			{
				pos[1] = m_density->getEnvironment().getGridSize() - 1;
			}
			// si on a pass toute la map et qu'on a pas trouv de solutions, on arrete et on laisse le centre  la
			// position trouve prcdement
			if (	std::abs(pos[0]) >= m_density->getEnvironment().getGridSize() 
				&&	std::abs(pos[1]) >= m_density->getEnvironment().getGridSize())
			{
				pos[0] = static_cast<int>(dir.x);
				pos[1] = static_cast<int>(dir.y);
				break;
			}
		}
		while (m_density->getEnvironment().getHeightMap()(pos[0], pos[1]) <= m_density->getEnvironment().getWaterLevel());

		// on a trouv une position hors de l'eau proche, on chage dir en consquence
		dir = D3DXVECTOR2(static_cast<float>(pos[0]), static_cast<float>(pos[1])) - c.pos;
		*/
	}

	return dir;
}

void GenCenters::computeClosestPointOutOfWater()
{
	// on rserve l'espace mmoire et on initialise
	m_closestPointOutOfWater = new NearPoint[m_density->getEnvironment().getGridSize() * m_density->getEnvironment().getGridSize()];
	std::queue<std::pair<int, int> > outWater;

	// on initialise avec les points hors de l'eau
	for (int i = 0; i < m_density->getEnvironment().getGridSize(); i++)
	{
		for (int j = 0; j < m_density->getEnvironment().getGridSize(); j++)
		{
			NearPoint &p = m_closestPointOutOfWater[i * m_density->getEnvironment().getGridSize() + j];
			if (m_density->getEnvironment().getHeightMap()(i, j) > m_density->getEnvironment().getWaterLevel())
			{
				p.x = i;
				p.y = j;
				p.distance = 0;
				outWater.push(std::pair<int, int>(i, j));
			}
			else
			{
				p.distance = 0xFFFFFF;
			}
		}
	}

	// On propage les points mergs les plus proches
	while(!outWater.empty())
	{
		int x = outWater.front().first;
		int y = outWater.front().second;
		outWater.pop();

		int dx = 1;
		int dy = 0;
		const NearPoint &ref = m_closestPointOutOfWater[x * m_density->getEnvironment().getGridSize() + y];

		for(unsigned i = 0; i < 4; ++i)
		{
			int nx = x + dx;
			int ny = y + dy;

			if(nx >= 0 && ny >= 0 && nx < m_density->getEnvironment().getGridSize() && ny < m_density->getEnvironment().getGridSize())
			{
				NearPoint &p = m_closestPointOutOfWater[nx * m_density->getEnvironment().getGridSize() + ny];

				if(p.distance > ref.distance + 1)
				{
					p.x = ref.x;
					p.y = ref.y;
					p.distance = ref.distance + 1;

					outWater.push(std::pair<int, int>(nx, ny));
				}
			}

			std::swap(dx, dy);
			dx = -dx;
		}
	}
}