#include "GenDensity.h"

#include "utils.h"
#include "Density.h"
#include "Environment.h"
#include <cmath>
#include <list>
#include "Control.h"
#include "MemoryLog.h"

#define DENSITY_RADIUS			10
#define VARIANCE_AMPLIFICATION	10.0f
#define CENTER_IMPORTANCE		0.7f	// compris entre 0 et 1, donne la force de la prsence du centre par rapport au relief

GenDensity::GenDensity(Density &density):
Generator("Density"),
m_density(&density),
m_x(0),
m_y(0)
{
}

GenDensity::~GenDensity(void)
{
}

bool GenDensity::isFinished() const
{
	// on a fini quand on a parcouru toute la map
	if (m_y >= m_density->getGridSize())
	{
		m_density->displayCenters(false);
		return true;
	}
	return false;
}

void GenDensity::createParameters(Control *control)
{
	control->addParameter("innerradiusdensity", "Inner Radius Density", 0.1f, 1.0f, 0.8f);
	control->addParameter("densityattenuation", "Density Attenuation", 1.0f, 2.0f, 1.2f);
	control->addParameter("elevationattenuation", "Elevation Factor Attenuation", 1.0f, 100.0f, 10.0f);
}

void GenDensity::attachControl(Control *control)
{
	control->attachValue("innerradiusdensity", m_borderCenterDensity);
	control->attachValue("densityattenuation", m_pow);
	control->attachValue("elevationattenuation", m_elevationRange);
}

void GenDensity::proceed()
{
	// on calcule la densit pour le point actuel, si il n'est pas sous le niveau de la mer
	if (m_density->getEnvironment().getHeightMap()(m_x, m_y) >= m_density->getEnvironment().getWaterLevel())
	{
		float d = computeZoneDensity();
		m_density->getDensityMap()(m_x, m_y) = d;
		m_density->getTextureMap()(m_x, m_y) = static_cast<unsigned short>(d * 65535.0f);
	}
	else
	{
		// on est sous le niveau de l'eau. On assigne tout de mme la couleur noire  la texture
		m_density->getTextureMap()(m_x, m_y) = 0;
	}
	// puis on passe au point suivant
	m_x++;
	if (m_x >= m_density->getGridSize())
	{
		m_x = 0;
		m_y++;
	}
}

void GenDensity::commitChange()
{
	m_density->update();
}

float GenDensity::computeZoneDensity() const
{
	float	moyenne = 0.0f, 
			moyenneCarree = 0.0f;
	int		nbPoints = 0;

	// on parcours le carr de ct DENSITY_RADIUS
	for (int i = -DENSITY_RADIUS + m_x; i < DENSITY_RADIUS + m_x; i++)
	{
		// si on est sur une ligne en dehors de la map, on passe
		if (i < 0 || i >= m_density->getGridSize())
		{
			continue;
		}	
			
		for (int j = -DENSITY_RADIUS + m_y; j < DENSITY_RADIUS + m_y; j++)
		{
			// si on est sur une colonne en dehors de la map, on passe
			if (j < 0 || j >= m_density->getGridSize())
			{
				continue;
			}

			D3DXVECTOR2 distFromPosition(	static_cast<float>(m_x - i), 
											static_cast<float>(m_y - j));
			// on vrifie que le plan soit dans le rayon de rfrence
			if (	DENSITY_RADIUS * DENSITY_RADIUS 
				>=	D3DXVec2LengthSq(&distFromPosition))
			{
				// on compte un point de plus
				nbPoints++;
				// puis on ajoute l'altitude dans la moyenne
				moyenne += m_density->getEnvironment().getHeightMap()(i, j);
				// et dans la moyenne au carre
				moyenneCarree += m_density->getEnvironment().getHeightMap()(i, j) * m_density->getEnvironment().getHeightMap()(i, j);
			}
		}
	}

	// on calcule la moyenne et la moyenne carre, et on les ramne entre 0 et 1
	moyenne /= static_cast<float>(nbPoints);
	moyenneCarree /= static_cast<float>(nbPoints);
	float variance = (moyenneCarree - (moyenne * moyenne) ) / m_elevationRange;

	// une fois qu'on a la variance de la zone, on regarde la distance de la position par rapport aux centres
	float weight = 0.0f;
	for (	std::vector<Center>::const_iterator it = m_density->getCenters().begin() ;
			it != m_density->getCenters().end()	;
			++it )
	{
		D3DXVECTOR2 vDist = (*it).pos - D3DXVECTOR2(static_cast<float>(m_x), static_cast<float>(m_y));
		float dist = D3DXVec2Length(&vDist);

		// si notre position est  l'intrieur du innerRadius d'un centre, on a la densit max
		if (dist < (*it).innerRadius)
		{
			float d = dist / (*it).innerRadius;
			weight = 1.0f - pow(d, m_pow) * (1.0f - m_borderCenterDensity);
			break;
		}
		// si on est hors du rayon externe du Centre, celui-ci n'a aucun effet sur la position actuelle
		else if (dist > (*it).outerRadius)
		{
			continue;
		}
		// sinon, on applique un coefficient proportionnel  la distance jusqu'au centre, en considrant la distance ventuelle
		// jusqu'au centres prcdents (on fait une moyenne)
		else
		{
			float d = 1.0f - ((dist - it->innerRadius) / (it->outerRadius - it->innerRadius));
			weight = max(weight, pow(d, m_pow) * m_borderCenterDensity);
		
		}
	}
	
	//weight *= CENTER_IMPORTANCE;
	float density = weight - variance;
	clamp(density, 0.0f, 1.0f);

	return density;
}