#include "GenEnvironment.h"

#include "utils.h"
#include "Control.h"
#include <cmath>
#include "MemoryLog.h"

/*****************************************************************************************************************************
											CONSTRUCTEUR / DESTRUCTEUR
*****************************************************************************************************************************/

GenEnvironment::GenEnvironment(Environment &env):
Generator("Landscape"),
m_env(&env),
m_step(1)
{
}

GenEnvironment::~GenEnvironment(void)
{
}

/*****************************************************************************************************************************
											FONCTIONS PUBLIQUES
*****************************************************************************************************************************/

bool GenEnvironment::isFinished() const
{
	return m_step > static_cast<int>(m_octave);
}

void GenEnvironment::commitChange()
{
	m_env->synchroHeights();
}

void GenEnvironment::createParameters(Control *control)
{
	control->addParameter("persistence", "Persistence", 1.0f, 6.0f, 2.0f);
	control->addParameter("amplitude", "Amplitude", 0.0f, 50.0f, 10.0f);
	control->addParameter("octave", "Octave", 1.0f, static_cast<float>(m_env->getGridPrecision() - 1), 5.0f, 1.0f);
}

void GenEnvironment::attachControl(Control *control)
{
	Generator::attachControl(control);

	control->attachValue("persistence", m_persistence);
	control->attachValue("amplitude", m_amplitude);
	control->attachValue("octave", m_octave);
}

/*****************************************************************************************************************************
											FONCTIONS PROTECTED
*****************************************************************************************************************************/

void GenEnvironment::proceed()
{
	// N reprsente le nombre de cases sur un ct. On en a deux de plus pour les alatoires hors maps ncssaires  l'interpolation
	int n = m_env->getGridSize()+2;
	// les cases que l'on vient de bruiter ont leurs deux coordonnes qui sont des multiples de l'incrment (on soustrait 1 pour
	// se recaler sur les indice et 2 pour compenser les deux ajouts ci-dessus).
	int inc = (n-3) / (1 << (m_step-1));
	//float coef = 10.0f/static_cast<float>(1 << (2*(m_step-1)));
	float coef = m_amplitude / std::pow(m_persistence, m_step - 1);

	/*if(m_step > 2 && m_step < 5)
	{
		m_step++;
		return;
	}*/

	// on cr le tableau d'interpolation.
	float *interArray = new float[n*n];
	// on initialise les altitudes  0
	fillArray(interArray, 0.0f, n*n);

	// on gnre les alatoires "in-map"...
	for (int i = 0; i < n; i += inc)
	{
		for (int j = 0; j < n; j += inc)
		{
			interArray[i*n+j] = randomFloat(-coef, coef);
		}
	}

	// ... puis les alatoires "hors-map"
	for (int i = n - 2; i < n; i++ )
	{
		for (int j = 0; j < n-2; j += inc )
		{
			// on assigne en symtrie interArray[x][y] et interArray[y][x], car se sont toutes deux des valeurs utilises
			interArray[i*n+j] = randomFloat(-coef, coef);
			interArray[j*n+i] = randomFloat(-coef, coef);
		}
	}
	// on gnre aussi les quatres points des "coins"
	interArray[n*n - 1] = randomFloat(-coef, coef);
	interArray[n*n - 2] = randomFloat(-coef, coef);
	interArray[n*(n-1) - 1] = randomFloat(-coef, coef);
	interArray[n*(n-1) - 0] = randomFloat(-coef, coef);

	// on fait un premier passage o l'on remplit les lignes o l'on a dj des points de prciss. on ne passe pour le moment
	// que les lignes "in-map"
	for (int i=0; i<n-2; i+=inc)
	{
		// on passe tout les points "in-map"... 
		for (int j=0; j<n-2; j++)
		{
			// ... o on n'a pas dj tir la position au hasard
			if (j%n)
			{
				// on fait l'interpolation cubique pour le point actuel avec les quatres points gnrs qui l'entourent
				interArray[i*n+j] = cubicInterpolation (	
						interArray[findXReferencePoint(i, j, -1, inc, n)],
						interArray[findXReferencePoint(i, j,  0, inc, n)],
						interArray[findXReferencePoint(i, j,  1, inc, n)],
						interArray[findXReferencePoint(i, j,  2, inc, n)],
						static_cast<float>(j % inc)/static_cast<float>(inc)
					);
			}
			// sinon, on ne fait rien, le tableau contient dj la valeur gnre
		}
	}

	// on fait les deux lignes "hors-map"
	for (int i=n-2; i<n; i++)
	{
		// on passe tout les points "in-map"... 
		for (int j=0; j<n-2; j++)
		{
			// ... o on n'a pas dj tir la position au hasard
			if (j%n)
			{
				// on fait l'interpolation cubique pour le point actuel avec les quatres points gnrs qui l'entourent
				interArray[i*n+j] = cubicInterpolation (	
						interArray[findXReferencePoint(i, j, -1, inc, n)],
						interArray[findXReferencePoint(i, j,  0, inc, n)],
						interArray[findXReferencePoint(i, j,  1, inc, n)],
						interArray[findXReferencePoint(i, j,  2, inc, n)],
						static_cast<float>(j % inc)/static_cast<float>(inc)
					);
			}
			// sinon, on ne fait rien, le tableau contient dj la valeur gnre
		}
	}

	// on fait un second passage o l'on remplit les cases manquantes en utilisant les colonnes
	for (int i=0; i<n; i++)
	{
		// on zappe les lignes o l'on est dj pass prcdemment
		if (i % n)
		{
			for (int j = 0; j<n; j++)
			{
				// on fait l'interpolation cubique pour le point actuel avec les quatres points gnrs qui l'entourent
				interArray[i*n+j] = cubicInterpolation (	
						interArray[findYReferencePoint(i, j, -1, inc, n)],
						interArray[findYReferencePoint(i, j,  0, inc, n)],
						interArray[findYReferencePoint(i, j,  1, inc, n)],
						interArray[findYReferencePoint(i, j,  2, inc, n)],
						static_cast<float>(i % inc)/static_cast<float>(inc)
					);
			}
		}
	}

	// on additionne l'interpolation de l'tape et la heightMap actuelle
	for (int i = 0; i < (n-2); i++)
	{
		for (int j = 0; j < (n-2); j++)
		{
			m_env->getHeightMap()(i, j) += interArray[i*n+j];
			// on vrifie au passage les latitudes min et max, et on les met ventuellement  jour
			if (m_env->getMinHeight() > m_env->getHeightMap()(i, j))
			{
				m_env->setMinHeight(m_env->getHeightMap()(i, j));
			}
			else if (m_env->getMaxHeight() < m_env->getHeightMap()(i, j))
			{
				m_env->setMaxHeight(m_env->getHeightMap()(i, j));
			}
		}
	}

	// on libre l'espace mmoire
	delete[] interArray;
	// et on passe  l'tape suivante
	m_step++;
}

/*****************************************************************************************************************************
											GETTERS / SETTERS
*****************************************************************************************************************************/

float GenEnvironment::getAmplitude() const
{
	return m_amplitude;
}

float GenEnvironment::getPersistence() const
{
	return m_persistence;
}

void GenEnvironment::setAmplitude(float amplitude)
{
	m_amplitude = amplitude;
}

void GenEnvironment::setPersistence(float persistence)
{
	m_persistence = persistence;
}

/*****************************************************************************************************************************
										FONCTIONS PRIVEES ET UTILITAIRES
*****************************************************************************************************************************/
float GenEnvironment::cubicInterpolation(float y0, float y1, float y2, float y3, float x)
{
	float a = y3 - y2 - y0 + y1;
	float b = y0 - y1 - a;
	float c = y2 - y0;
	float d = y1;

	return a*x*x*x + b*x*x + c*x + d;
}

int GenEnvironment::findXReferencePoint(int i, int j, int refPointPosition, int inc, int width)
{
	// on travaille sur X, donc pas de modification de l'indice ligne, mais modification de l'indice colonne
	j = ((j/inc) + refPointPosition) * inc;
	// si on est sur une colonne d'indice ngatif, on se place dans la dernire colonne...
	if (j < 0)
	{
		j = width - 1;
	}
	// ... et si l'indice est en dehors du tableau "in-map", on se place dans l'avant dernire colonne
	else if (j > width -2)
	{
		j = width - 2;
	}
	return i * width + j;
}

int GenEnvironment::findYReferencePoint(int i, int j, int refPointPosition, int inc, int width)
{
	// on travaille sur Y, donc pas de modification de l'indice colonne, mais modification de l'indice ligne
	i = ((i/inc) + refPointPosition) * inc;
	// si on est sur une colonne d'indice ngatif, on se place dans la dernire ligne...
	if (i < 0)
	{
		i = width - 1;
	}
	// ... et si l'indice est en dehors du tableau "in-map", on se place dans l'avant dernire ligne
	else if (i > width -2)
	{
		i = width - 2;
	}
	return i * width + j;
}