#include "GenRoad.h"
#include "Environment.h"
#include "Density.h"
#include "Road.h"
#include "Crossing.h"
#include "RoadNetwork.h"
#include "utils.h"
#include "Control.h"
#include <limits>
#include "MemoryLog.h"

#define MIN_ROAD_LENGTH				  1.5f
#define SAMPLE_LENGTH				 30.0f
#define SAMPLE_INC					  1.0f
#define STREET_DELAY				250
#define NB_TEST_ROAD				  5
#define MIN_PRUNE_COEFF				  0.3f
#define INC_PRUNE_COEFF				  0.1f;
#define MAX_ANGLE_MOD				  (5.0f * M_PI / 12.0f)
#define INC_ANGLE_MOD				  0.1f
#define BRIDGE_INC					  1.0f

static const float s_minAngle[] = { -M_PI / 12.0f,	5.0f * M_PI / 12.0f,	-7.0f * M_PI / 12.0f };
static const float s_maxAngle[] = {  M_PI / 12.0f,	7.0f * M_PI / 12.0f,	-5.0f * M_PI / 12.0f };

static const float s_rasterAngle[] = { -M_PI / 2.0f, 0.0f, M_PI / 2.0f };

#undef max
#undef min

GenRoad::GenRoad(RoadNetwork &roads, const Environment &env, const Density &density):
Generator("Roads Network"),
m_roads(roads),
m_env(&env),
m_density(&density),
m_crossingsGrid(NULL),
m_roadsGrid(NULL)
{
}

GenRoad::~GenRoad()
{
	delete[] m_crossingsGrid;
	delete[] m_roadsGrid;

	while(!m_unassignedCrossings.empty())
	{
		Road *r = m_unassignedCrossings.front().first;
		delete &r->getEndCrossing();
		delete r;
		m_unassignedCrossings.pop();
	}
}

void GenRoad::init()
{
	m_crossingsGrid = new std::vector<Crossing *>[m_env->getGridSize() * m_env->getGridSize()];
	m_roadsGrid = new std::vector<Road*>[m_env->getGridSize() * m_env->getGridSize()];
	const std::vector<Center> &centers = m_density->getCenters();

	for(unsigned i = 0; i < centers.size(); ++i)
	{
		const D3DXVECTOR2 &pos = centers[i].pos;
		Crossing *center = new Crossing(D3DXVECTOR3(pos.x, m_env->getHeightMap().interpolate(pos.x, pos.y), pos.y));
		assignCrossing(center);
		
		initBasicCenter(center);
	}
}

void GenRoad::initBasicCenter(Crossing *center)
{
	const D3DXVECTOR2 &pos = center->get2DPosition();
	for(unsigned i = 0; i < 4; ++i)
	{
		float angle = randomFloat(0.0f, 2.0f * M_PI);
		float length = randomFloatInRange(m_streetLength, m_streetRange);
		D3DXVECTOR2 dir(std::cos(angle) * length, std::sin(angle) * length);
		Crossing *end = new Crossing(D3DXVECTOR3(pos.x + dir.x, 0.0f, pos.y + dir.y));
		Road *r = new Road(*center, *end, RT_HIGHWAY);
		end->addRoad(*r);

		m_unassignedCrossings.push(std::pair<Road *, RoadAttrib>(r,
			RoadAttrib(0, randomFloatInRange(m_highwayLength, m_highwayRange))));
	}
}

void GenRoad::initRasterCenter(Crossing *center)
{
	float angle = randomFloat(0.0f, 2.0f * M_PI);
	const D3DXVECTOR2 &pos = center->get2DPosition();
	for(unsigned i = 0; i < 4; ++i)
	{
		float length = randomFloatInRange(m_streetLength, m_streetRange);
		D3DXVECTOR2 dir(std::cos(angle) * length, std::sin(angle) * length);
		Crossing *end = new Crossing(D3DXVECTOR3(pos.x + dir.x, 0.0f, pos.y + dir.y));
		Road *r = new Road(*center, *end, RT_HIGHWAY);
		end->addRoad(*r);
		m_unassignedCrossings.push(std::pair<Road *, RoadAttrib>(r,
			RoadAttrib(0, randomFloatInRange(m_highwayLength, m_highwayRange))));

		angle += M_PI * 0.5f;
	}
}

void GenRoad::finish()
{
	m_roads.removeOneway();
	m_roads.sortRoadsOnCrossing();
	m_roads.commitChange();
}

void GenRoad::commitChange()
{
	m_roads.commitChange();
}

bool GenRoad::isFinished() const
{
	return m_unassignedCrossings.empty();
}

void GenRoad::createParameters(Control *control)
{
	control->addParameter("streetlength", "Street Length", MIN_ROAD_LENGTH, 10.0f, 5.0f);
	control->addParameter("streetrange", "Street Length Variance", 0.0f, 1.0f, 0.2f);
	control->addParameter("highwaylength", "Highway Length", MIN_ROAD_LENGTH, 100.0f, 30.0f);
	control->addParameter("highwayrange", "Highway Length Variance", 0.0f, 1.0f, 0.4f);
	control->addParameter("probhighwaychangeangle", "Prob. Highway Change Angle", 0.0f, 1.0f, 0.3f);
	control->addParameter("probstreetchangelength", "Prob. Street Change Length", 0.0f, 1.0f, 0.02f);
	control->addParameter("probstreetbranch", "Prob. Street Branch", 0.0f, 1.0f, 0.4f);
	control->addParameter("mergedistance", "Merging Ratio Distance", 0.0f, 1.0f, 0.7f);
	control->addParameter("densitytreshold", "Density Treshold", 0.0f, 1.0f, 0.2f);
	control->addParameter("maxbridgelength", "Max Bridge Length", 5.0f, 50.0f, 20.0f, 0.5f);
}

void GenRoad::attachControl(Control *control)
{
	Generator::attachControl(control);

	control->attachValue("streetlength", m_streetLength);
	control->attachValue("streetrange", m_streetRange);
	control->attachValue("highwaylength", m_highwayLength);
	control->attachValue("highwayrange", m_highwayRange);
	control->attachValue("probhighwaychangeangle", m_probHighwayChangeAngle);
	control->attachValue("probstreetchangelength", m_probStreetChangeLength);
	control->attachValue("probstreetbranch", m_probStreetBranch);
	control->attachValue("mergedistance", m_mergeDistance);
	control->attachValue("densitytreshold", m_densityTreshold);
	control->attachValue("maxbridgelength", m_maxBridgeLength);
}

ELocalConstraintResult GenRoad::localConstraint(Road *r)
{
	Crossing *endCrossing = &r->getEndCrossing();

	if(r->getStartCrossing().getNbStreet() >= 4)
	{
		return LCR_FAILED;
	}

	if(!adjustValidLocation(r))
	{
		return LCR_FAILED;
	}

	bool newCrossing = true;
	Road *intersectedRoad = NULL;
	ELocalConstraintResult result = tryIntersectingBackward(r, intersectedRoad);

	if(result == LCR_FAILED)
	{
		return LCR_FAILED;
	}
	else if(result == LCR_EXISTING_CROSSING)
	{
		newCrossing = false;
	}

	result = tryMergeCrossing(r);

	if(result == LCR_FAILED)
	{
		return LCR_FAILED;
	}
	else if(result == LCR_EXISTING_CROSSING)
	{
		intersectedRoad = NULL;
		newCrossing = false;
	}

	if(newCrossing)
	{
		result = tryIntersectingForward(r, intersectedRoad);

		if(result == LCR_FAILED)
		{
			return LCR_FAILED;
		}
		else if(result == LCR_EXISTING_CROSSING)
		{
			newCrossing = false;
		}

		result = tryMergeCrossing(r);

		if(result == LCR_FAILED)
		{
			return LCR_FAILED;
		}
		else if(result == LCR_EXISTING_CROSSING)
		{
			intersectedRoad = NULL;
			newCrossing = false;
		}
	}

	if(r->getStartCrossing().hasColinearRoad(r, 0.9f))
	{
		r->setCrossing(r->getEndCrossing(), *endCrossing);
		return LCR_FAILED;
	}

	if(r->getEndCrossing().hasColinearRoad(r, -0.9f))
	{
		r->setCrossing(r->getEndCrossing(), *endCrossing);
		return LCR_FAILED;
	}

	if(r->getEndCrossing().getNbStreet() >= 4)
	{
		r->setCrossing(r->getEndCrossing(), *endCrossing);
		return LCR_FAILED;
	}

	if(r->get2DLength() < MIN_ROAD_LENGTH)
	{
		r->setCrossing(r->getEndCrossing(), *endCrossing);
		return LCR_FAILED;
	}

	if(intersectedRoad)
	{
		if(canDivideRoad(*intersectedRoad, r->getEndCrossing()))
		{
			assignCrossing(&r->getEndCrossing());

			Road *newRoad = m_roads.divideRoad(*intersectedRoad, r->getEndCrossing());
			assignRoad(newRoad);
			r->getEndCrossing().removeRoad(*r);
		}
		else
		{
			r->setCrossing(r->getEndCrossing(), *endCrossing);
			return LCR_FAILED;
		}
	}

	if(endCrossing != &r->getEndCrossing())
	{
		delete endCrossing;
	}
	else
	{
		const D3DXVECTOR2 &pos = endCrossing->get2DPosition();
		endCrossing->getPosition().y = m_env->getHeightMap().interpolate(pos.x, pos.y);
	}

	if(newCrossing)
	{
		return LCR_NEW_CROSSING;
	}
	else
	{
		r->getEndCrossing().addRoad(*r);
		return LCR_EXISTING_CROSSING;
	}
}

bool GenRoad::adjustValidLocation(Road *r) const
{
	if(isValidLocation(r->getEndCrossing().get2DPosition()))
	{
		return true;
	}

	const D3DXVECTOR2 &start = r->getStartCrossing().get2DPosition();
	D3DXVECTOR2 dir = r->get2DVector();
	float length = D3DXVec2Length(&dir);
	dir /= length;
	float coeff = 1.0f - INC_PRUNE_COEFF;

	while(coeff >= MIN_PRUNE_COEFF)
	{
		D3DXVECTOR2 p = start + dir * (length * coeff);
		if(isValidLocation(p))
		{
			r->getEndCrossing().setPosition(D3DXVECTOR3(p.x, 0.0f, p.y));
			return true;
		}
		coeff -= INC_PRUNE_COEFF;
	}

	float angle = vectorAngle(dir);
	float modAngle = INC_ANGLE_MOD;

	while(std::fabs(modAngle) <= MAX_ANGLE_MOD)
	{
		D3DXVECTOR2 newdir(std::cos(angle + modAngle), std::sin(angle + modAngle));
		D3DXVECTOR2 p = start + newdir * length;
		if(isValidLocation(p))
		{
			r->getEndCrossing().setPosition(D3DXVECTOR3(p.x, 0.0f, p.y));
			return true;
		}
		modAngle = (modAngle > 0.0f) ? -modAngle : -modAngle + INC_ANGLE_MOD;
	}

	if(r->getType() == RT_HIGHWAY)
	{
		length += BRIDGE_INC;
		while(length <= m_maxBridgeLength)
		{
			D3DXVECTOR2 p = start + dir * length;
			if(isValidLocation(p))
			{
				r->setBridge();
				r->getEndCrossing().setPosition(D3DXVECTOR3(p.x, 0.0f, p.y));
				return true;
			}
			length += BRIDGE_INC;
		}
	}

	return false;
}

void GenRoad::globalConstraint(Road *r, const RoadAttrib &attrib)
{
	if(r->getType() == RT_STREET)
	{
		globalConstraintStreet(r, attrib);
	}
	else
	{
		const D3DXVECTOR2 &pos = r->getEndCrossing().get2DPosition();
		float minDistance = std::numeric_limits<float>::infinity();
		const Center *nearCenter = NULL;

		for(unsigned i = 0; i < m_density->getCenters().size(); ++i)
		{
			D3DXVECTOR2 diff = pos - m_density->getCenters()[i].pos;
			float length = D3DXVec2LengthSq(&diff);
			if(length < minDistance)
			{
				nearCenter = &m_density->getCenters()[i];
				minDistance = length;
			}
		}

		switch(nearCenter->type)
		{
		case CENTER_BASIC:
			globalConstraintBasic(r, attrib);
			break;
		case CENTER_RASTER:
			globalConstraintRaster(r, attrib);
			break;
		case CENTER_RADIAL:
			globalConstraintRadial(r, attrib, nearCenter->pos);
			break;
		}
	}
}

void GenRoad::globalConstraintStreet(Road *r, const RoadAttrib &attrib)
{
	D3DXVECTOR2 dir = r->get2DVector();
	float length = D3DXVec2Length(&dir);

	if(randomBool(m_probStreetChangeLength))
	{
		dir /= length;
		length = randomFloatInRange(m_streetLength, m_streetRange);
		dir *= length;
	}

	D3DXVECTOR2 perp = D3DXVECTOR2(-dir.y, dir.x);
	D3DXVec2Normalize(&perp, &perp);
	perp *= attrib.data;

	createNewRoad(r, dir, RT_STREET, RoadAttrib(0, attrib.data));
	createNewRoad(r, perp, RT_STREET, RoadAttrib(0, length));
	createNewRoad(r, -perp, RT_STREET, RoadAttrib(0, length));
}

void GenRoad::globalConstraintBasic(Road *r, const RoadAttrib &attrib)
{
	float length = r->get2DLength();
	if(length >= attrib.data)
	{
		D3DXVECTOR2 diff = r->get2DVector();

		D3DXVec2Normalize(&diff, &diff);

		float angle = std::acos(diff.x);

		if(diff.y < 0.0f)
		{
			angle = -angle;
		}

		for(unsigned i = 0; i < sizeof(s_minAngle) / sizeof(*s_minAngle); ++i)
		{
			float population = 0.0f;
			const D3DXVECTOR2 &dir = computeBestDir(
				r->getEndCrossing().get2DPosition(),
				s_minAngle[i] + angle,
				s_maxAngle[i] + angle,
				randomFloatInRange(m_streetLength, m_streetRange),
				population);

			if(population >= m_densityTreshold)
			{
				createNewRoad(r, dir, RT_HIGHWAY,
					RoadAttrib(0, randomFloatInRange(m_highwayLength, m_highwayRange)));
			}
		}
	}
	else
	{
		D3DXVECTOR2 dir;

		if(randomBool(m_probHighwayChangeAngle))
		{
			D3DXVECTOR2 diff = r->get2DVector();

			D3DXVec2Normalize(&diff, &diff);

			float angle = std::acos(diff.x);

			if(diff.y < 0.0f)
			{
				angle = -angle;
			}
			
			float population = -1.0f;
			dir = computeBestDir(
				r->getEndCrossing().get2DPosition(),
				s_minAngle[0] + angle,
				s_maxAngle[0] + angle,
				randomFloatInRange(m_streetLength, m_streetRange),
				population);
		}
		else
		{
			dir = r->get2DVector();
			dir /= length;
			dir *= randomFloatInRange(m_streetLength, m_streetRange);
		}

		createNewRoad(r, dir, RT_HIGHWAY, RoadAttrib(0, attrib.data - length));

		if(randomBool(m_probStreetBranch))
		{
			D3DXVECTOR2 perp(-dir.y, dir.x);
			D3DXVec2Normalize(&perp, &perp);
			float length = randomFloatInRange(m_streetLength, m_streetRange);
			float lengthPerp = randomFloatInRange(0.75f, 0.1f) * length;
			perp *= length;

			createNewRoad(r, perp, RT_STREET, RoadAttrib(STREET_DELAY, lengthPerp));
			createNewRoad(r, -perp, RT_STREET, RoadAttrib(STREET_DELAY, lengthPerp));
		}
	}
}

void GenRoad::globalConstraintRaster(Road *r, const RoadAttrib &attrib)
{
	float length = r->get2DLength();
	if(length >= attrib.data)
	{
		float angle = vectorAngle(r->get2DVector());
		const D3DXVECTOR2 &start = r->getEndCrossing().get2DPosition();

		for(unsigned i = 0; i < sizeof(s_rasterAngle) / sizeof(*s_rasterAngle); ++i)
		{
			float length = randomFloatInRange(m_streetLength, m_streetRange);
			D3DXVECTOR2 dir = D3DXVECTOR2(std::cos(s_rasterAngle[i] + angle), std::sin(s_rasterAngle[i] + angle)) * length;
			float population = computeDensityWeight(r->getEndCrossing().get2DPosition(), start + dir);
			if(population >= m_densityTreshold)
			{
				createNewRoad(r, dir, RT_HIGHWAY,
					RoadAttrib(0, randomFloatInRange(m_highwayLength, m_highwayRange)));
			}
		}
	}
	else
	{
		D3DXVECTOR2 dir = r->get2DVector();
		dir /= length;
		dir *= randomFloatInRange(m_streetLength, m_streetRange);

		createNewRoad(r, dir, RT_HIGHWAY, RoadAttrib(0, attrib.data - length));

		if(randomBool(m_probStreetBranch))
		{
			D3DXVECTOR2 perp(-dir.y, dir.x);
			D3DXVec2Normalize(&perp, &perp);
			float length = randomFloatInRange(m_streetLength, m_streetRange);
			float lengthPerp = randomFloatInRange(0.75f, 0.1f) * length;
			perp *= length;

			createNewRoad(r, perp, RT_STREET, RoadAttrib(STREET_DELAY, lengthPerp));
			createNewRoad(r, -perp, RT_STREET, RoadAttrib(STREET_DELAY, lengthPerp));
		}
	}
}

void GenRoad::globalConstraintRadial(Road *r, const RoadAttrib &attrib, const D3DXVECTOR2 &center)
{
	float length = r->get2DLength();
	D3DXVECTOR2 toCenter = r->getEndCrossing().get2DPosition() - center;
	D3DXVec2Normalize(&toCenter, &toCenter);
	D3DXVECTOR2 dir = r->get2DVector() / length;
	float cos = D3DXVec2Dot(&toCenter, &dir);

	if(length >= attrib.data)
	{
		D3DXVECTOR2 perp(-toCenter.y, toCenter.x);
		const D3DXVECTOR2 &start = r->getEndCrossing().get2DPosition();

		if(std::fabs(cos) < 0.7071f)
		{
			const D3DXVECTOR2 dirs[] = { -toCenter, (D3DXVec2Dot(&perp, &dir) < 0.0f) ? -perp : perp };

			for(unsigned i = 0; i < sizeof(dirs) / sizeof(*dirs); ++i)
			{
				length = randomFloatInRange(m_streetLength, m_streetRange);
				float population = computeDensityWeight(start, start + dirs[i] * length);
				if(population >= m_densityTreshold)
				{
					createNewRoad(r, dirs[i] * length, RT_HIGHWAY,
						RoadAttrib(0, randomFloatInRange(m_highwayLength, m_highwayRange)));
				}
			}
		}
		else
		{
			const D3DXVECTOR2 dirs[] = { (cos < 0.0f) ? -toCenter : toCenter, -perp, perp };

			for(unsigned i = 0; i < sizeof(dirs) / sizeof(*dirs); ++i)
			{
				length = randomFloatInRange(m_streetLength, m_streetRange);
				float population = computeDensityWeight(start, start + dirs[i] * length);
				if(population >= m_densityTreshold)
				{
					createNewRoad(r, dirs[i] * length, RT_HIGHWAY,
						RoadAttrib(0, randomFloatInRange(m_highwayLength, m_highwayRange)));
				}
			}
		}
	}
	else
	{
		D3DXVECTOR2 newDir;
		if(std::fabs(cos) < 0.7071f)
		{
			D3DXVECTOR2 perp(-toCenter.y, toCenter.x);
			newDir = (D3DXVec2Dot(&dir, &perp) < 0.0f) ? -perp : perp;			
		}
		else
		{
			newDir = (cos < 0.0f) ? -toCenter : toCenter;
		}

		createNewRoad(r, newDir * randomFloatInRange(m_streetLength, m_streetRange), RT_HIGHWAY, RoadAttrib(0, attrib.data - length));

		if(randomBool(m_probStreetBranch))
		{
			D3DXVECTOR2 perp(-newDir.y, newDir.x);
			float length = randomFloatInRange(m_streetLength, m_streetRange);
			float lengthPerp = randomFloatInRange(0.75f, 0.1f) * length;
			perp *= length;

			createNewRoad(r, perp, RT_STREET, RoadAttrib(STREET_DELAY, lengthPerp));
			createNewRoad(r, -perp, RT_STREET, RoadAttrib(STREET_DELAY, lengthPerp));
		}
	}
}

void GenRoad::createNewRoad(Road *prevRoad, const D3DXVECTOR2 &dir, ERoadType type, const RoadAttrib &attrib)
{
	D3DXVECTOR2 newPoint = prevRoad->getEndCrossing().get2DPosition() + dir;
	Crossing * c = new Crossing(D3DXVECTOR3(newPoint.x, 0.0f, newPoint.y));
	Road *newRoad = new Road(prevRoad->getEndCrossing(), *c, type);
	c->addRoad(*newRoad);
	m_unassignedCrossings.push(std::pair<Road *, RoadAttrib>(newRoad, attrib));
}

float GenRoad::computeDensityWeight(const D3DXVECTOR2 &start, const D3DXVECTOR2 &end)
{
	D3DXVECTOR2 dir = end - start;
	float length = D3DXVec2Length(&dir);
	dir /= length;
	float sum = 0.0f;
	float bias = 0.0f;

	for(float d = SAMPLE_INC; d <= length; d += SAMPLE_INC)
	{
		const D3DXVECTOR2 &p = start + d * dir;
		float w = m_density->getDensityMap().interpolate(p.x, p.y);

		if(w > 0.0f)
		{
			sum += w / d;
		}
		bias += 1.0f / d;
	}

	return sum / bias;
}

D3DXVECTOR2 GenRoad::computeBestDir(const D3DXVECTOR2 &start, float minAngle, float maxAngle, float length, float &population)
{
	D3DXVECTOR2 bestDir;

	for(unsigned i = 0; i < NB_TEST_ROAD; ++i)
	{
		float angle = randomFloat(minAngle, maxAngle);
		D3DXVECTOR2 dir(std::cos(angle), std::sin(angle));
		const D3DXVECTOR2 &p = start + dir * SAMPLE_LENGTH;

		float pop = computeDensityWeight(start, p);

		if(pop > population)
		{
			population = pop;
			bestDir = dir * length;
		}
	}

	return bestDir;
}

void GenRoad::assignCrossing(Crossing *c)
{
	int x = static_cast<int>(c->getPosition().x);
	int y = static_cast<int>(c->getPosition().z);

	m_crossingsGrid[x + y * m_env->getGridSize()].push_back(c);
	m_roads.addCrossing(*c);
}

void GenRoad::assignRoad(Road *r)
{
	float x1 = r->getStartCrossing().getPosition().x;
	float y1 = r->getStartCrossing().getPosition().z;
	float x2 = r->getEndCrossing().getPosition().x;
	float y2 = r->getEndCrossing().getPosition().z;

	int start, end, inc;
	int i, j;
	int *it, *o;
	float a, b;

	if(std::fabs(x2 - x1) > std::fabs(y2 - y1))
	{
		it = &i;
		o = &j;
		start = static_cast<int>(x1);
		end = static_cast<int>(x2);
		a = (y2 - y1) / (x2 - x1);
		b = y1 - x1 * a;
	}
	else
	{
		it = &j;
		o = &i;
		start = static_cast<int>(y1);
		end = static_cast<int>(y2);
		a = (x2 - x1) / (y2 - y1);
		b = x1 - y1 * a;
		y1 = x1;
		y2 = x2;
	}

	if(end >= start)
	{
		inc = 1;
		if(end >= m_env->getGridSize())
			end = m_env->getGridSize() - 1;
		if(start < 0)
			start = 0;
	}
	else
	{
		inc = -1;
		if(start >= m_env->getGridSize())
			start = m_env->getGridSize();
		if(end < 0)
			end = 0;
	}

	int max = std::min(static_cast<int>(std::max(y1, y2)), m_env->getGridSize() - 1);

	for(*it = start; *it != end + inc; *it += inc)
	{
		*o = static_cast<int>(static_cast<float>(*it) * a + b);

		if(*o >= 0 && *o <= max)
		{
			m_roadsGrid[i + j * m_env->getGridSize()].push_back(r);
		}

		int prec = *o;
		*o = static_cast<int>(static_cast<float>(*it + 1) * a + b);

		if(*o != prec && *o >= 0 && *o <= max)
		{
			m_roadsGrid[i + j * m_env->getGridSize()].push_back(r);
		}
	}

	m_roads.addRoad(*r);
}

Crossing * GenRoad::getNearestCrossingInArea(const D3DXVECTOR2 &pos, float radius)
{
	Crossing *nearest = NULL;
	float nearLength = radius * radius;
	int xmin = static_cast<int>(pos.x - radius);
	int xmax = static_cast<int>(pos.x + radius + 1);
	int ymin = static_cast<int>(pos.y - radius);
	int ymax = static_cast<int>(pos.y + radius + 1);

	if(xmin < 0) xmin = 0;
	if(ymin < 0) ymin = 0;

	if(xmax >= m_env->getGridSize()) xmax = m_env->getGridSize() - 1;
	if(ymax >= m_env->getGridSize()) ymax = m_env->getGridSize() - 1;

	for(int j = ymin; j <= ymax; ++j)
	{
		for(int i = xmin; i <= xmax; ++i)
		{
			std::vector<Crossing *> &array = m_crossingsGrid[i + j * m_env->getGridSize()];

			for(std::vector<Crossing *>::iterator it = array.begin(); it != array.end(); ++it)
			{
				const D3DXVECTOR2 &crossPos = (*it)->get2DPosition();
				D3DXVECTOR2 diff = crossPos - pos;

				float length = D3DXVec2LengthSq(&diff);

				if(length < nearLength)
				{
					nearLength = length;
					nearest = *it;
				}
			}
		}
	}

	return nearest;
}

float roadIntersection(const Road &r1, const Road &r2)
{
	const D3DXVECTOR2 &start1 = r1.getStartCrossing().get2DPosition();
	const D3DXVECTOR2 &end1 = r1.getEndCrossing().get2DPosition();
	const D3DXVECTOR2 &start2 = r2.getStartCrossing().get2DPosition();
	const D3DXVECTOR2 &end2 = r2.getEndCrossing().get2DPosition();

	const D3DXVECTOR2 &dir1 = end1 - start1;
	const D3DXVECTOR2 &dir2 = end2 - start2;
	const D3DXVECTOR2 &vec = start1 - start2;

	float d = D3DXVec2CCW(&dir1, &dir2);

	if(std::fabs(d) < 0.00001f)
	{
		float t = D3DXVec2CCW(&dir2, &vec);

		if(std::fabs(t) < 0.00001f)
		{
			float t1 = -D3DXVec2Dot(&vec, &dir1);
			const D3DXVECTOR2 &vec2 = end2 - start1;
			float t2 = D3DXVec2Dot(&vec2, &dir1);

			if(t2 < t1)
			{
				std::swap(t1, t2);
			}
			float l = 1.0f / D3DXVec2Length(&dir1);
			t1 *= l;

			if(t1 > 0.0f && t1 < 1.0f)
			{
				return t1;
			}

			if(t1 < 0.0f && t2 > 0.0f)
			{
				return 0.0f;
			}
		}

		return std::numeric_limits<float>::infinity();
	}

	d = 1.0f / d;

	float t1 = D3DXVec2CCW(&dir2, &vec) * d;
	float t2 = D3DXVec2CCW(&dir1, &vec) * d;

	if(t1 < 0.0f || t1 > 1.0f || t2 < 0.0f || t2 > 1.0f)
	{
		return std::numeric_limits<float>::infinity();
	}

	return t1;
}

Road * GenRoad::getNearestIntersectedRoadInCell(const Road &r, float &t, int i, int j) const
{
	t = std::numeric_limits<float>::infinity();
	Road *out = NULL;

	std::vector<Road *> &array = m_roadsGrid[i + j * m_env->getGridSize()];

	for(std::vector<Road *>::iterator it = array.begin(); it != array.end(); ++it)
	{
		if(&(*it)->getStartCrossing() == &r.getStartCrossing() ||
			&(*it)->getStartCrossing() == &r.getEndCrossing() ||
			&(*it)->getEndCrossing() == &r.getStartCrossing() ||
			&(*it)->getEndCrossing() == &r.getEndCrossing())
		{
			continue;
		}

		float tempT = roadIntersection(r, **it);

		if(tempT < t)
		{
			t = tempT;
			out = *it;
		}
	}

	return out;
}

Road * GenRoad::getNearestIntersectedRoad(const Road &r, float &t) const
{
	float x1 = r.getStartCrossing().getPosition().x;
	float y1 = r.getStartCrossing().getPosition().z;
	float x2 = r.getEndCrossing().getPosition().x;
	float y2 = r.getEndCrossing().getPosition().z;

	int start, end, inc;
	int d1, d2;
	int i, j;
	int *it, *o;
	float a, b;

	if(std::fabs(x2 - x1) > std::fabs(y2 - y1))
	{
		it = &i;
		o = &j;
		start = static_cast<int>(x1);
		end = static_cast<int>(x2);
		a = (y2 - y1) / (x2 - x1);
		b = y1 - x1 * a;
	}
	else
	{
		it = &j;
		o = &i;
		start = static_cast<int>(y1);
		end = static_cast<int>(y2);
		a = (x2 - x1) / (y2 - y1);
		b = x1 - y1 * a;
		y1 = x1;
		y2 = x2;
	}

	if(end >= start)
	{
		inc = 1;
		d1 = 0;
		d2 = 1;
		if(end >= m_env->getGridSize())
			end = m_env->getGridSize() - 1;
		if(start < 0)
			start = 0;
	}
	else
	{
		inc = -1;
		d1 = 1;
		d2 = 0;
		if(start >= m_env->getGridSize())
			start = m_env->getGridSize();
		if(end < 0)
			end = 0;
	}

	int max = std::min(static_cast<int>(std::max(y1, y2)), m_env->getGridSize() - 1);

	for(*it = start; *it != end + inc; *it += inc)
	{
		*o = static_cast<int>(static_cast<float>(*it + d1) * a + b);

		if(*o >= 0 && *o <= max)
		{
			Road *result = getNearestIntersectedRoadInCell(r, t, i, j);

			if(result)
			{
				return result;
			}
		}

		int prec = *o;
		*o = static_cast<int>(static_cast<float>(*it + d2) * a + b);

		if(*o != prec && *o >= 0 && *o <= max)
		{
			Road *result = getNearestIntersectedRoadInCell(r, t, i, j);

			if(result)
			{
				return result;
			}
		}
	}

	return NULL;
}

bool GenRoad::isValidLocation(const D3DXVECTOR2 &pos) const
{
	if(!isInRect(
		D3DXVECTOR2(0.0f, 0.0f),
		D3DXVECTOR2(
			static_cast<float>(m_env->getGridSize() - 1),
			static_cast<float>(m_env->getGridSize() - 1)),
		pos))
	{
		return false;
	}

	float height = m_env->getHeightMap().interpolate(pos.x, pos.y);

	if(height < m_env->getWaterLevel() + 0.05f)
	{
		return false;
	}

	return true;
}

bool GenRoad::canDivideRoad(const Road &r, const Crossing &c) const
{
	const D3DXVECTOR2 &crossingPos = c.get2DPosition();
	const D3DXVECTOR2 &start = r.getStartCrossing().get2DPosition();
	const D3DXVECTOR2 &end = r.getEndCrossing().get2DPosition();

	D3DXVECTOR2 vector = start - crossingPos;
	if(D3DXVec2LengthSq(&vector) < MIN_ROAD_LENGTH * MIN_ROAD_LENGTH)
	{
		return false;
	}

	vector = end - crossingPos;
	if(D3DXVec2LengthSq(&vector) < MIN_ROAD_LENGTH * MIN_ROAD_LENGTH)
	{
		return false;
	}
	return true;
}

ELocalConstraintResult GenRoad::tryMergeCrossing(Road *r)
{
	Crossing *nearest = getNearestCrossingInArea(r->getEndCrossing().get2DPosition(), m_mergeDistance * r->get2DLength());

	if(nearest)
	{
		Crossing &endCrossing = r->getEndCrossing();

		if(nearest == &r->getStartCrossing() || r->getStartCrossing().isNeighbor(nearest))
		{
			return LCR_FAILED;
		}

		r->setCrossing(endCrossing, *nearest);
		
		float t = std::numeric_limits<float>::infinity();
		if(getNearestIntersectedRoad(*r, t) == NULL)
		{
			return LCR_EXISTING_CROSSING;
		}

		r->setCrossing(*nearest, endCrossing);
	}

	return LCR_NEW_CROSSING;
}

ELocalConstraintResult GenRoad::tryIntersectingBackward(Road *r, Road *&intersectedRoad)
{
	float t = std::numeric_limits<float>::infinity();
	intersectedRoad = getNearestIntersectedRoad(*r, t);

	if(intersectedRoad)
	{
		if(r->getStartCrossing().isNeighbor(intersectedRoad))
		{
			return LCR_FAILED;
		}

		D3DXVECTOR2 crossingPos = r->getStartCrossing().get2DPosition() + t * r->get2DVector();
		r->getEndCrossing().setPosition(D3DXVECTOR3(crossingPos.x, 0.0f, crossingPos.y));

		return LCR_EXISTING_CROSSING;
	}

	return LCR_NEW_CROSSING;
}

ELocalConstraintResult GenRoad::tryIntersectingForward(Road *r, Road *&intersectedRoad)
{
	D3DXVECTOR2 dir = r->get2DVector();

	dir *= 1.0f + m_mergeDistance;
	D3DXVECTOR3 lastPos = r->getEndCrossing().getPosition();
	r->getEndCrossing().setPosition(r->getStartCrossing().getPosition() + D3DXVECTOR3(dir.x, 0.0f, dir.y));

	ELocalConstraintResult result = tryIntersectingBackward(r, intersectedRoad);

	if(result == LCR_NEW_CROSSING)
	{
		r->getEndCrossing().setPosition(lastPos);
	}

	return result;
}

void GenRoad::proceed()
{
	Road *r = m_unassignedCrossings.front().first;
	RoadAttrib attrib = m_unassignedCrossings.front().second;
	int delay = attrib.delay;
	m_unassignedCrossings.pop();

	if(delay > 0)
	{
		--attrib.delay;
		m_unassignedCrossings.push(std::pair<Road *, RoadAttrib>(r, attrib));
	}
	else if(delay == 0)
	{
		ELocalConstraintResult result = localConstraint(r);

		switch(result)
		{
		case LCR_NEW_CROSSING:
			assignRoad(r);
			assignCrossing(&r->getEndCrossing());
			r->getStartCrossing().addRoad(*r);

			globalConstraint(r, attrib);
			break;

		case LCR_EXISTING_CROSSING:
			assignRoad(r);
			r->getStartCrossing().addRoad(*r);
			break;

		default:
		case LCR_FAILED:
			delete &r->getEndCrossing();
			delete r;
			break;
		}
	}
	else
	{
		delete &r->getEndCrossing();
		delete r;
	}
}
