#ifndef H_GENROAD
#define H_GENROAD

#include "Generator.h"
#include <d3dx9.h>
#include <vector>
#include <queue>
#include <utility>

class Environment;
class Density;
class Road;
class Crossing;
class RoadNetwork;
enum ERoadType;

enum ELocalConstraintResult
{
	LCR_NEW_CROSSING,
	LCR_EXISTING_CROSSING,
	LCR_FAILED
};

class GenRoad: public Generator
{
public:
	GenRoad(RoadNetwork &roads, const Environment &env, const Density &density);
	~GenRoad();

	virtual bool isFinished() const;
	virtual void commitChange();

	virtual void attachControl(Control *control);

protected:
	struct RoadAttrib
	{
		int delay;
		float data;

		RoadAttrib(int d, float da): delay(d), data(da) {};
	};

	virtual void init();
	virtual void finish();
	virtual void proceed();
	virtual void createParameters(Control *control);

	void initBasicCenter(Crossing *center);
	void initRasterCenter(Crossing *center);

	ELocalConstraintResult localConstraint(Road *r);
	bool adjustValidLocation(Road *r) const;

	void globalConstraint(Road *r, const RoadAttrib &attrib);
	void globalConstraintStreet(Road *r, const RoadAttrib &attrib);
	void globalConstraintBasic(Road *r, const RoadAttrib &attrib);
	void globalConstraintRaster(Road *r, const RoadAttrib &attrib);
	void globalConstraintRadial(Road *r, const RoadAttrib &attrib, const D3DXVECTOR2 &center);

	void createNewRoad(Road *prevRoad, const D3DXVECTOR2 &dir, ERoadType type, const RoadAttrib &attrib);
	float computeDensityWeight(const D3DXVECTOR2 &start, const D3DXVECTOR2 &end);
	D3DXVECTOR2 computeBestDir(const D3DXVECTOR2 &start, float minAngle, float maxAngle, float length, float &population);
	void assignCrossing(Crossing *c);
	void assignRoad(Road *r);
	Crossing * getNearestCrossingInArea(const D3DXVECTOR2 &pos, float radius);
	Road * getNearestIntersectedRoadInCell(const Road &r, float &t, int i, int j) const;
	Road * getNearestIntersectedRoad(const Road &r, float &t) const;
	bool isValidLocation(const D3DXVECTOR2 &pos) const;
	bool canDivideRoad(const Road &r, const Crossing &c) const;
	ELocalConstraintResult tryMergeCrossing(Road *r);
	ELocalConstraintResult tryIntersectingBackward(Road *r, Road *&intersectedRoad);
	ELocalConstraintResult tryIntersectingForward(Road *r, Road *&intersectedRoad);

private:
	GenRoad(const GenRoad &);
	GenRoad & operator =(const GenRoad &);

	const Environment		 *m_env;
	const Density			 *m_density;

	std::queue<std::pair<Road *, RoadAttrib> >	 m_unassignedCrossings;
	std::vector<Crossing *>						*m_crossingsGrid;
	std::vector<Road *>							*m_roadsGrid;
	RoadNetwork									&m_roads;

	float	m_streetLength;
	float	m_streetRange;
	float	m_highwayLength;
	float	m_highwayRange;
	float	m_probHighwayChangeAngle;
	float	m_probStreetChangeLength;
	float	m_probStreetBranch;
	float	m_mergeDistance;
	float	m_densityTreshold;
	float	m_maxBridgeLength;
};

#endif