#include "Geometry.h"
#include "MediaManager.h"
#include "Renderer.h"
#include "MemoryLog.h"

static DeclarationElement elems[] =
{
	{ 0, D3DDECLUSAGE_POSITION, D3DDECLTYPE_FLOAT3 },
	{ 0, D3DDECLUSAGE_NORMAL, D3DDECLTYPE_FLOAT3 }
};

Geometry::Layer::Layer():
m_vb(NULL),
m_ib(NULL)
{
}

Geometry::Layer::~Layer()
{
	delete m_vb;
	delete m_ib;
}

Geometry::Geometry(MediaManager &media):
m_media(&media)
{
	m_shader = media.getObject<Shader>("Building.fx");
	m_declaration = media.getRenderer().createDeclaration(elems);
}

Geometry::~Geometry()
{
	delete m_declaration;
}

unsigned short Geometry::addVertices(const std::vector<VertexStruct> &vertices, unsigned layer)
{
	unsigned start = m_layers[layer].m_vertices.size() + m_layers[layer].m_newVertices.size();
	m_layers[layer].m_newVertices.insert(m_layers[layer].m_newVertices.end(),
		vertices.begin(), vertices.end());
	return static_cast<unsigned short>(start);
}

void Geometry::addIndices(const std::vector<unsigned short> &indices, unsigned layer)
{
	m_layers[layer].m_newIndices.insert(m_layers[layer].m_newIndices.end(),
		indices.begin(), indices.end());
}

void Geometry::commitChange()
{
	recomputeBuffer();
}

static const D3DXVECTOR3 layerColor[GEO_NB_LAYERS] =
{
	D3DXVECTOR3(0.5f, 0.5f, 0.5f),
	D3DXVECTOR3(0.5f, 0.5f, 0.5f),
	D3DXVECTOR3(0.4f, 0.4f, 0.4f)
};

void Geometry::render(Renderer &renderer) const
{
	renderer.setShader(m_shader);
	m_shader->setValue("World", renderer.getMatrix());
	renderer.setDeclaration(m_declaration);
	for(unsigned layer = 0; layer < GEO_NB_LAYERS; ++layer)
	{
		if(m_layers[layer].m_ib)
		{
			m_shader->setValue("Color", &layerColor[layer]);
			renderer.setVertexBuffer<VertexStruct>(0, m_layers[layer].m_vb, 0);
			renderer.drawIndexedPrimitives<unsigned short>(
				D3DPT_TRIANGLELIST, m_layers[layer].m_ib, 0, m_layers[layer].m_ib->getSize() / 3);
		}
	}
}

void Geometry::reset()
{
	for(unsigned i = 0; i < GEO_NB_LAYERS; ++i)
	{
		m_layers[i].m_vertices.clear();
		m_layers[i].m_indices.clear();
		m_layers[i].m_newVertices.clear();
		m_layers[i].m_newIndices.clear();
	}
	recomputeBuffer();
}

void Geometry::recomputeBuffer(bool forceVerticesUpdate)
{
	for(unsigned layer = 0; layer < GEO_NB_LAYERS; ++layer)
	{
		if(m_layers[layer].m_vertices.empty() && m_layers[layer].m_newVertices.empty())
		{
			delete m_layers[layer].m_vb;
			delete m_layers[layer].m_ib;
			m_layers[layer].m_vb = NULL;
			m_layers[layer].m_ib = NULL;
		}
		else
		{
			if(forceVerticesUpdate || !m_layers[layer].m_newVertices.empty())
			{
				delete m_layers[layer].m_vb;
				m_layers[layer].m_vertices.insert(m_layers[layer].m_vertices.end(),
					m_layers[layer].m_newVertices.begin(), m_layers[layer].m_newVertices.end());
				m_layers[layer].m_newVertices.clear();

				m_layers[layer].m_vb = m_media->getRenderer().createVertexBuffer<VertexStruct>(
					m_layers[layer].m_vertices.size(), &m_layers[layer].m_vertices[0]);
			}
			if(!m_layers[layer].m_newIndices.empty())
			{
				delete m_layers[layer].m_ib;
				m_layers[layer].m_indices.insert(m_layers[layer].m_indices.end(),
					m_layers[layer].m_newIndices.begin(), m_layers[layer].m_newIndices.end());
				m_layers[layer].m_newIndices.clear();

				m_layers[layer].m_ib = m_media->getRenderer().createIndexBuffer<unsigned short>(
					m_layers[layer].m_indices.size(), &m_layers[layer].m_indices[0]);
			}
		}
	}
}

void Geometry::computeNormals()
{
	for(unsigned layer = 0; layer < GEO_NB_LAYERS; ++layer)
	{
		for(unsigned i = 0; i < m_layers[layer].m_vertices.size(); ++i)
		{
			m_layers[layer].m_vertices[i].normal = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
		}

		for(unsigned i = 0; i < m_layers[layer].m_indices.size(); i += 3)
		{
			D3DXVECTOR3 p1 = m_layers[layer].m_vertices[m_layers[layer].m_indices[i + 0]].position;
			D3DXVECTOR3 p2 = m_layers[layer].m_vertices[m_layers[layer].m_indices[i + 1]].position;
			D3DXVECTOR3 p3 = m_layers[layer].m_vertices[m_layers[layer].m_indices[i + 2]].position;

			p2 = p2 - p1;
			p3 = p3 - p1;
			D3DXVec3Cross(&p1, &p2, &p3);

			if(D3DXVec3LengthSq(&p1) >= 0.00000001f)
			{
				D3DXVec3Normalize(&p1, &p1);

				for(unsigned j = 0; j < 3; ++j)
				{
					m_layers[layer].m_vertices[m_layers[layer].m_indices[i + j]].normal += p1;
				}
			}
		}

		for(unsigned i = 0; i < m_layers[layer].m_vertices.size(); ++i)
		{
			D3DXVec3Normalize(&m_layers[layer].m_vertices[i].normal, &m_layers[layer].m_vertices[i].normal);
		}
	}

	recomputeBuffer(true);
}