#include "Graph.h"

#include "Way.h"
#include "Junction.h"
#include "HashMap.h"
#include "RoadNetwork.h"
#include "Road.h"
#include "Crossing.h"
#include <cmath>
#include "Debug.h"
#include "Intersection.h"
#include "IntersectionTrafficLight.h"
#include "IntersectionPriority.h"
#include "Renderer.h"
#include "MediaManager.h"
#include "MemoryLog.h"

#define NB_TRAFFICLIGHT_PER_BATCH	125

DeclarationElement elemsTrafficLight[] =
{
	{ 0, D3DDECLUSAGE_POSITION, D3DDECLTYPE_FLOAT3 },
	{ 0, D3DDECLUSAGE_NORMAL, D3DDECLTYPE_FLOAT3 },
	{ 0, D3DDECLUSAGE_BLENDINDICES, D3DDECLTYPE_SHORT4 }
};

Graph::Graph(RoadNetwork & roadNetwork, MediaManager & media):
m_media(&media),
m_vb(NULL),
m_ib(NULL)
{
	HashMap<const Crossing*, Intersection*> connectionMap;

	// on convertit RoadNetwork en graphe
	// on commence par crer les junctions
	for (unsigned i = 0; i < roadNetwork.getCrossings().size(); i++)
	{
		const Crossing *cross = roadNetwork.getCrossings()[i];
		unsigned begin = m_junctions.size();

		const std::vector<const Road*>& sortedRoad = cross->getRoads();
		ERoadType type = RT_STREET;

		// on regarde si une avenue se trouve dans l'intersection
		for (unsigned j = 0; j < sortedRoad.size(); j++)
		{
			if (sortedRoad[j]->getType() == RT_HIGHWAY)
			{
				type = RT_HIGHWAY;
				break;
			}
		}

		Intersection *intersection;
		// si le crossing ne contient pas plus de deux routes, il n'y a pas croisement de route, donc pas de rgles
		if (sortedRoad.size() < 3)
		{
			intersection = new Intersection();
		}
		// sinon, on cr une intersection avec des rgles d'arrt pour viter les collisions
		else if (type == RT_STREET)
		{
			intersection = new IntersectionPriority();
		}
		else
		{
			TrafficLight trafficLight;
			trafficLight.intersection = new IntersectionTrafficLight();
			for(unsigned i = 0; i < sortedRoad.size(); ++i)	
				m_trafficLights.push_back(trafficLight);
			intersection = trafficLight.intersection;
		}

		connectionMap[cross] = intersection;

		// pour chaque Crossing, on cr 3 junctions : 
		for (unsigned j = 0; j < sortedRoad.size()*3; j++)
		{
			Junction *junc = new Junction(cross->getJunctionPosition(j), *intersection);
			m_junctions.push_back(junc);
		}

		// on ajoute l'intersection correspondante  la liste
		m_intersections.push_back(intersection);

		// on cr les ways internes
		for (unsigned j = 0; j < sortedRoad.size(); j++)
		{
			// way partant d'un point d'attente
			Way* way = new Way(m_junctions[begin + 3 * j], m_junctions[begin + 3 * j + 1], sortedRoad[j]->getType(), RZT_UNDEFINED);
			m_ways.push_back(way);
			// way partant d'un point central
			for (unsigned k = 0; k < sortedRoad.size() - 1; k++)
			{
				unsigned ind = begin + ((3 * j + 1 + 3 * k + 1) % (sortedRoad.size() * 3));
				way = new Way(m_junctions[begin + 3 * j + 1], m_junctions[ind], sortedRoad[j]->getType(), RZT_UNDEFINED);
				m_ways.push_back(way);
			}
		}
	}

	// la deuxime tape est de crer les routes entres les carrefours nouvellement gnrs
	for (unsigned i = 0; i < roadNetwork.getRoads().size(); i++)
	{
		const Road* r = roadNetwork.getRoads()[i];
		const Crossing	&c1 = r->getStartCrossing(),
						&c2 = r->getEndCrossing();

		unsigned index1 = c1.getRoadIndex(r);
		unsigned index2 = c2.getRoadIndex(r);

		unsigned indexBegin1 = index1 * 3;
		unsigned indexBegin2 = (index1 == 0) ? (3 * c1.getRoads().size() - 1) : (3 * index1 - 1);
		unsigned indexEnd1 = index2 * 3;
		unsigned indexEnd2 = (index2 == 0) ? (3 * c2.getRoads().size() - 1) : (3 * index2 - 1);

		m_ways.push_back(new Way(
			connectionMap[&c1]->getJunctions()[indexBegin2],
			connectionMap[&c2]->getJunctions()[indexEnd1], r->getType(), r->getZoneType()));
		m_ways.push_back(new Way(
			connectionMap[&c2]->getJunctions()[indexEnd2],
			connectionMap[&c1]->getJunctions()[indexBegin1], r->getType(), r->getZoneType()));
	}

	// enfin, on initialise les intersections
	for (unsigned i = 0; i < m_intersections.size(); i++)
	{
		m_intersections[i]->init();
	}

	IntersectionTrafficLight *lastIntersection = NULL;
	unsigned index = 0;
	for(unsigned i = 0; i < m_trafficLights.size(); ++i)
	{
		if(m_trafficLights[i].intersection != lastIntersection)
		{
			index = 0;
			lastIntersection = m_trafficLights[i].intersection;
		}
		m_trafficLights[i].junction = lastIntersection->getJunctions()[index];
		index += 3;
	}

	// On cr la gomtrie
	m_shader = m_media->getObject<Shader>("TrafficLight.fx");
	m_declaration = m_media->getRenderer().createDeclaration(elemsTrafficLight);

	m_vb = m_media->getRenderer().createVertexBuffer<Vertex>(NB_TRAFFICLIGHT_PER_BATCH * 4);
	m_ib = m_media->getRenderer().createIndexBuffer<unsigned short>(NB_TRAFFICLIGHT_PER_BATCH * 6);

	Vertex *vertices = m_vb->lock();
	unsigned short *indices = m_ib->lock();

	for(unsigned short i = 0; i < NB_TRAFFICLIGHT_PER_BATCH; ++i)
	{
		vertices[0].position = D3DXVECTOR3(-0.02f, 0.01f, 0.02f);
		vertices[1].position = D3DXVECTOR3( 0.02f, 0.01f, 0.02f);
		vertices[2].position = D3DXVECTOR3( 0.02f, 0.01f,-0.02f);
		vertices[3].position = D3DXVECTOR3(-0.02f, 0.01f,-0.02f);

		vertices[0].index[0] = static_cast<char>(i);
		vertices[1].index[0] = static_cast<char>(i);
		vertices[2].index[0] = static_cast<char>(i);
		vertices[3].index[0] = static_cast<char>(i);

		indices[0] = 4 * i + 3;
		indices[1] = 4 * i;
		indices[2] = 4 * i + 2;
		indices[3] = 4 * i + 2;
		indices[4] = 4 * i;
		indices[5] = 4 * i + 1;

		vertices += 4;
		indices += 6;
	}

	m_vb->unlock();
	m_ib->unlock();
}

Graph::~Graph(void)
{
	for (unsigned i = 0; i < m_junctions.size(); i++)
	{
		delete m_junctions[i];
	}
	for (unsigned i = 0; i < m_ways.size(); i++)
	{
		delete m_ways[i];
	}
	for (unsigned i = 0; i < m_intersections.size(); i++)
	{
		delete m_intersections[i];
	}
	delete m_vb;
	delete m_ib;
	delete m_declaration;
}

const std::vector<Way*>& Graph::getWays() const
{
	return m_ways;
}

float Graph::computeRoadAngle(const Road *r, const Crossing *current)
{
	D3DXVECTOR2 vRoad = r->getEndCrossing(*current).get2DPosition() - current->get2DPosition();
	D3DXVec2Normalize(&vRoad, &vRoad);

	float dot = vRoad.x;
	if (dot > 1.0f)
	{
		dot = 1.0f;
	}
	else if (dot < -1.0f)
	{
		dot = -1.0f;
	}

	if (vRoad.y < 0.0f)
	{
		return - std::acos(dot);
	}
	return std::acos(dot);
}

void Graph::computeOrderedRoadVector(const Crossing *c, std::vector<const Road*>& out)
{
	std::vector<RoadAngle> roadsAngles;
	roadsAngles.clear();
	for (unsigned i = 0; i < c->getRoads().size(); i++)
	{
		// on calcule l'angle de la route actuelle...
		RoadAngle ra(c->getRoads()[i], computeRoadAngle(c->getRoads()[i], c));
		// ... puis on la place  sa place dans la liste
		unsigned k = 0;
		while (k < roadsAngles.size())
		{
			if (roadsAngles[k].angle > ra.angle)
			{
				break;
			}
			k++;
		}
		roadsAngles.insert(roadsAngles.begin() + k, ra);
	}
	// une fois finie la liste, on la recopie en enlevant les angles
	for (unsigned i = 0; i < roadsAngles.size(); i++)
	{
		out.push_back(roadsAngles[i].road);
	}
}

Junction * Graph::getPreviousJunction(Junction *j)
{
	unsigned ind = (j->getIntersectionIndice() == 0) ? 
							j->getIntersection().getJunctions().size() - 1 
						:	j->getIntersectionIndice() - 1;
	return j->getIntersection().getJunctions()[ind];
}

D3DXVECTOR3 Graph::getJunctionPosition(const Road *r1, const Road *r2) const
{
	// on trouve le crossing commun
	const Crossing *common;
	if (&r1->getStartCrossing() == &r2->getStartCrossing() || &r1->getStartCrossing() == &r2->getEndCrossing())
	{
		common = &r1->getStartCrossing();
	}
	else
	{
		common = &r1->getEndCrossing();
	}

	// on calcule les vecteurs des routes
	D3DXVECTOR3 vR1 = r1->get3DVector(common),
				vR2 = r2->get3DVector(common);
	// on dcale les vecteurs pour les placer sur la voie correspondante
	D3DXVec3Normalize(&vR1, &vR1);
	D3DXVec3Normalize(&vR2, &vR2);
	D3DXVECTOR2 vp1(vR1.z, -vR1.x);
	D3DXVECTOR2 vR2_2D(vR2.x, vR2.z);
	D3DXVec2Normalize(&vp1, &vp1);
	D3DXVec2Normalize(&vR2_2D, &vR2_2D);
	float cos = D3DXVec2Dot(&vR2_2D, &vp1);
	if (std::fabs(cos) < 0.001f)
	{
		// les deux vecteurs sont collinaires
		return common->getPosition() - D3DXVECTOR3(vp1.x, 0.0f, vp1.y) * (r1->getWidth() + r2->getWidth()) / 8.0f;
	}

	return common->getPosition() - ((r1->getWidth() / 4.0f) * vR2 + (r2->getWidth() / 4.0f) * vR1) / cos;
}

const std::vector<Junction*>& Graph::getJunctions()const 
{	
	return m_junctions;
}

const std::vector<Intersection*>& Graph::getIntersections() const
{
	return m_intersections;
}

void Graph::render(Renderer &renderer)const
{
	D3DXVECTOR4 positions[NB_TRAFFICLIGHT_PER_BATCH];
	D3DXVECTOR4 colors[NB_TRAFFICLIGHT_PER_BATCH];

	renderer.setShader(m_shader);
	renderer.setDeclaration(m_declaration);
	renderer.setVertexBuffer<Vertex>(0, m_vb, 0);
	
	for (unsigned i = 0; i < m_trafficLights.size(); i += NB_TRAFFICLIGHT_PER_BATCH)
	{
		unsigned count = min(m_trafficLights.size() - i, NB_TRAFFICLIGHT_PER_BATCH);

		for(unsigned j = 0; j < count; ++j)
		{
			const D3DXVECTOR3 &p = m_trafficLights[i + j].junction->getPosition();
			D3DXVECTOR3 dir = m_trafficLights[i + j].junction->getWaysIn()[0]->getVector();
			dir.y = 0.0f;
			D3DXVec3Normalize(&dir, &dir);
			const D3DXVECTOR3 &c = m_trafficLights[i + j].intersection->getColor(*m_trafficLights[i + j].junction);

			positions[j] = D3DXVECTOR4(p.x, p.y, p.z, dir.x);
			colors[j] = D3DXVECTOR4(c.x, c.y, c.z, dir.z);
		}
		m_shader->setValue("Positions", positions, count);
		m_shader->setValue("Colors", colors, count);
		renderer.drawIndexedPrimitives<unsigned short>(D3DPT_TRIANGLELIST, m_ib, 0, 2 * count);
	}
}