#include "HRTF.h"
#include "Debug.h"
#include "FFT.h"
#include "Sound.h"
#include <sstream>
#include <iomanip>

HRTF g_HRTF;

static const int nbDataPerElev[HRTF_NB_ELEV] =
{ 56, 60, 72, 72, 72, 72, 72, 60, 56, 45, 36, 24, 12, 1 };

static inline int indexToElev(int index)
{
	return (index * HRTF_INC_ELEV) + HRTF_MIN_ELEV;
}

static inline int indexToAzimuth(int index, int indexElev)
{
	return static_cast<int>(((index * 360.0f) / nbDataPerElev[indexElev]) + 0.5f);
}

static inline unsigned nbFileInElev(int index)
{
	return (nbDataPerElev[index] / 2) + 1;
}

HRTF::HRTF()
{
	for(unsigned i = 0; i < HRTF_NB_ELEV; ++i)
	{
		m_data[i] = NULL;
	}
}

HRTF::~HRTF()
{
	for(unsigned i = 0; i < HRTF_NB_ELEV; ++i)
	{
		delete[] m_data[i];
	}
}

void HRTF::load(const std::string &path)
{
	for(unsigned i = 0; i < HRTF_NB_ELEV; ++i)
	{
		unsigned nbFile = nbFileInElev(i);
		m_data[i] = new std::complex<float>[nbFile * FFT_NB_POINTS * 2];

		for(unsigned j = 0; j < nbFile; ++j)
		{
			int elev = indexToElev(i);
			int azimuth = indexToAzimuth(j, i);
			
			loadFile(path + getFileName(elev, azimuth), i, j);
		}
	}
}

const std::complex<float> * HRTF::get(const HRTFIndex &index, int channel) const
{
	if(index.flipChannel) channel = 1 - channel;
	return &m_data[index.indexElev][(index.indexAzimuth * 2 + channel) * FFT_NB_POINTS];
}

HRTFIndex HRTF::getIndex(float elev, float azimuth) const
{
	HRTFIndex out;
	out.indexElev = static_cast<int>((elev - HRTF_MIN_ELEV) / HRTF_INC_ELEV + 0.5f);
	if(out.indexElev < 0)
	{
		out.indexElev = 0;
	}
	else if(out.indexElev >= HRTF_NB_ELEV)
	{
		out.indexElev = HRTF_NB_ELEV - 1;
	}
	azimuth = std::fmod(azimuth, 360.0f);
	if(azimuth < 0.0f)
	{
		azimuth += 360.0f;
	}
	if(azimuth > 180.0f)
	{
		azimuth = 360.0f - azimuth;
		out.flipChannel = true;
	}
	else
	{
		out.flipChannel = false;
	}
	out.indexAzimuth = static_cast<int>(azimuth * nbDataPerElev[out.indexElev] / 360.0f + 0.5f);

	return out;
}

void HRTF::loadFile(const std::string &filename, int elevIndex, int azimuthIndex)
{
	Sound file(filename);
	float leftChannel[FFT_NB_POINTS];
	float rightChannel[FFT_NB_POINTS];
	const short *data = reinterpret_cast<const short *>(file.getData());

	for(unsigned i = 0; i < FFT_NB_POINTS - HRTF_SAMPLE_LENGTH; ++i)
	{
		leftChannel[i] = 0.0f;
		rightChannel[i] = 0.0f;
	}

	for(unsigned i = 0; i < HRTF_SAMPLE_LENGTH; ++i)
	{
		leftChannel[FFT_NB_POINTS - HRTF_SAMPLE_LENGTH + i] = static_cast<float>(data[2 * i]) / 32768.0f;
		rightChannel[FFT_NB_POINTS - HRTF_SAMPLE_LENGTH + i] = static_cast<float>(data[2 * i + 1]) / 32768.0f;
	}

	fastFourierTransform(leftChannel, &m_data[elevIndex][azimuthIndex * 2 * FFT_NB_POINTS]);
	fastFourierTransform(rightChannel, &m_data[elevIndex][(azimuthIndex * 2 + 1) * FFT_NB_POINTS]);
}

std::string HRTF::getFileName(int elev, int azimuth) const
{
	std::ostringstream buffer;
	buffer << "elev" << elev << "/H" << elev << "e" << std::setfill('0') << std::setw(3) << azimuth << "a.wav";
	return buffer.str();
}