/*****************************************************************************/
/* Name:				HashMap.inl
/* Purpose:				Dfinition de la classe template HashMap
/* Author:				Bonnet Benjamin
/* Created:				03/11/2008
/* Last Modification:	06/11/2008
/*****************************************************************************/

#include <string>

/*****************************************************************************/
/* Fonctions diverses
/*****************************************************************************/

/*****************************************************************************/
/* hashString
/* Retourne un index de tableau  partir du chaine de caractres
/*	S:		Taille du tableau (puissance de 2)
/*	str:	Chaine de caractres  hasher
/*	size:	Taille de la chane de caractres
/*****************************************************************************/

// Version ANSI
template <int S>
unsigned hashString(const char *str, unsigned size)
{
	unsigned hash = 5381;

	while(size--)
	{
		hash = ((hash << 5) + hash) + *str;
		str++;
	}

	return hash & (S - 1); // hash modulo S
}

// Version UNICODE
template <int S>
unsigned hashString(const wchar_t *str, unsigned size)
{
	unsigned hash = 5381;

	while(size--)
	{
		hash = ((hash << 5) + hash) + *str;
		str++;
	}

	return hash & (S - 1); // hash modulo S
}

/*****************************************************************************/
/* Classes locales
/*****************************************************************************/

/*****************************************************************************/
/* HashObjectFunc:
/* Classe utilitaire template permettant de hasher l'objet diffremment selon
/* son type
/*	K:		Type de l'objet
/*	size:	Taille du tableau
/*****************************************************************************/

// Classe de base pour tous les objets non particulier
template <class K, int size>
class HashObjectFunc
{
public:
	static unsigned hash(const K &object)
	{
		return hashString<size>(reinterpret_cast<const char *>(&object), sizeof(K));
	}
};

// Spcialisation pour les chanes de caractres
template <int size>
class HashObjectFunc<std::string, size>
{
public:
	static unsigned hash(const std::string &str)
	{
		return hashString<size>(str.c_str(), str.size());
	}
};


/*****************************************************************************/
/* Dfinition de HashMap
/*****************************************************************************/

/*****************************************************************************/
/* Constructeurs et Destructeurs
/*****************************************************************************/

template <class K, class T, int size>
HashMap<K, T, size>::HashMap()
{
	for(unsigned i = 0; i < size; ++i)
	{
		m_data[i] = NULL;
	}
}

template <class K, class T, int size>
HashMap<K, T, size>::~HashMap()
{
	clear();
}

/*****************************************************************************/
/* Fonctions publiques
/*****************************************************************************/

template <class K, class T, int size>
bool HashMap<K, T, size>::isExist(const K &key) const
{
	// On rcupre l'index du tableau o devrait se trouver l'objet
	Item *ptr = m_data[HashObjectFunc<K, size>::hash(key)];

	// On regarde si un des lments de la liste chane est celui recherch
	while(ptr)
	{
		if(ptr->key == key)
		{
			return true;
		}
		ptr = ptr->next;
	}

	return false;
}

template <class K, class T, int size>
void HashMap<K, T, size>::remove(const K &key)
{
	// On rcupre l'index du tableau o devrait se trouver l'objet
	unsigned index = HashObjectFunc<K, size>::hash(key);

	if(m_data[index])
	{
		Item *ptr = m_data[index];

		if(ptr->key == key)
		{
			// L'lment a supprimer est le premier lment de la liste
			m_data[index] = ptr->next;
			delete ptr;
			return;
		}
		else
		{
			// On recherche l'lment dans la liste chane
			while(ptr->next)
			{
				if(ptr->next->key == key)
				{
					Item *temp = ptr->next;
					ptr->next = temp->next;
					delete temp;
					return;
				}
				ptr = ptr->next;
			}
		}
	}
}

template <class K, class T, int size>
void HashMap<K, T, size>::clear()
{
	for(unsigned i = 0; i < size; ++i)
	{
		Item *ptr = m_data[i];

		// On supprime tous les lments de la liste chane
		while(ptr)
		{
			Item *next = ptr->next;
			delete ptr;
			ptr = next;
		}

		m_data[i] = NULL;
	}
}

template <class K, class T, int size>
T & HashMap<K, T, size>::operator[] (const K &key)
{
	// On rcupre l'index o devrait se trouver l'objet
	unsigned index = HashObjectFunc<K, size>::hash(key);

	Item *ptr = m_data[index];

	while(ptr)
	{
		if(ptr->key == key)
		{
			// On a trouv l'objet
			return ptr->item;
		}
		ptr = ptr->next;
	}

	// Il n'existe encore aucun objet pour cette cl, on le cr
	m_data[index] = new Item(key, m_data[index]);
	return m_data[index]->item;
}

template <class K, class T, int size>
T * HashMap<K, T, size>::get (const K &key) const
{
	// On rcupre l'index o devrait se trouver l'objet
	unsigned index = HashObjectFunc<K, size>::hash(key);

	Item *ptr = m_data[index];

	while(ptr)
	{
		if(ptr->key == key)
		{
			// On a trouv l'objet
			return &ptr->item;
		}
		ptr = ptr->next;
	}

	// Il n'existe encore aucun objet pour cette cl, on retourne NULL
	return NULL;
}

template <class K, class T, int size>
void HashMap<K, T, size>::forEach(void (*func)(T &))
{
	for(unsigned i = 0; i < size; ++i)
	{
		Item *ptr  = m_data[i];

		// On applique la fonction sur chaque lment de la liste
		while(ptr)
		{
			func(ptr->item);
			ptr = ptr->next;
		}
	}
}

template <class K, class T, int size>
template <class Param>
void HashMap<K, T, size>::forEach(void (*func)(const K &, T &, Param), Param param)
{
	for(unsigned i = 0; i < size; ++i)
	{
		Item *ptr  = m_data[i];

		// On applique la fonction sur chaque lment de la liste
		while(ptr)
		{
			func(ptr->key, ptr->item, param);
			ptr = ptr->next;
		}
	}
}

template <class K, class T, int size>
template <class Param>
void HashMap<K, T, size>::forEach(void (*func)(const K &, const T &, Param), Param param) const
{
	for(unsigned i = 0; i < size; ++i)
	{
		Item *ptr  = m_data[i];

		// On applique la fonction sur chaque lment de la liste
		while(ptr)
		{
			func(ptr->key, ptr->item, param);
			ptr = ptr->next;
		}
	}
}

/*****************************************************************************/
/* Dfinition de la sous classe HashMap::Item
/*****************************************************************************/

/*****************************************************************************/
/* Constructeurs et Destructeurs
/*****************************************************************************/

// Classe de base pour toutes les cls non particulires
template <class K, class T, int size>
HashMap<K, T, size>::Item::Item(const K &key, Item *next):
key(key),
next(next)
{
}