#include "IntersectionTrafficLight.h"

#include "Intersection.h"
#include "Agent.h"
#include "Junction.h"
#include "Way.h"
#include <vector>
#include <algorithm>
#include "Road.h"
#include "MediaManager.h"
#include "MemoryLog.h"

#define TIME_ON_HIGHWAY	20.0f
#define TIME_ON_STREET	10.0f
#define DISTANCE_CHECKED 0.5f
#define MAX_ANGLE_TO_ROAD_CONTINUATION 0.0f

IntersectionTrafficLight::IntersectionTrafficLight():
Intersection(),
m_timeSinceLastSwitch(0.0f),
m_safetyTime(3.0f),
m_indiceCurrentsGreen(0),
m_orange(false)
{
}

IntersectionTrafficLight::~IntersectionTrafficLight(void)
{
	for (unsigned i = 0; i < m_synchronizedLights.size(); i++)
	{
		delete m_synchronizedLights[i];
	}
}

void IntersectionTrafficLight::init()
{
	Intersection::init();

	std::vector<const Way*> highwaysIn, streetsIn;

	// on parcourt toutes les ways arrivant sur notre intersection
	for (unsigned i = 0; i < m_junctions.size(); i += 3)
	{
		if (&m_junctions[i]->getWaysIn()[0]->getBegin().getIntersection() != this)
		{
			if (m_junctions[i]->getWaysIn()[0]->getRoadType() == RT_HIGHWAY)
			{
				highwaysIn.push_back(m_junctions[i]->getWaysIn()[0]);
			}
			else
			{
				streetsIn.push_back(m_junctions[i]->getWaysIn()[0]);
			}
		}
	}

	// de l, on cr les groupes synchroniss, en commencant par les highways
	createSynchro(highwaysIn);
	createSynchro(streetsIn);
}

void IntersectionTrafficLight::pUpdate(float timeElapsed)
{
	if (m_timeSinceLastSwitch + timeElapsed > m_switchTime[m_indiceCurrentsGreen] + m_safetyTime)
	{
		m_orange = false;
		m_timeSinceLastSwitch = 0.0f;
		m_indiceCurrentsGreen = (m_indiceCurrentsGreen + 1) % m_switchTime.size();
	}
	else if (m_timeSinceLastSwitch + timeElapsed > m_switchTime[m_indiceCurrentsGreen])
	{
		m_orange = true;
		m_timeSinceLastSwitch += timeElapsed;
	}
	else 
	{
		m_timeSinceLastSwitch += timeElapsed;
	}
}

bool IntersectionTrafficLight::passable(const Agent& agent) const
{
	// si on est dj  l'intrieur du croisement, on peut circuler
	if (	&agent.getCurrentWay()->getBegin().getIntersection() == this
		&&	&agent.getCurrentWay()->getEnd().getIntersection() == this)
	{
		return true;
	}

	// si le feu est rouge ou orange, on ne passe pas
	if (!isGreen(agent.getCurrentDestination()) ||	m_orange)
	{
		return false;
	}

	// on ne passe que si il y a la place...
	if (!enoughSpaceInDestination(agent))
	{
		return false;
	}

	// ... si on veut passer  gauche, on regarde si personne ne passe  droite ...
	Way* out = getWayTo(*agent.getNextIntersection());
	if (	out  // devrait toujours tre le cas 
		&&	m_synchronizedLights[m_indiceCurrentsGreen]->size() == 2 ) // si il n'y a qu'une voie d'active, pas de risques
	{
		// on rcupre la voie face  nous
		Way* front = NULL;
		if (agent.getCurrentWay() == m_junctions[(*m_synchronizedLights[m_indiceCurrentsGreen])[0]]->getWaysIn()[0])
		{
			front = m_junctions[(*m_synchronizedLights[m_indiceCurrentsGreen])[1]]->getWaysIn()[0];
		}
		else
		{
			front = m_junctions[(*m_synchronizedLights[m_indiceCurrentsGreen])[0]]->getWaysIn()[0];
		}

		
		// si on croise la voie d'en face
		if (agentCrossFrontWay(agent, *front, *out))
		{
			// on regarde l'agent le plus proche de la voie d'en face
			const Agent* agentFront = front->agentAhead();
			// si il y a un agent en face, qu'il ne croise pas notre voie et que nous n'avons pas le temps de passer
			if (agentFront)
			{
				const Way* agentFrontOut = getWayTo(*agentFront->getNextIntersection());
				if (	agentFrontOut	// on vrifie qu'il ne soit pas arriv
					&&	!agentCrossFrontWay(*agentFront, *agent.getCurrentWay(), *agentFrontOut)
					&&	front->timeBeforeNextAgent() < TIMETOPASS)
				{
					// on reste sur place
					return false;
				}
			}
		}
	}

	return true;
}

void IntersectionTrafficLight::createSynchro(std::vector<const Way*>& waysList)
{
	if (waysList.size() == 0)
	{
		return;
	}

	std::vector<int> couples;
	std::vector<float> angles;
	std::vector<unsigned> disposed;
	// les listes sont uniformes
	float time = (waysList[0]->getRoadType() == RT_STREET) ? TIME_ON_STREET  : TIME_ON_HIGHWAY;

	// on commence par crer tous les couples disponibles de ways, et de calculer leurs angles
	for (unsigned i = 0; i < waysList.size(); i++)
	{
		D3DXVECTOR2 v1(waysList[i]->getVector().x, waysList[i]->getVector().z);
		for (unsigned j = i+1; j < waysList.size(); j++)
		{
			D3DXVECTOR2 v2(waysList[j]->getVector().x, waysList[j]->getVector().z);
			couples.push_back(i*10 + j);
			angles.push_back(D3DXVec2Dot(&v1, &v2));
		}
		disposed.push_back(waysList[i]->getEnd().getIntersectionIndice());
	}

	// on fait l'opration autant de fois que de couples dans l'intersection
	for (unsigned cpt = 0; cpt < waysList.size() / 2; cpt ++)
	{
		// on cherche le couple ayant l'angle le plus important (proche de -1)
		int indMax = 0;
		for (unsigned i = 1; i < angles.size(); i++)
		{
			if (angles[i] < angles[indMax])
			{
				indMax = i;
			}
		}
		int maxCouple = couples[indMax];
		// on enregistre le couple ...
		if (angles[indMax] < MAX_ANGLE_TO_ROAD_CONTINUATION) // angle valide
		{
			m_synchronizedLights.push_back(new std::vector<unsigned>());
			( (m_synchronizedLights.back()) )->push_back( disposed[maxCouple / 10] );
			( (m_synchronizedLights.back()) )->push_back( disposed[maxCouple % 10] );
			m_switchTime.push_back(time);
		}
		else // angle trop important, on ne cr pas de groupe
		{
			m_synchronizedLights.push_back(new std::vector<unsigned>());
			( (m_synchronizedLights.back()) )->push_back( disposed[maxCouple / 10] );
			m_switchTime.push_back(time);
			m_synchronizedLights.push_back(new std::vector<unsigned>());
			( (m_synchronizedLights.back()) )->push_back( disposed[maxCouple % 10] );
			m_switchTime.push_back(time);
		}
		// ... et on le drfrence
		for (unsigned i = 0; i < angles.size(); i++)
		{
			if (	maxCouple / 10 == couples[i] / 10 || maxCouple / 10 == couples[i] % 10
				||	maxCouple % 10 == couples[i] / 10 || maxCouple % 10 == couples[i] % 10)
			{
				angles[i] = 2.0f;
			}
		}
		disposed[maxCouple / 10] = 999;
		disposed[maxCouple % 10] = 999;
	}

	// si on avait un nombre impair de ways, on doit crer un groupe pour la dernire
	if (waysList.size() % 2 == 1)
	{
		// on trouve l'indice de la way restante
		unsigned indLast;
		for (indLast = 0; indLast < disposed.size(); indLast++)
		{
			if (disposed[indLast] != 999)
			{
				break;
			}
		}
		// on enregistre la way ...
		m_synchronizedLights.push_back(new std::vector<unsigned>());
		( *(m_synchronizedLights.end() - 1) )->push_back(disposed[indLast]);
		// ... on ajoute son temps d'attente  la liste ...
		m_switchTime.push_back(time);
	}
}

bool IntersectionTrafficLight::isGreen(const Junction& junction) const
{
	return	std::find(	m_synchronizedLights[m_indiceCurrentsGreen]->begin(), 
						m_synchronizedLights[m_indiceCurrentsGreen]->end(),
						junction.getIntersectionIndice())
				!= m_synchronizedLights[m_indiceCurrentsGreen]->end()
			&& !m_orange;
}

D3DXVECTOR3 IntersectionTrafficLight::getColor(const Junction& junction) const
{
	if (	std::find(	m_synchronizedLights[m_indiceCurrentsGreen]->begin(), 
						m_synchronizedLights[m_indiceCurrentsGreen]->end(),
						junction.getIntersectionIndice())
				!= m_synchronizedLights[m_indiceCurrentsGreen]->end() )
	{
		// l'intersection est verte ou orange
		if (m_orange)
		{
			return D3DXVECTOR3(1.0f, 0.5f, 0.0f);
		}
		return D3DXVECTOR3(0.0f, 1.0f, 0.0f);
	}
	return D3DXVECTOR3(1.0f, 0.0f, 0.0f);
}

bool IntersectionTrafficLight::agentCrossFrontWay(const Agent& agent, const Way& front, const Way& out) const
{
	unsigned	iIn = agent.getCurrentDestination().getIntersectionIndice(),
				iOut = out.getEnd().getIntersectionIndice(),
				iFront = front.getEnd().getIntersectionIndice();
	return		((iFront < iIn) ? (iFront + m_junctions.size() - iIn) : (iFront - iIn))
			<	((iOut < iIn) ? (iOut + m_junctions.size() - iIn) : (iOut));
}