#include "Pathfinding.h"

#include "Graph.h"
#include "Way.h"
#include "Intersection.h"
#include <fstream>
#include "Junction.h"
#include "MemoryLog.h"

#define INFINITY 1000.0f

Pathfinding::Pathfinding(Graph& g):
m_W(NULL),
m_nextIntersection(0),
m_graph(&g),
m_lastHighwayRatio(-1.0f)
{
}

Pathfinding::~Pathfinding()
{
	delete[] m_nextIntersection;
}

void Pathfinding::setGraph(Graph *graph)
{
	m_graph = graph;
}

void Pathfinding::init()
{
	const unsigned n = m_graph->getIntersections().size();
	m_W = new float[n * n];
	m_nextIntersection = new int[n * n];
	for(unsigned i = 0; i < n; i++)
	{
		m_graph->getIntersections()[i]->setIndex(i);
	}
	for(unsigned i = 0; i < n; i++)
	{
		for(unsigned j = 0; j < n; j++)
		{
			m_W[i * n + j] = (i == j) ? 0.0f : INFINITY;
			m_nextIntersection[i * n + j] = (i == j) ? j : -1;
			const std::vector<Intersection*> &intersections = m_graph->getIntersections()[i]->getNextIntersections();
			for(unsigned k = 0; k < intersections.size(); ++k)
			{
				if(intersections[k]->getIndex() == j)
				{  
					Way *way = m_graph->getIntersections()[i]->getWayTo(*intersections[k]);
					m_W[i * n + j] = (way->getRoadType() == RT_STREET) ? way->getLength() : way->getLength() / m_lastHighwayRatio;
					m_nextIntersection[i * n + j] = j;
					break;
				}
			}
		}
	}
}

void Pathfinding::FloydWarshall(float highwayRatio)
{
	if(m_lastHighwayRatio == highwayRatio)
	{
		return;
	}
	m_lastHighwayRatio = highwayRatio;

	init();
	unsigned n = m_graph->getIntersections().size();
	for(unsigned k = 0; k < n; ++k) {
		for(unsigned i = 0; i < n; ++i) {
			float startLength = m_W[i * n + k];
			int nextIntersection = m_nextIntersection[i * n + k];
			for(unsigned j = 0; j < n; ++j)
			{
				int index = i * n + j;
				float newLength = startLength + m_W[k * n + j];
				if(m_W[index] > newLength)
				{
					m_W[index] = newLength;
					m_nextIntersection[index] = nextIntersection;
				}
			}
		}
	}
	delete[] m_W;
}

float* Pathfinding::getW()const
{
	return m_W;
}

Intersection * Pathfinding::next(const Intersection& current, const Intersection& destination) const
{
	unsigned index = current.getIndex() * m_graph->getIntersections().size() + destination.getIndex();
	return (m_nextIntersection[index] == -1) ? NULL : m_graph->getIntersections()[m_nextIntersection[index]];
}
