#include "Renderer.h"
#include <exception>
#include "MemoryLog.h"

Renderer::Renderer(LPDIRECT3DDEVICE9 device):
m_device(device),
m_matrixStackIndex(0),
m_pool()
{
}

Renderer::~Renderer()
{
}

const D3DXMATRIX * Renderer::getMatrix() const
{
	return &m_matrixStack[m_matrixStackIndex];
}

void Renderer::setMatrix(const D3DXMATRIX &matrix)
{
	m_matrixStack[m_matrixStackIndex] = matrix;
}

void Renderer::multMatrix(const D3DXMATRIX &matrix)
{
	m_matrixStack[m_matrixStackIndex] *= matrix;
}

void Renderer::popMatrix()
{
	--m_matrixStackIndex;
}

void Renderer::pushMatrix()
{
	m_matrixStack[m_matrixStackIndex + 1] = m_matrixStack[m_matrixStackIndex];
	++m_matrixStackIndex;
}

void Renderer::setDeclaration(const Declaration *declaration)
{
	m_device->SetVertexDeclaration(declaration->getDeclaration());
}

void Renderer::drawPrimitives(D3DPRIMITIVETYPE type, unsigned start, unsigned count)
{
	m_shader->begin();
	do
	{
		m_device->DrawPrimitive(type, start, count);
	}
	while(m_shader->nextPass());
	m_shader->end();
}

LPD3DXEFFECT Renderer::createShader(const std::string &src)
{
	LPD3DXEFFECT effect;
	LPD3DXBUFFER error;

	if(FAILED(D3DXCreateEffect(m_device, static_cast<const void *>(src.c_str()), src.length(),
		NULL, NULL, 0, m_pool.getPool(), &effect, &error)))
	{
		std::string msg("Can't create shader:\n");

		if(error)
		{
			msg += static_cast<char *>(error->GetBufferPointer());
		}

		throw std::exception(msg.c_str());
	}

	m_pool.registerSharedParameters(effect);

	return effect;
}

Texture * Renderer::createTexture(const std::string &fileName)
{
	LPDIRECT3DTEXTURE9 texture;

	if(FAILED(D3DXCreateTextureFromFile(m_device, fileName.c_str(), &texture)))
	{
		throw std::exception(("Can't create texture from file " + fileName).c_str());
	}

	return new Texture(texture);
}

DynamicTexture * Renderer::createDynamicTexture(const Image<unsigned short> &image)
{
	LPDIRECT3DTEXTURE9 texture;

	if(FAILED(m_device->CreateTexture(image.getWidth(), image.getHeight(), 0, D3DUSAGE_AUTOGENMIPMAP,
		D3DFMT_L16, D3DPOOL_MANAGED, &texture, NULL)))
	{
		throw std::exception("Can't create dynamic texture");
	}

	return new DynamicTexture(texture, image);
}

void Renderer::setShader(Shader *shader)
{
	m_shader = shader;
}

void Renderer::setTexture(const std::string &name, const Texture *texture)
{
	m_shader->setTexture(name, texture);
}

const ParametersPool & Renderer::getPool() const
{
	return m_pool;
}

Font * Renderer::createFont(const std::string &fontName, unsigned height)
{
	LPD3DXFONT font;

	D3DXCreateFont(m_device, height, 0, FW_NORMAL, 0, FALSE, DEFAULT_CHARSET,
		OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, fontName.c_str(), &font);

	return new Font(font);
}

void Renderer::drawText(const std::string &text, const D3DXVECTOR2 &position, const Font &font, const D3DXVECTOR4 &color)
{
	RECT rc =
	{
		static_cast<LONG>(position.x),
		static_cast<LONG>(position.y),
		static_cast<LONG>(position.x) + 1,
		static_cast<LONG>(position.y) + 1
	};

	font.getD3DPtr()->DrawText(NULL, text.c_str(), -1, &rc, DT_SINGLELINE | DT_NOCLIP, D3DCOLOR_COLORVALUE(color.x, color.y, color.z, color.w));
}

void Renderer::clearZBuffer()
{
	m_device->Clear(0, NULL, D3DCLEAR_ZBUFFER, D3DCOLOR(), 1.0f, 0);
}