#include "RoadNetwork.h"
#include "Road.h"
#include "MediaManager.h"
#include "Renderer.h"
#include "HashMap.h"
#include "MemoryLog.h"

static DeclarationElement elems[] =
{
	{ 0, D3DDECLUSAGE_POSITION, D3DDECLTYPE_FLOAT3 },
	{ 0, D3DDECLUSAGE_COLOR, D3DDECLTYPE_FLOAT3 }
};

RoadNetwork::RoadNetwork(MediaManager &media):
m_vb(NULL),
m_ib(NULL),
m_media(&media)
{
	m_shader = media.getObject<Shader>("Road.fx");
	m_declaration = media.getRenderer().createDeclaration(elems);
}

RoadNetwork::~RoadNetwork()
{
	for(std::vector<Road *>::iterator it = m_roads.begin(); it != m_roads.end(); ++it)
	{
		delete *it;
	}

	for(std::vector<Crossing *>::iterator it = m_crossings.begin(); it != m_crossings.end(); ++it)
	{
		delete *it;
	}

	delete m_declaration;
	delete m_ib;
	delete m_vb;
}

void RoadNetwork::addRoad(Road &road)
{
	m_roads.push_back(&road);
}

void RoadNetwork::addCrossing(Crossing &c)
{
	m_crossings.push_back(&c);
}

Road * RoadNetwork::divideRoad(Road &road, Crossing &c)
{
	Crossing &end = road.getEndCrossing();
	road.setCrossing(end, c);
	end.removeRoad(road);
	c.addRoad(road);
	
	Road *newRoad = new Road(c, end, road.getType());
	c.addRoad(*newRoad);
	end.addRoad(*newRoad);

	return newRoad;
}

void RoadNetwork::commitChange()
{
	recomputeBuffer();
}

void RoadNetwork::render(Renderer &renderer) const
{
	if(m_vb)
	{
		renderer.setShader(m_shader);
		m_shader->setValue("World", renderer.getMatrix());
		renderer.setDeclaration(m_declaration);
		renderer.setVertexBuffer<Vertex>(0, m_vb, 0);
		renderer.drawIndexedPrimitives<unsigned short>(D3DPT_LINELIST, m_ib, 0, m_roads.size());
	}
}

void RoadNetwork::reset()
{
	for(std::vector<Road *>::iterator it = m_roads.begin(); it != m_roads.end(); ++it)
	{
		delete *it;
	}
	m_roads.clear();

	for(std::vector<Crossing *>::iterator it = m_crossings.begin(); it != m_crossings.end(); ++it)
	{
		delete *it;
	}
	m_crossings.clear();

	recomputeBuffer();
}

void RoadNetwork::recomputeBuffer()
{
	HashMap<std::pair<Crossing *, ERoadType>, unsigned, 1024> crossIndex;
	unsigned nbCrossing = 0;

	delete m_vb;
	delete m_ib;

	if(m_roads.empty())
	{
		m_vb = NULL;
		m_ib = NULL;
	}
	else
	{
		for(std::vector<Road *>::iterator it = m_roads.begin(); it != m_roads.end(); ++it)
		{
			ERoadType type = (*it)->getType();
			std::pair<Crossing *, ERoadType> pair(&(*it)->getStartCrossing(), type);
			if(!crossIndex.isExist(pair))
			{
				crossIndex[pair] = nbCrossing++;
			}

			pair.first = &(*it)->getEndCrossing();
			if(!crossIndex.isExist(pair))
			{
				crossIndex[pair] = nbCrossing++;
			}
		}

		m_vb = m_media->getRenderer().createVertexBuffer<Vertex>(nbCrossing);
		m_ib = m_media->getRenderer().createIndexBuffer<unsigned short>(2 * m_roads.size());

		Vertex *vertices = m_vb->lock();
		unsigned short *indices = m_ib->lock();

		for(unsigned i = 0; i < m_roads.size(); ++i)
		{
			ERoadType type = m_roads[i]->getType();
			Crossing *c1 = &m_roads[i]->getStartCrossing();
			Crossing *c2 = &m_roads[i]->getEndCrossing();

			unsigned i1 = crossIndex[std::pair<Crossing *, ERoadType>(c1, type)];
			unsigned i2 = crossIndex[std::pair<Crossing *, ERoadType>(c2, type)];

			vertices[i1].position = c1->getPosition();
			vertices[i2].position = c2->getPosition();

			if(type == RT_STREET)
			{
				vertices[i1].color = D3DXVECTOR3(0.5f, 0.5f, 0.5f);
				vertices[i2].color = D3DXVECTOR3(0.5f, 0.5f, 0.5f);
			}
			else
			{
				vertices[i1].color = D3DXVECTOR3(1.0f, 1.0f, 1.0f);
				vertices[i2].color = D3DXVECTOR3(1.0f, 1.0f, 1.0f);
			}

			indices[2 * i] = static_cast<unsigned short>(i1);
			indices[2 * i + 1] = static_cast<unsigned short>(i2);
		}

		m_ib->unlock();
		m_vb->unlock();
	}
}

const std::vector<Road*>& RoadNetwork::getRoads() const
{
	return m_roads;
}

const std::vector<Crossing*>& RoadNetwork::getCrossings() const
{
	return m_crossings;
}

std::vector<Crossing*>& RoadNetwork::getCrossings()
{
	return m_crossings;
}

void RoadNetwork::sortRoadsOnCrossing()
{
	for(unsigned i = 0; i < m_crossings.size(); ++i)
	{
		m_crossings[i]->sortRoad();
	}
}

void RoadNetwork::removeOneway()
{
	for(unsigned i = 0; i < m_crossings.size(); ++i)
	{
		unsigned index = i;
		while(m_crossings[index]->getRoads().size() == 1)
		{
			const Road *r = m_crossings[index]->getRoads()[0];
			const Crossing *other = &r->getEndCrossing(*m_crossings[index]);
			r->detach();
			m_roads.erase(std::find(m_roads.begin(), m_roads.end(), r));
			delete r;
			delete m_crossings[index];
			m_crossings.erase(m_crossings.begin() + index);
			--i;

			index = std::find(m_crossings.begin(), m_crossings.end(), other) - m_crossings.begin();
		}
	}
}