#include "SoundEngine.h"
#include "SoundException.h"
#include "Speaker.h"
#include "Camera.h"
#include "Debug.h"
#include <xaudio2.h>
#include <algorithm>
#include "MemoryLog.h"

#define TEST_CALL(func, errorMsg) {HRESULT hr; if(FAILED(hr = func)) throw SoundException(hr, errorMsg);}
#define MAX_ACTIVE_SPEAKER 5

SoundEngine::SoundEngine():
m_objectXAudio2(NULL),
m_masterVoice(NULL)
{
	CoInitializeEx(NULL, COINIT_MULTITHREADED);
	TEST_CALL(
		XAudio2Create(&m_objectXAudio2, 0, XAUDIO2_DEFAULT_PROCESSOR),
		"Can't create Audio device");

	unsigned nbDevices;
	int maxChannel = 0;
	int bestDevice = 0;

	m_objectXAudio2->GetDeviceCount(&nbDevices);
	for(unsigned i = 0; i < nbDevices; ++i)
	{
		XAUDIO2_DEVICE_DETAILS deviceDetails;

		m_objectXAudio2->GetDeviceDetails(i, &deviceDetails);

		if(deviceDetails.OutputFormat.Format.nChannels > maxChannel)
		{
			maxChannel = deviceDetails.OutputFormat.Format.nChannels;
			bestDevice = i;
		}
	}

	TEST_CALL(
		m_objectXAudio2->CreateMasteringVoice(&m_masterVoice, maxChannel,
			XAUDIO2_DEFAULT_SAMPLERATE, 0, bestDevice, NULL),
		"Can't create Master Voice");
}

SoundEngine::~SoundEngine()
{
	Assert(m_speakers.empty());

	if(m_masterVoice)
		m_masterVoice->DestroyVoice();
	if(m_objectXAudio2)
		m_objectXAudio2->Release();
}

Speaker * SoundEngine::createSpeaker(const Sound *sound)
{
	Speaker *speaker = new Speaker(sound, m_objectXAudio2, this);
	m_speakers.push_back(speaker);
	return speaker;
}

void SoundEngine::eraseSpeaker(const Speaker *speaker)
{
	for(std::vector<Speaker *>::iterator it = m_speakers.begin(); it != m_speakers.end(); ++it)
	{
		if(*it == speaker)
		{
			m_speakers.erase(it);
			return;
		}
	}
}

struct SpeakerComp
{
	bool operator ()(const Speaker *s1, const Speaker *s2)
	{
		return (s1->isPlaying() && (!s2->isPlaying() || (s1->getRealVolume() > s2->getRealVolume())));
	}
};

void SoundEngine::update(float elapsedTime, const Camera &camera)
{
	const D3DXMATRIX &view = camera.getViewMatrix();
	D3DXVECTOR3 position = -camera.getPosition();
	D3DXVECTOR3 cameraVelocity = (position - m_cameraLastPosition) / elapsedTime;

	for(std::vector<Speaker *>::iterator it = m_speakers.begin(); it != m_speakers.end(); ++it)
	{
		(*it)->computeRealVolume(position);
	}

	std::sort<std::vector<Speaker *>::iterator, const SpeakerComp &>(
		m_speakers.begin(), m_speakers.end(), SpeakerComp());

	for(unsigned i = 0; i < MAX_ACTIVE_SPEAKER && i < m_speakers.size(); ++i)
	{
		m_speakers[i]->activate();
		m_speakers[i]->update(position, view, cameraVelocity, elapsedTime);
	}
	for(unsigned i = MAX_ACTIVE_SPEAKER; i < m_speakers.size(); ++i)
	{
		m_speakers[i]->desactivate();
	}
	m_cameraLastPosition = position;
}