#include "Speaker.h"
#include "Sound.h"
#include "SoundEngine.h"
#include "Debug.h"
#include "APOLoc.h"
#include <cmath>
#include "MemoryLog.h"

#define EAR_DISTANCE 0.5f
#define SOUND_SPEED 793.0f
#define MAX_VOLUME 4.0f

Speaker::Speaker(const Sound *sound, IXAudio2 *device, SoundEngine *engine):
m_sound(sound),
m_engine(engine),
m_position(NULL),
m_volume(1.0f)
{
	IUnknown *effect = static_cast<IXAPO *>(new APOLoc());
	XAUDIO2_EFFECT_DESCRIPTOR descriptor;
	descriptor.InitialState = true;
	descriptor.OutputChannels = 2;
	descriptor.pEffect = effect;
	XAUDIO2_EFFECT_CHAIN chain;
	chain.EffectCount = 1;
	chain.pEffectDescriptors = &descriptor;

	if(FAILED(device->CreateSourceVoice(&m_voice, &sound->getFormat(), 0, XAUDIO2_DEFAULT_FREQ_RATIO, NULL, NULL, &chain)))
	{
		return;
	}

	effect->Release();

	APOLocParams params;
	params.azim = 0.0f;
	params.elev = 0.0f;

	m_voice->SetEffectParameters(0, &params, sizeof(APOLocParams));
}

Speaker::~Speaker()
{
	if(m_voice)
	{
		m_voice->DestroyVoice();
	}
	m_engine->eraseSpeaker(this);
}

void Speaker::play(bool looped)
{
	Assert(m_voice);

	XAUDIO2_BUFFER buffer = { 0 };
	buffer.AudioBytes = m_sound->getDataLength();
	buffer.pAudioData = const_cast<BYTE*>(reinterpret_cast<const BYTE *>(m_sound->getData()));
	buffer.Flags = XAUDIO2_END_OF_STREAM;
	if(looped)
	{
		buffer.LoopBegin = 0;
		buffer.LoopLength = 0;
		buffer.LoopCount = XAUDIO2_LOOP_INFINITE;
	}

	if(FAILED(m_voice->SubmitSourceBuffer(&buffer)))
	{
		return;
	}

	m_looped = looped;
	m_state = DESACTIVATE;
}

void Speaker::stop()
{
	Assert(m_voice);

	m_voice->Stop();
	m_voice->FlushSourceBuffers();
	m_state = STOP;
}

void Speaker::activate()
{
	Assert(m_voice);

	if(m_state == DESACTIVATE)
	{
		m_voice->Start();
		m_state = PLAYING;
	}
}

void Speaker::desactivate()
{
	Assert(m_voice);

	if(m_state == PLAYING)
	{
		m_voice->Stop();
		m_state = DESACTIVATE;
	}
}

void Speaker::computeRealVolume(const D3DXVECTOR3 &position)
{
	if(m_position)
	{
		D3DXVECTOR3 toSpeaker(*m_position - position);
		float distanceSq = D3DXVec3LengthSq(&toSpeaker);
		m_realVolume = min(m_volume / distanceSq, MAX_VOLUME);
	}
}

void Speaker::update(const D3DXVECTOR3 &position, const D3DXMATRIX &viewMatrix, const D3DXVECTOR3 &cameraVelocity, float elapsedTime)
{
	Assert(m_voice);

	if(m_position)
	{
		APOLocParams params;
		D3DXVECTOR3 toSpeaker(*m_position - position);
		D3DXVec3Normalize(&toSpeaker, &toSpeaker);
		
		D3DXVECTOR3 toSpeakerLocal;
		D3DXVec3TransformNormal(&toSpeakerLocal, &toSpeaker, &viewMatrix);
		D3DXVECTOR2 toSpeakerMedian(toSpeakerLocal.z, -toSpeakerLocal.x);
		D3DXVec2Normalize(&toSpeakerMedian, &toSpeakerMedian);
		params.azim = std::acos(toSpeakerMedian.x);
		if(toSpeakerMedian.y > 0.0f)
		{
			params.azim = -params.azim;
		}
		params.elev = std::asin(toSpeakerLocal.y);
		m_voice->SetEffectParameters(0, &params, sizeof(APOLocParams));

		m_voice->SetVolume(m_realVolume);

		D3DXVECTOR3 speakerVelocity = (*m_position - m_lastPosition) / elapsedTime;
		float scalarSpeakerVelocity = -D3DXVec3Dot(&speakerVelocity, &toSpeaker);
		float scalarCameraVelocity = -D3DXVec3Dot(&cameraVelocity, &toSpeaker);
		m_voice->SetFrequencyRatio((SOUND_SPEED - scalarCameraVelocity) / (SOUND_SPEED - scalarSpeakerVelocity));
		m_lastPosition = *m_position;
	}

	if(m_state == PLAYING && !m_looped)
	{
		XAUDIO2_VOICE_STATE state;
		m_voice->GetState(&state);
		if(state.BuffersQueued == 0)
		{
			m_voice->Stop();
			m_state = STOP;
		}
	}
}

void Speaker::attachPosition(const D3DXVECTOR3 &position)
{
	m_position = &position;
	m_lastPosition = *m_position;
}

void Speaker::setVolume(float volume)
{
	m_volume = volume;
}

const D3DXVECTOR3 & Speaker::getPosition() const
{
	return *m_position;
}

float Speaker::getRealVolume() const
{
	return m_realVolume;
}

bool Speaker::isPlaying() const
{
	return m_state != STOP;
}