#include "Way.h"

#include "Junction.h"
#include "Agent.h"
#include <cmath>
#include "Debug.h"
#include "Intersection.h"
#include "MemoryLog.h"

static const float TYPE_SPEED_LIMIT[] = {0.5f, 0.9f};

Way::Way(Junction* const begin, Junction* const end, ERoadType type, unsigned zone):
m_begin(begin),
m_end(end),
m_type(type),
m_zone(zone)
{
	m_vector = end->getPosition() - begin->getPosition();
	m_length = D3DXVec3Length(&m_vector);
	D3DXVec3Normalize(&m_vector, &m_vector);
	begin->addWayOut(this);
	end->addWayIn(this);
}

Way::~Way(void)
{
}

void Way::addAgent(const Agent &agent)
{
	m_agents.push_back(&agent);
}

void Way::removeAgent(const Agent &agent)
{
	// thoriquement, l'agent supprim doit tre le premier
	for (unsigned i = 0; i < m_agents.size(); i++)
	{
		if (m_agents[i] == &agent)
		{
			m_agents.erase(m_agents.begin() + i);
			return;
		}
	}
	// on n'a pas supprim d'agent, il y a problme
	Assert(true);
}

const Junction& Way::getBegin() const
{
	return *m_begin;
}

const Junction& Way::getEnd() const
{
	return *m_end;
}

float Way::getLength() const
{
	return m_length;
}

const D3DXVECTOR3& Way::getVector() const
{
	return m_vector;
}

const Agent* Way::agentAhead(const Agent& agent) const
{
	float dist = 0.0f;
	const Agent *res;
	res = NULL;
	for (unsigned i = 0; i < m_agents.size(); i++)
	{
		if (	m_agents[i] != &agent
			&&	m_agents[i]->getDistanceFromCurrentJunction() > dist	// on se trouve entre l'agent...
			&&	m_agents[i]->getDistanceFromCurrentJunction() < agent.getDistanceFromCurrentJunction() // ...et l'obstacle en cours
		   )
		{
			res = m_agents[i];
			dist = m_agents[i]->getDistanceFromCurrentJunction();
		}
	}
	return res;
}

const Agent* Way::agentAhead(float positionOnRoad) const
{
	float dist = 0.0f;
	const Agent *res;
	res = NULL;
	for (unsigned i = 0; i < m_agents.size(); i++)
	{
		if (	m_agents[i]->getDistanceFromCurrentJunction() > dist	// on se trouve entre l'agent...
			&&	m_agents[i]->getDistanceFromCurrentJunction() < positionOnRoad // ...et l'obstacle en cours
		   )
		{
			res = m_agents[i];
			dist = m_agents[i]->getDistanceFromCurrentJunction();
		}
	}
	return res;
}

const Agent* Way::agentAhead() const
{
	if (m_agents.size() == 0)
	{
		return NULL;
	}
	const Agent* result = m_agents[0];
	for (unsigned i = 1; i < m_agents.size(); i++)
	{
		if (result->getDistanceFromCurrentJunction() > m_agents[i]->getDistanceFromCurrentJunction())
		{
			result = m_agents[i];
		}
	}

	return result;
}

const Agent* Way::agentBehind() const
{
	if (m_agents.size() == 0)
	{
		return NULL;
	}
	const Agent* result = m_agents[0];
	for (unsigned i = 1; i < m_agents.size(); i++)
	{
		if (result->getDistanceFromCurrentJunction() < m_agents[i]->getDistanceFromCurrentJunction())
		{
			result = m_agents[i];
		}
	}

	return result;
}

const std::vector<const Agent*>& Way::getAgents() const
{
	return m_agents;
}

bool Way::isCrossed(const Way& way) const
{
	// si les deux way partagent une junction, elles se croient
	if (	m_begin == &way.getBegin() || m_begin == &way.getEnd()
		||	m_end == &way.getBegin() || m_end == &way.getEnd())
	{
		return true;
	}

	D3DXVECTOR3	vec = m_begin->getPosition() - way.getBegin().getPosition();

	float den = m_vector.x * way.getVector().z - way.getVector().x * m_vector.z;
	float num = (way.getVector().x * vec.z - vec.x * way.getVector().z);
	if (den == 0.0f)
	{
		// les segments sont parallles, ils se croisent si ils sont confondus
		return num == 0.0f;
	}
	float r = num / den;
	float s = (m_vector.x * vec.z - vec.x * m_vector.z) / den;

	return r >= 0.0f && r <= 1.0f && s >= 0.0f && s <= 1.0f;
}

float Way::timeBeforeNextAgent() const
{
	return pTimeBeforeNextAgent(3);
}

float Way::pTimeBeforeNextAgent(unsigned execNb) const
{
	float result = 1000.0f;

	const Agent* ahead = agentAhead();
	if (ahead)
	{
		float speed = ahead->getCurrentSpeed();
		result = ahead->getDistanceFromCurrentJunction() / speed;
	}
	else if (execNb > 1)
	{
		// pas d'agents, on regarde sur le voies prcdentes
		for (unsigned i = 0; i < m_begin->getWaysIn().size(); i++)
		{
			float tmp = m_begin->getWaysIn()[i]->pTimeBeforeNextAgent(execNb-1);
			if (tmp < result)
			{
				result = tmp;
			}
		}
	}

	return result;
}

Junction& Way::getBegin()
{
	return *m_begin;
}

Junction& Way::getEnd()
{
	return *m_end;
}

float Way::getSpeedLimit() const
{
	return TYPE_SPEED_LIMIT[m_type];
}

ERoadType Way::getRoadType() const
{
	return m_type;
}

bool Way::isZoneType(ERoadZoneType type) const
{
	return (m_zone & type) != 0;
}